/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_ALFcnl_D_H_
#define HCL_ALFcnl_D_H_

#include "HCL_Functional_d.h"
#include "HCL_LinearSolver_d.h"
#include "HCL_CompLinearOp_d.h"
#include "HCL_LinCombLinearOp_d.h"
#include "HCL_Adjoint_d.h"
#include "HCL_BiLinearOp_d.h"
#include "HCL_Op_d.h"
#include "Table.h"

/** {\bf HCL\_EvalALFcnl\_d} is the evaluation object for the class
    {\bf HCL\_ALFcnl\_d}, which implements an Augmented Lagrangian functional.
    For documentation, see the base class {\bf HCL\_EvaluateFunctional\_d}
    and the functional class {\tt HCL\_ALFcnl\_d}. */
class HCL_EvalALFcnl_d: public HCL_EvaluateFunctional_d
{
   private:
      HCL_EvaluateFunctional_d * feval;
      HCL_Vector_d * lambda;
      HCL_Vector_d * Slambda;
      HCL_Vector_d * mu;
      HCL_EvaluateOp_d * Heval;
      HCL_Vector_d * Hx;
      HCL_Vector_d * SHx;
      HCL_LinearOp_d * DHx;
      HCL_BiLinearOp_d * D2Hx;
      HCL_LinearOp_d * S;
      HCL_Vector_d * LamPlusRhoHx;
      HCL_Vector_d * SLamPlusRhoHx;
      double rho;
   public:

   // Default constructor---disabled.
      HCL_EvalALFcnl_d();
   // Copy constructor---disabled.
      HCL_EvalALFcnl_d( const HCL_EvalALFcnl_d & );
   /** Usual constructor.  The arguments are: a pointer to the evaluation
       object of the objective functional, a pointer to the Lagrange
       multiplier vector, a pointer to the evaluation object of the operator
       defining the equality constraints, and a pointer to
       a linear operator defining a scaled norm for the constraint space
       (this pointer may be NULL). (If $H:X\rightarrow Z$ is the constraint
       operator, then the scaled norm on $Z$ is $\|z\|_S=\sqrt{(z,Sz)}$.  The
       operator $S$ must be self-adjoint and positive definite.) */
      HCL_EvalALFcnl_d( HCL_EvaluateFunctional_d * fev,
                            HCL_Vector_d * l,
                            HCL_EvaluateOp_d * Hev,double r,
                            HCL_LinearOp_d * SS );

   // Destructor
      virtual ~HCL_EvalALFcnl_d();

   /// Domain returns a reference to the domain of the underlying function.
      virtual HCL_VectorSpace_d & Domain() const { return feval->Domain(); }

   /// Value computes the value of the function.
      virtual double Value() const;
   /// Gradient computes the gradient of the function.
      virtual void Gradient( HCL_Vector_d & g ) const;
   /// Hessian computes the Hessian of the function.
      virtual HCL_LinearOp_d * Hessian() const;
   /// fValue computes the value of the objective function.
      virtual double fValue() const;
   /** ConstraintViolation computes the image of the (equality) constraint
       operator. */
      virtual void ConstraintViolation( HCL_Vector_d & h );
   /// GradientLag computes the gradient of the Lagrangian.
      virtual void GradientLag( HCL_Vector_d & g );
   /** {\bf MultiplierUpdate} computes the Hestenes-Powell update to the
       multiplier. */
      virtual void MultiplierUpdate( HCL_Vector_d & l );

   /// Write prints out some useful information about the object.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_ALFcnl\_d} implements an Augmented Lagrangian functional; it is
    intended for use with {\bf HCL\_CMinAL\_d}. If $f$ is a real-valued
    (objective) functional with domain $X$, and $H:X\rightarrow Z$ is a
    (equality constraint) operator, then the augmented Lagrangian functional
    $F$ is defined by

       $F(x) = f(x) + (\lambda,H(x)) + \frac{\rho}{2}\|H(x)\|_Z^2$,

    where $\lambda$ is the Lagrange multiplier and $\rho$ is the penalty
    parameter.  This class also allows the inner product on $Z$ to be
    scaled by a self-adjoint, positive definite operator $S:Z\rightarrow Z$:

       $F(x) = f(x) + (\lambda,SH(x)) + \frac{\rho}{2}(H(x),SH(x))_Z$.

    {\bf HCL\_ALFcnl\_d} has the usual properties (methods) of the
    {\bf HCL\_Functional\_d} base class; the documentation of the base class
    should be consulted for more information.
*/

class HCL_ALFcnl_d: public HCL_Functional_d
{
   private:
      Table * ParamTable;
      HCL_Functional_d * f;
      HCL_Op_d * H;
      HCL_Vector_d * lam;
      HCL_LinearOp_d * S;
   public:

   // Default constructor---disabled.
      HCL_ALFcnl_d();
   // Copy constructor---disabled.
      HCL_ALFcnl_d( const HCL_ALFcnl_d & F );
   /** Usual constructor.  The arguments are: a pointer to the objective
       functional, a pointer to the constraint operator, a pointer to the
       (initial) Lagrange multiplier, (optionally) an initial penalty
       parameter, and (optionally) a pointer to a self-adjoint positive-
       definite linear operator $S$, used to scale the inner product on
       the constraint space.  The default value for the penalty parameter
       is $1.0$.  Note that the penalty parameter can be updated using
       the {\bf Parameters()} method as follows:
           Parameters().PutValue( "PenParam",newvalue );
	   */
      HCL_ALFcnl_d( HCL_Functional_d * ff,HCL_Op_d * HH,
                    HCL_Vector_d * l,double rho=1.0,
                    HCL_LinearOp_d * SS=NULL );
   // Destructor.
      virtual ~HCL_ALFcnl_d();

   /// Domain space access.
      virtual HCL_VectorSpace_d & Domain() const { return f->Domain(); }

    /** {\bf MaxStep} computes the longest vector from x in the direction dir
	that will lie in the domain of the functional, i.e. it computes
        the greatest $a$ such that $x + a*dir$ lies in the domain. */
      virtual double MaxStep( const HCL_Vector_d & x, 
			      const HCL_Vector_d &dir) const;

   /// Access to parameter table
      Table & Parameters() { return *ParamTable; }

   /** {\bf Evaluate} creates an "evaluation" object, which knows how to
       compute the function value and gradient at the given x. */
      virtual HCL_EvaluateFunctional_d * 
                  Evaluate( const HCL_Vector_d & x ) const;

   /// Write prints out some useful information about the function. 
      virtual ostream & Write( ostream & str ) const;
};

#endif
