/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_ADJOINT_D_H_
#define HCL_ADJOINT_D_H_

#include "HCL_LinearOp_d.h"
#include "HCL_LinearSolver_d.h"

/** {\bf HCL\_Adjoint\_d} creates the adjoint operator of an
{\bf HCL\_LinearOp\_d} as an operator in its own right.  Note that
this object points to the original linear operator.  The primary
methods are precisely those of {\bf HCL\_LinearOp\_d}; in particular, the
image methods such as {\bf Image} and {\bf AdjImage} work by simply
calling the appropriate methods of the original
linear operator. */
class HCL_Adjoint_d: public HCL_LinearOp_d
{
   private:
      HCL_LinearOp_d * L;
   public:

   // Default constructor---disabled.
      HCL_Adjoint_d();
   // Copy constructor---disabled.
      HCL_Adjoint_d( const HCL_Adjoint_d & x );
   /// Usual constructor.  Takes a pointer to the original operator.
      HCL_Adjoint_d( HCL_LinearOp_d * l );
   // Destructor.
      virtual ~HCL_Adjoint_d();

   /** Domain space access.  Returns a reference to the domain,
       which is a vector space. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access.  Returns a reference to the range,
       which is a vector space. */
      virtual HCL_VectorSpace_d & Range() const;

   /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                HCL_Vector_d & x ) const;
   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b z$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b z$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b y$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b y$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /// {\bf Write} calls the Write method of the original operator.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_Normal\_d} creates the normal operator of a
{\bf HCL\_LinearOp\_d} as an operator in its own right.  Note that this
object points to the original linear operator.  The primary methods are
precisely those of {\bf HCL\_LinearOp\_d}; in particular, the various
image methods are implemented to call the corresponding methods of the
original linear operator. */
class HCL_Normal_d: public HCL_LinearOp_d
{
   private:
      HCL_LinearOp_d * L;
   public:

   // Default constructor---disabled.
      HCL_Normal_d();
   // Copy constructor---disabled.
      HCL_Normal_d( const HCL_Normal_d & x );
   /// Constructs the object from a {\bf HCL\_LinearOp\_d} pointer.
      HCL_Normal_d( HCL_LinearOp_d * l );
   // Destructor.
      virtual ~HCL_Normal_d();

   /** Domain space access.  Returns a reference to the domain,
       which is a vector space. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access.  Returns a reference to the range,
       which is a vector space. */
      virtual HCL_VectorSpace_d & Range() const;

   /// {\bf Image} computes the action of the operator on x, giving z.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & z ) const;
   /// {\bf AdjImage} computes the adjoint of the operator on x, giving z.
      virtual void AdjImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & z ) const;
   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /// {\bf Write} invokes the {\bf Write} method of the original operator.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_InvLinearOp\_d} creates the inverse operator of a
{\bf HCL\_LinearOp\_d} as an operator in its own right.  If the original
operator does not have an implemented {\bf InvImage} method, then a
{\bf HCL\_LinearSolver\_d} must be provided.  Note that this object points
to the original linear operator and linear solver. */
class HCL_InvLinearOp_d: public HCL_LinearOp_d
{
   private:
      HCL_LinearOp_d * L;
      HCL_LinearSolver_d * lsolver;
   public:

   // Default constructor---disabled.
      HCL_InvLinearOp_d();
   // Copy constructor---disabled.
      HCL_InvLinearOp_d( const HCL_InvLinearOp_d & x );
   /** Constructs the inverse operator based on the forward operator with
       {\bf InvImage} method; pointer version. */
      HCL_InvLinearOp_d( HCL_LinearOp_d * l );
   /** Constructs the inverse operator based on the forward operator
       and a linear solver; pointer version. */
      HCL_InvLinearOp_d( HCL_LinearOp_d * l, HCL_LinearSolver_d *ls );
   // Destructor.
      virtual ~HCL_InvLinearOp_d();

   /** Domain space access.  Returns a reference to the domain,
       which is a vector space. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access.  Returns a reference to the range,
       which is a vector space. */
      virtual HCL_VectorSpace_d & Range() const;

   /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the adjoint of the operator on x, giving z.
      virtual void AdjImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & z ) const;
   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage( const HCL_Vector_d & y,
                                HCL_Vector_d & x ) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b z$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b z$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b y$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b y$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /// {\bf Write} invokes the {\bf Write} method of the original operator.
      virtual ostream & Write( ostream & str ) const;
};

#endif
