/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_BASIS_D_H_
#define HCL_BASIS_D_H_

#include "HCL_Rn_d.h"
#include "HCL_ProductVector_d.h"
#include "HCL_Arch_d.h"
#include "maxmin_d.h"

/** {\bf HCL\_Basis\_d} is a concrete class implementing a basis for
a finite-dimensional subspace of a vector space.  Essentially, this
class implements a finite sequence of vectors, together with some operations
that are natural to perform on bases.

The provided operations are:
\begin{itemize}
\item {\bf Dim} Return the dimension of the subspace;
\item {\bf LinearCombination} Forms a linear combination of the basis vectors;
the weights are specified as the components of an instance of
{\bf HCL\_RnVector\_d}.
\item {\bf ModifiedGramSchmidt} Orthonormalizes the basis using the
modified Gram Schmidt algorithm.
\item {\bf ModifiedGramSchmidtWithReduction} Orthonormalizes the basis using
the modified Gram Schmidt algorithm; reduces the number of vectors in the
basis if it is detected that the vectors are not linearly independent.
\end{itemize}
*/

class HCL_Basis_d: public HCL_Base
{
   private:
      HCL_VectorSpace_d * sp;
      HCL_Vector_d ** vlist;
      int n;
      int nmax;
   public:

   // Default constructor---disabled.
      HCL_Basis_d();
   // Copy constructor---disabled.
      HCL_Basis_d( const HCL_Basis_d & );
   /** Usual constructor; needs the space and the dimension of the subspace.
       When the basis is constructed, space is allocated for all n vectors.
       The vector should then be assigned values before the basis is used.
       Note that access to the individual vectors is provided through the
       operator() index operator. */
      HCL_Basis_d( HCL_VectorSpace_d * sp,int m );
   // Destructor.
      ~HCL_Basis_d();

   /// Returns a reference to the space.
      HCL_VectorSpace_d & Space() const;
   /// Returns dimension of subspace.
      int Dim();
   /// Returns a reference to the ith basis vector.
      HCL_Vector_d & operator()( int i );
   /// Computes a linear combination of the basis vectors.
      void LinearCombination( const HCL_RnVector_d & a,
                                    HCL_Vector_d & x ) const;
   /// Performs modified Gram-Schmidt to orthogonalize the basis.
      int ModifiedGramSchmidt();
   /** Performs modified Gram-Schmidt to orthogonalize the basis;
       reduces dimension of space if linear dependence is discovered. */
      int ModifiedGramSchmidtWithReduction();
   /** Computes and displays on standard out the matrix of inner products
       to check orthogonality. */
      void OrthoCheck() const;

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;

};

/** {\bf HCL\_BasisProductSpace\_d} is a concrete class implementing a basis
for a finite-dimensional subspace of a product space.  Essentially, this
class implements a finite sequence of vectors, together with some operations
that are natural to perform on bases.

The provided operations are:
\begin{itemize}
\item {\bf Dim} Return the dimension of the subspace;
\item {\bf LinearCombination} Forms a linear combination of the basis vectors;
the weights are specified as the components of an instance of
{\bf HCL\_RnVector\_d}.
\item {\bf ModifiedGramSchmidt} Orthonormalizes the basis using the
modified Gram Schmidt algorithm.
\item {\bf ModifiedGramSchmidtWithReduction} Orthonormalizes the basis using
the modified Gram Schmidt algorithm; reduces the number of vectors in the
basis if it is detected that the vectors are not linearly independent.
\end{itemize}
*/

class HCL_BasisProductSpace_d: public HCL_Base
{
   private:
      HCL_ProductSpace_d * sp;
      HCL_ProductVector_d ** vlist;
      int n;
   public:

   // Default constructor---disabled.
      HCL_BasisProductSpace_d();
   // Copy constructor---disabled.
      HCL_BasisProductSpace_d( const HCL_BasisProductSpace_d & );
   /** Usual constructor; needs the space and the dimension of the subspace.
       When the basis is constructed, space is allocated for all n vectors.
       The vector should then be assigned values before the basis is used.
       Note that access to the individual vectors is provided through the
       operator() index operator. */
      HCL_BasisProductSpace_d( HCL_ProductSpace_d * sp,int m );
   // Destructor.
      ~HCL_BasisProductSpace_d();

   /// Returns a reference to the space.
      HCL_ProductSpace_d & Space() const;
   /// Returns dimension of subspace.
      int Dim();
   /// Returns a reference to the ith basis vector.
      HCL_ProductVector_d & operator()( int i );
   /// Computes a linear combination of the basis vectors.
      void LinearCombination( const HCL_RnVector_d & a,
                                    HCL_ProductVector_d & x ) const;
   /// Performs modified Gram-Schmidt to orthogonalize the basis.
      int ModifiedGramSchmidt();
   /** Performs modified Gram-Schmidt to orthogonalize the basis;
       reduces dimension of space if linear dependence is discovered. */
      int ModifiedGramSchmidtWithReduction();
   /** Computes and displays on standard out the matrix of inner products
       to check orthogonality. */
      void OrthoCheck() const;

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
