/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_BIGMFCNL_H_
#define HCL_BIGMFCNL_H_

#include "HCL_Functional_d.h"
#include "HCL_GenericProductVector_d.h"
#include "HCL_Rn_d.h"
#include "Table.h"

/** {\bf HCL\_BigMFcnlHessian\_d} implements the Hessian of the big-M
functional represented by {\bf HCL\_BigMFcnl\_d} as an object derived
from {\bf HCL\_LinearOp\_d}.  For more information, see the documentation
for {\bf HCL\_BigMFcnl\_d} and for the base class {\bf HCL\_LinearOp\_d}.
*/

class HCL_BigMFcnlHessian_d: public HCL_LinearOp_d
{
   private:
      HCL_GenericProductSpace_d * dom;
      HCL_LinearOp_d * A;
      double M,msum;
      HCL_Vector_d * Ate;
      HCL_Vector_d * temp;
      HCL_Vector_d * mask;
   public:

  // Default constructor---disabled.
      HCL_BigMFcnlHessian_d();
  // Copy constructor---disabled.
      HCL_BigMFcnlHessian_d( const HCL_BigMFcnlHessian_d & L );
  /** Usual constructor; takes a pointer to the domain, a pointer to the
      linear operator, the parameter $M$, and a "mask" vector (indicates
      which components of the residual are positive). */
      HCL_BigMFcnlHessian_d( HCL_GenericProductSpace_d * d,
                             HCL_LinearOp_d * AA,
                             double MM,HCL_Vector_d * msk );
  // destructor
      virtual ~HCL_BigMFcnlHessian_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is a {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
  /** Range space access. Returns a reference to the range of
      the operator, which is a {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;
  /// Image compute the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
  /** Debugging information.*/
      virtual ostream & Write( ostream & str ) const;
};


/** {\bf HCL\_EvalBigMFcnl\_d} is the "evaluation object" for the functional
class {\bf HCL\_BigMFcnl\_d}, which implements a "big-M" functional for the
problem of finding a vector $x$ satisfying $Ax-b+\epsilon\le 0$
(that is, the goal is to find a point $x$ strictly satisfying $Ax-b\le 0$).
The big-M functional is defined by

$f_{\epsilon,M}(x)=\frac{1}{2}\sum_{i=1}^m\max\{(Ax-b)_i-(\tau-\epsilon),0\}^2
                  +\frac{M}{2}\tau^2.$

For a general description of the methods of an evaluation class, see the
documentation for the base class {\bf HCL\_EvaluateFunctional}.  See also
the documentation for {\bf HCL\_BigMFcnl\_d}.
*/

class HCL_EvalBigMFcnl_d: public HCL_EvaluateFunctional_d
{
   private:
      HCL_Vector_d * x;
      double tau;
      HCL_GenericProductSpace_d * dom;
      HCL_LinearOp_d * A;
      double M,eps;

      HCL_Vector_d * res;
      HCL_Vector_d * mask;
   public:

   // Default constructor---disabled.
      HCL_EvalBigMFcnl_d();
   // Copy constructor---disabled.
      HCL_EvalBigMFcnl_d( const HCL_EvalBigMFcnl_d & );
   /** Usual constructor; takes (a reference to) the input vector, a pointer
       to the linear operator, a pointer to the right-hand side vector,
       and the parameters $M$ and $\epsilon$. */
      HCL_EvalBigMFcnl_d( const HCL_GenericProductVector_d & x,
                          HCL_LinearOp_d * AA,HCL_Vector_d * bb,
                          double M,double eps );

   // Destructor.
      virtual ~HCL_EvalBigMFcnl_d();

   /// {\bf Domain} returns a reference to the domain space.
      virtual HCL_VectorSpace_d & Domain() const;

   /// {\bf Value} computes the value of the functional.
      virtual double Value() const;
   /// {\bf Gradient} computes the gradient of the functional.
      virtual void Gradient( HCL_Vector_d & g ) const;
   /** {\bf Hessian} computes the Hessian of the functional.  Note that
       this method creates an instance of {\bf HCL\_LinearOp\_d}, which the
       calling routine must delete using {\bf HCL\_delete}. */
      virtual HCL_LinearOp_d * Hessian() const;
   /// {\bf Write} outputs a description of the object.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_BigMFcnl\_d} implements a "big-M" functional for the
problem of finding a vector $x$ satisfying $Ax-b+\epsilon\le 0$
(that is, the goal is to find a point $x$ strictly satisfying $Ax-b\le 0$).
The big-M functional is defined by

$f_{\epsilon,M}(x,\tau)=
     \frac{1}{2}\sum_{i=1}^m\max\{(Ax-b)_i-(\tau-\epsilon),0\}^2
                  +\frac{M}{2}\tau^2.$

Note that the domain (the space of $(x,\tau)$) is implemented as a product
space using {\bf HCL\_GenericProductSpace\_d}.

For a general description of the methods of a functional class, see the
documentation for the base class {\bf HCL\_Functional}.
*/

class HCL_BigMFcnl_d: public HCL_Functional_d
{
   private:
      HCL_GenericProductSpace_d * dom;
      HCL_LinearOp_d * A;
      HCL_Vector_d * b;
      double M,eps;

   public:  

   // Default constructor---disabled.
      HCL_BigMFcnl_d();
   // Copy constructor---disabled.
      HCL_BigMFcnl_d( const HCL_BigMFcnl_d & );
   /** Usual constructor; takes a pointer to the linear operator, a pointer
       to the right-hand side vector, the scalar $M$, and the scalar
       $\epsilon$. */
      HCL_BigMFcnl_d( HCL_LinearOp_d * AA,
                      HCL_Vector_d * bb,double MM,double eeps );
   // destructor
      virtual ~HCL_BigMFcnl_d();

   /** Domain space access.  Returns a reference to the
       {\bf HCL\_VectorSpace\_d} which forms the domain of the functional.*/
      virtual HCL_VectorSpace_d & Domain() const;

   /** {\bf Evaluate} creates an "evaluation" object, which knows how
       to compute the function value, gradient,and Hessian at the given x.*/
      virtual HCL_EvaluateFunctional_d * 
                  Evaluate( const HCL_Vector_d & x ) const;

   /** Debugging information.*/
      virtual ostream & Write( ostream & str ) const;
};

#endif
