#ifndef HCL_BLOCKBILINEAROP_D_H__
#define HCL_BLOCKBILINEAROP_D_H__

#include "HCL_BiLinearOp_d.h"
#include "HCL_ProductVector_d.h"

/** {\bf HCL\_BlockBiLinearOp\_d} implements a bilinear operator of
the following form:
\TEX{\[
B:(X_1\times X_2\times\cdots\times X_J)\times
  (Y_1\times Y_2\times\cdots\times Y_I)\rightarrow Z,
\]},
where
\TEX{\[
B(x,y)=\sum_{i=1}^I\sum_{j=1}^J B_{ij}(x_j,y_i)
\]}
and each \TEX{$B_{ij}:X_j\times Y_i\rightarrow Z$} is a bilinear operator.
The primary purpose for this class is to represent the second derivative
of an operator defined on a product space in terms of it partial derivatives.
In that case, \TEX{$I=J$}, \TEX{$Y_i=X_i$},
\TEX{\[
B_{ij}=D_{x_ix_j}^2F(x),
\]}
and \TEX{$B=D^2F(x)$}.

For more details on bilinear operators, see the documentation for
{\bf HCL\_BiLinearOp\_d}. */

class HCL_BlockBiLinearOp_d: public HCL_BiLinearOp_d
{
   private:
      int I,J;   // J is the number of components in the first variable;
                 // I is the number of components in the second variable.
      HCL_BiLinearOp_d ** blocks;
      HCL_ProductSpace_d * X;
      HCL_ProductSpace_d * Y;
      HCL_VectorSpace_d * Z;

   protected:
      virtual void PartialAdjImage( int i,const HCL_Vector_d & xy,
                                              const HCL_Vector_d & z,
                                                    HCL_Vector_d & yx ) const;

   public:

   // Default constructor---disabled.
      HCL_BlockBiLinearOp_d();
   // Copy constructor---disabled.
      HCL_BlockBiLinearOp_d( const HCL_BlockBiLinearOp_d & );
   /** Usual constructor; just needs domains and range (blocks must be
       set using {\bf Set}. */
      HCL_BlockBiLinearOp_d( HCL_ProductSpace_d * d1,HCL_ProductSpace_d * d2,
                             HCL_VectorSpace_d * r );
   // Destructor.
      ~HCL_BlockBiLinearOp_d();

   /** {\bf Set} sets the (i,j) bilinear operator. Note that this method
       must be called for each (i,j) after the block bilinear operator is
       constructed in order to initialize it. */
      void Set( int i,int j,HCL_BiLinearOp_d * Bij );
   /** {\bf Initialized} is a boolean that indicates whether the operator
       has been initialized or not. */
      int Initialized() const;
   /** {\bf IsSet} is a boolean that indicates whether the (i,j) block
       has been initialized. */
      int IsSet( int i,int j ) const;
   /** The index operator returns a reference to the (i,j) block. */
      HCL_BiLinearOp_d & operator()( int i,int j ) const
      {
         return *blocks[(i-1)*J+j-1];
      }

   /** Access to the domain of the first argument.  Returns a reference
       to an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain1() const { return *X; }
   /** Access to the domain of the second argument.  Returns a reference
       to an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain2() const { return *Y; }
   /** Access to the range.  Returns a reference to an
       {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return *Z; }

   /// Image computes the action of the operator on the pair (x,y), giving z.
      virtual void Image( const HCL_Vector_d & x,const HCL_Vector_d & y,
                                HCL_Vector_d & z ) const;

  /** {\bf Write} prints a description of the of the operator. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
