/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _HCL_BLOCKLINEAROP_D_H
#define _HCL_BLOCKLINEAROP_D_H

#include "Table.h"
#include "HCL_LinearOp_d.h"
#include "HCL_ProductVector_d.h"

#define COLUMNOP 1
#define ROWOP 2

/** {\bf HCL\_BlockLinearOp\_d} implements a 2D array of linear operators
acting on a product vector space.  For example, if the array is 2x2, the
operator is of the form
\TEX{
\[
(x,y) \mapsto (Ax+By,Cx+Dy).
\]}
In general, {\bf HCL\_BlockLinearOp\_d} represents an operator
\TEX{\[
L:X_1\times X_2\times\cdots\times X_n\rightarrow
Y_1\times Y_2\times\cdots\times Y_m
\]}
defined by
\TEX{\[
(Lx)_i=\sum_{j=1}^nL_{ij}x_j.
\]}

In order to construct a general block operator, one must specify the
domain and range, which are product spaces represented by instances of
{\bf HCL\_ProductSpace\_d}.  The blocks \TEX{$L_{ij}$} are then specified
{\em by rows} using the {\bf SetNext} method, or in any order using the
{\bf Set} method.  Here is an example of creating a block operator using the
{\bf SetNext} method.  In this example, the operator is \TEX{$3\times 2$},
that is, the domain space has two factors, while the range space has three.
\begin{verbatim}
// X and Y are instances of HCL_ProductSpace_d with 2 and 3 factors,
// respectively.  The operator L will map X into Y.

   HCL_BlockLinearOp_d L( &X,&Y );

// L11, L12, L21, L22, L31, L32 are instances of HCL_LinearOp_d with the
// appropriate domains and ranges.  For example, L21 must map X(1) into Y(2).
// Note that, when using the SetNext method, the blocks are specified by
// rows.

   L.SetNext( &L11 );
   L.SetNext( &L12 );
   L.SetNext( &L21 );
   L.SetNext( &L22 );
   L.SetNext( &L31 );
   L.SetNext( &L32 );

// L is now ready to use.
\end{verbatim}
The same thing can be accomplished using the {\bf Set} method:
\begin{verbatim}
// X and Y are instances of HCL_ProductSpace_d with 2 and 3 factors,
// respectively.  The operator L will map X into Y.

   HCL_BlockLinearOp_d L( &X,&Y );

// L11, L12, L21, L22, L31, L32 are instances of HCL_LinearOp_d with the
// appropriate domains and ranges.  For example, L21 must map X(1) into Y(2).
// Note that, when using the Set method, the blocks can be specified in
// any order.

   L.Set( 1,1,&L11 );
   L.Set( 2,1,&L21 );
   L.Set( 3,2,&L32 );
   L.Set( 2,2,&L22 );
   L.Set( 3,1,&L31 );
   L.Set( 1,2,&L12 );

// L is now ready to use.
\end{verbatim}
It is not intended that {\bf SetNext} and {\bf Set} both be used in
initializing a single operator (although they should work---{\bf SetNext}
sets the first unitialized block, scanning across the rows).

It is also possible to use {\bf HCL\_BlockLinearOp\_d} to represent a
"row" operator or a "column" operator.  A row operator maps a product space
into an ordinary vector space, while a column operator maps an ordinary
vector space into a product space.  This option is provided so that
users of the class do not need to construct a product space with one
factor to represent an ordinary vector space.  To construct a row or
column operator, pass the constructor the product space, the ordinary vector
space, and either the ROWOP or the COLUMNOP flag (these flags are defined
in the header file).
*/
class HCL_BlockLinearOp_d: public HCL_LinearOp_d
{
   private:
      int num,rows,cols;
      int RowOp,ColOp;
      HCL_VectorSpace_d * dom1;
      HCL_VectorSpace_d * range1;
      HCL_ProductSpace_d * dom;
      HCL_ProductSpace_d * range;
      HCL_LinearOp_d ** blist;
   public:  

   // Default constructor---disabled.
      HCL_BlockLinearOp_d();
   // Copy constructor---disabled.
      HCL_BlockLinearOp_d( const HCL_BlockLinearOp_d & L );
   /** Usual constructor; needs the domain and range spaces.  The operators
       are defined with either the {\bf SetNext} method or the {\bf Set}
       method. */
      HCL_BlockLinearOp_d( HCL_ProductSpace_d * d,
                              HCL_ProductSpace_d * r );
   /** Special constructor for the case of a ``row'' or ``column'' operator.
       The input flag must have value ROWOP or COLUMNOP. If the operator
       is a row, then the product space is the domain, and the vector space
       is the range; if the operator is a column, the reverse is true. */
      HCL_BlockLinearOp_d( HCL_ProductSpace_d * d,
                           HCL_VectorSpace_d * r,int flag );
   // Destructor.
      virtual ~HCL_BlockLinearOp_d();

   /** {\bf Initialized} is a boolean method indicating whether all blocks
       in the operator have been initialized. */
      int Initialized() const;

   /** {\bf IsSet} is a boolean indicated whether the (i,j) block has been
       initialized. */
      int IsSet( int i,int j ) const;

   /// Domain space access.  Returns a reference to the domain.
      virtual HCL_VectorSpace_d & Domain() const;
   /// Range space access.  Returns a reference to the range.
      virtual HCL_VectorSpace_d & Range() const;

   /// The () operator provides access to the individual operators.
      HCL_LinearOp_d & operator()( int i,int j ) const
      {
         return *(blist[(i-1)*Cols()+j-1]);
      }

   /// {\bf Rows} returns the number of rows in the array.
      int Rows() const { return rows; }
   /// {\bf Cols} returns the number of columns in the array.
      int Cols() const { return cols; }

   /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & y ) const;
   /** {\bf InvImage} computes the action of the inverse on y, giving x.
       This method is not implemented in this initial version of the class,
       and returns an error message. */
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y.  This method is not implemented in this initial version of
       the class, and returns an error message. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /** {\bf SetNext} sets the "next" block to L; when this method is used,
       the blocks must be specified by rows. */
      void SetNext( HCL_LinearOp_d * L );
   /** {\bf Set} sets the (i,j) block to L. */
      void Set( int i,int j,HCL_LinearOp_d * L );

   /// {\bf Write} invokes the {\bf Write} method for the blocks.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_DiagBlockLinearOp\_d} implements a diagonal array of linear
operators (with adjoints and inverses) acting on a product vector space.
For example, if the array is 2x2, the operator is of the form
$(x,y) \mapsto (Ax,By)$.

In order to construct a diagonal block operator, one must specify the
domain and range, which are product spaces represented by instances of
{\bf HCL\_ProductSpace\_d}; these product spaces must have the same number of
factors.  The diagonal blocks are then specified (in order) using the
{\bf SetNext} method, or in any order using the
{\bf Set} method.  Here is an example of creating a diagonal block operator
using the {\bf SetNext} method.  In this example, the operator is
\TEX{$3\times 3$}, that is, the domain space has three factors, as does the
range.
\begin{verbatim}
// X and Y are instances of HCL_ProductSpace_d, each with 3 factors.
// The operator L will map X into Y.

   HCL_DiagBlockLinearOp_d L( &X,&Y );

// L11, L22, L33 are instances of HCL_LinearOp_d with the
// appropriate domains and ranges.  For example, L22 must map X(2) into Y(2).
// Note that, when using the SetNext method, the blocks are specified in
// order.

   L.SetNext( &L11 );
   L.SetNext( &L22 );
   L.SetNext( &L33 );

// L is now ready to use.
\end{verbatim}
The same thing can be accomplished using the {\bf Set} method:
\begin{verbatim}
// X and Y are instances of HCL_ProductSpace_d, each with 3 factors.
// The operator L will map X into Y.

   HCL_BlockLinearOp_d L( &X,&Y );

// L11, L22, L33 are instances of HCL_LinearOp_d with the
// appropriate domains and ranges.  For example, L21 must map X(2) into Y(2).
// Note that, when using the Set method, the blocks can be specified in
// any order.

   L.Set( 3,&L33 );
   L.Set( 1,&L11 );
   L.Set( 2,&L22 );

// L is now ready to use.
\end{verbatim}
It is not intended that {\bf SetNext} and {\bf Set} both be used in
initializing a single operator (although they should work---{\bf SetNext}
sets the first unitialized block). */

class HCL_DiagBlockLinearOp_d: public HCL_LinearOp_d
{
   private:
      int num;
      HCL_ProductSpace_d * dom;
      HCL_ProductSpace_d * range;
      HCL_LinearOp_d ** blist;
   public:  

   // Default constructor---disabled.
      HCL_DiagBlockLinearOp_d();
   // Copy constructor---disabled.
      HCL_DiagBlockLinearOp_d( const HCL_DiagBlockLinearOp_d & L );
   /** Usual constructor; needs the domain and range spaces.  The operators
       are defined with the {\bf SetNext} method or the {\bf Set} method. */
      HCL_DiagBlockLinearOp_d( HCL_ProductSpace_d * d,
                               HCL_ProductSpace_d * r );
   // Destructor.
      virtual ~HCL_DiagBlockLinearOp_d();

   /// Domain space access.
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
   /// Domain space access.
      virtual HCL_VectorSpace_d & Range() const { return *range; }

   /** {\bf Initialized} is a boolean method indicating whether the blocks
       have all been set. */
      int Initialized() const;

   /** {\bf IsSet} is a boolean indicated whether the ith block has been
       initialized. */
      int IsSet( int i ) const;

   /// The () operator provides access to the operators on the diagonal.
      HCL_LinearOp_d & operator()( int i ) const { return *blist[i-1]; }
   /// Number returns the number of factors in the domain (and range).
      int Number() const { return dom->Number(); }

   /// Image computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// AdjImage computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;
   /// {\bf NormalImage} computes the action of the normal on x, giving z.
      virtual void NormalImage( const HCL_Vector_d & x,
                                      HCL_Vector_d & z ) const;
   /// InvAdjImage computes the action of the inverse of the adjoint.
      virtual void InvAdjImage( const HCL_Vector_d & x,
                                      HCL_Vector_d & y ) const;
   /// InvImage computes the action of the inverse on y, giving x.
      virtual void InvImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b z$}.  */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b z$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b z$}.  */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b y$}. */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b y$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b y$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /** SetNext sets the next operator on the diagonal to L. */
      void SetNext( HCL_LinearOp_d * L );
   /** {\bf Set} sets the (i,i) block to L. */
      void Set( int i,HCL_LinearOp_d * L );

   /// Write prints out the operators on the diagonal.
      virtual ostream & Write( ostream & str ) const;
};

#endif
