/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_COMPLINEAROP_D_H_
#define HCL_COMPLINEAROP_D_H_

#include "HCL_LinearOp_d.h"

/** {\bf HCL\_CompLinearOp\_d} is a concrete class implementing a composition
of two or more linear operators.   That is, if \TEX{$A_1,A_2,\ldots,A_n$}
are linear operators represented by instances of {\bf HCL\_LinearOp\_d}, then
this class represents the composition \TEX{$A_nA_{n-1}\cdots A_1$}.  */
class HCL_CompLinearOp_d: public HCL_LinearOp_d
{
   private:
      int num;
      int numinit;
      HCL_LinearOp_d ** ops;
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
   public:

   // Default constructor---disabled.
      HCL_CompLinearOp_d();
   // Copy constructor---disabled.
      HCL_CompLinearOp_d( const HCL_CompLinearOp_d & L );
   /** Usual constructor.  Requires that the domain and range space be
       given (for error checking), along with the number of factors in
       the composition.  The operators are added one-by-one using
       the {\bf SetNext} method, beginning with the "innermost" operator.
       That is, if the composition is \TEX{$A_nA_{n-1}\cdots A_1$}, then the
       method {\bf SetNext} must be invoked on \TEX{$A_1,A_2,\ldots,A_n$} in
       that order.  */
      HCL_CompLinearOp_d( HCL_VectorSpace_d * d,HCL_VectorSpace_d * r,int n );
   /** Provided for convenience.  This constructor makes it easy to
       create a composition with two factors.  Pointers to the
       two operators (A1 and A2) must be given; the composition is then
       \TEX{$A_2A_1$}. */
      HCL_CompLinearOp_d( HCL_LinearOp_d * A1,
                          HCL_LinearOp_d * A2 );
   // Destructor.
      virtual ~HCL_CompLinearOp_d();

   /** Domain space access. Returns a reference to the vector space
       forming the domain of the operator. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
   /** Range space access. Returns a reference to the vector space
       forming the range of the operator. */
      virtual HCL_VectorSpace_d & Range() const { return *range; }

   /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;
   /** {\bf NormalImage} computes the action of the normal operator on y,
       giving x. */
      virtual void NormalImage( const HCL_Vector_d & y,
                                      HCL_Vector_d & x ) const;
   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */ 
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b z$}. */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b z$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b z$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b y$}. */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b y$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b y$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /** {\bf SetNext} is used for initializing the terms in the composition.
       A pointer to the next operator (L) must be given.  Note that the
       operators must be specified in order, beginning with the "innermost".
       That is, if the composition is \TEX{$A_nA_{n-1}\cdots A_1$}, then the
       method {\bf SetNext} must be invoked on \TEX{$A_1,A_2,\ldots,A_n$} in
       that order.  */
      void SetNext( HCL_LinearOp_d * L );

   /// Write prints out the operators in the composition.
      virtual ostream & Write( ostream & str ) const;
};

#endif
