/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_DIAGFUNCTIONAL_D_H_
#define HCL_DIAGFUNCTIONAL_D_H_

#include "HCL_ProductVector_d.h"
#include "HCL_LinearOp_d.h"
#include "HCL_FunctionalProductDomain_d.h"
#include "HCL_ScaleLinearOp_d.h"
#include "HCL_BlockOp_d.h"

/** {\bf HCL\_EvalDiagFcnl\_d} is the evaluation object for 
{\bf HCL\_DiagFunctional\_d}, which implements a separable function defined
on a product space.  This class is derived from
{\bf HCL\_EvalFunctionalProductDomain\_d}.

For more information, see the documentation for {\bf HCL\_DiagFunctional\_d}
and {\bf HCL\_EvalFunctionalProductDomain\_d}. */

class HCL_EvalDiagFcnl_d: public HCL_EvalFunctionalProductDomain_d
{
   private:
      int num;
      HCL_ProductSpace_d * dom;
      HCL_EvaluateFunctional_d ** evallist;
      HCL_ZeroLinearOp_d ** Zlist;
   public:

   // Default constructor---disabled.
      HCL_EvalDiagFcnl_d();
   // Copy constructor---disabled.
      HCL_EvalDiagFcnl_d( const HCL_EvalDiagFcnl_d & );
   /** The constructor requires a vector \TEX{$x$} in the domain of the
       functional, as well as the list of functionals defining the sum.
       Note that the constructor makes an independent copy of the input
       vector. */
      HCL_EvalDiagFcnl_d( HCL_ProductVector_d & x,
                          HCL_Functional_d ** flist );
   // Destructor.
      virtual ~HCL_EvalDiagFcnl_d();

   /// Domain returns a reference to the domain of the underlying function.
      HCL_VectorSpace_d & Domain() const { return *dom; }

   /// Value computes the value of the function.
      virtual double Value() const;
   /// PartialGradient computes the gradient of the function.
      virtual void PartialGradient( int i,HCL_Vector_d & g ) const;
   /// Hessian computes the Hessian of the function.
      virtual HCL_LinearOp_d * Hessian() const;
   /** HessianRef computes the Hessian of the function and returns
       a reference to it.  This causes the Evaluation object to manage
       the storage. */
      virtual const HCL_LinearOp_d & HessianRef() const;
   /// HessianBlock computes a block of the Hessian operator of the function.
      virtual HCL_LinearOp_d * HessianBlock( int i,int j ) const;
   /** {\bf HessianBlockRef} returns a reference to the (i,j) block of
       the Hessian. */
      virtual const HCL_LinearOp_d & HessianBlockRef( int i,int j ) const;

   /** {\bf Write} invokes the {\bf Write} methods of the evaluation objects
       of the terms in the sum. */
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_DiagFunctional\_d} implements a separable function defined on a
product space, that is, a function 
\TEX{$f:X\rightarrow {\bf{\rm R}},\ X=X_1\times X_2 \times ... \times X_n$},
of the form \TEX{$f(x)=f_1(x_1)+f_2(x_2)+...+f_n(x_n)$}.
The terms in the sum must be represented by instances of
{\bf HCL\_Functional\_d}; this class then puts the terms together to form
a single, separable, function.  The domain of the functional must be
represented by an instance of {\bf HCL\_ProductSpace\_d}.

For more information about functionals defined on product spaces, see
the documentation for {\bf HCL\_FunctionalProductDomain\_d}. */

class HCL_DiagFunctional_d: public HCL_FunctionalProductDomain_d
{
   private:
      int num;
      HCL_ProductSpace_d * dom;
      HCL_Functional_d ** flist;

   public:

   // Default constructor---disabled.
      HCL_DiagFunctional_d();
   // Copy constructor---disabled.
      HCL_DiagFunctional_d( const HCL_DiagFunctional_d & F );
   /** The constructor requires (a pointer to) the functional's domain,
       which must be a product space.  The terms defining the functional,
       which are functionals defined on the factors of the domain, must
       be specified using the {\bf SetNext} method. */
      HCL_DiagFunctional_d( HCL_ProductSpace_d * d );
   /// Destructor
      virtual ~HCL_DiagFunctional_d();

   /** Domain space access.  Returns a reference to the
       {\bf HCL\_VectorSpace\_d} which forms the domain of the functional.
       Note that, in the current design, the domain of a functional
       must be an entire vector space, although many nonlinear functionals
       are only defined on an open subset of a vector space.  This is
       a flaw in the design of HCL which may be corrected in some future
       version of the software.  A partial work-around to this problem
       is provided by the {\bf MaxStep} method. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
    /** {\bf MaxStep} computes the longest vector from x in the direction dir
	that will lie in the domain of definition of the functional, i.e.
        it computes the greatest a such that x + a*dir lies in the domain.
        By default (i.e. unless this virtual function is implemented
        in a derived class), the domain is assumed to be the whole space,
        and MaxStep return the largest floating point number.  This method
        is used in some of the optimization algorithms and provides a partial
        solution to the problem that many functionals are not defined on an
        entire vector space, but rather only on a subset. */
      virtual double MaxStep( const HCL_Vector_d & x, 
                              const HCL_Vector_d &dir) const;

   /** {\bf Initialized} is a boolean method indicating whether the terms
       in the sum have all been set.  The method returns 1 if the terms
       have all been set and 0 otherwise. */
      int Initialized() const;

   /** Access to components.  If {\bf f} is an instance of
       {\bf HCL\_DiagFunctional\_d}, then {\bf f(i)} returns a reference to
       the {\bf HCL\_Functional\_d} representing the ith term in the sum. */
      virtual HCL_Functional_d & operator()( int i ) const;
   /// Number of terms in the sum.
      virtual int Number() const { return num; }

   /** {\bf Evaluate} creates an "evaluation" object, which knows how to
       compute the function value, gradient, and Hessian at the given x. */
      virtual HCL_EvaluateFunctional_d * 
                          Evaluate( const HCL_Vector_d & x ) const;

   /** {\bf SetNext} allows the user to specify the next term in the
       sum defining the funcional.  If the functional has n terms, then
       {\bf SetNext} must be called n times before the functional can be
       used.  The terms must be set in the order they appear in the
       {\bf HCL\_ProductSpace\_d} defining the domain. */
      void SetNext( HCL_Functional_d * f );

   /// {\bf Write} invokes the {\bf Write} method for the terms in the sum.
      virtual ostream & Write( ostream & str ) const;
};

#endif
