/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_ITEIGSOLVERS_D_H
#define HCL_ITEIGSOLVERS_D_H

#include <iostream.h>
#include <fstream.h>

#include "Table.h"
#include "timer_d.h"

#include "HCL_MatrixOps_d.h"
#include "HCL_LinearOp_d.h"
#include "HCL_Rn_d.h"

/** Abstract base class for iterative eigenvalue solvers. */
class HCL_ItEigSolver_d: public HCL_Base
{
 public:
   
   /**@name        Methods      */
   //@{

   /** Solve A*x = lambda*x 
       @param nev Number of eigenvalues wanted.  On output this contains
		  the number of eigenvalues actually computed.
       @param A Linear operator A.
       @param x Starting vector for iteration.
    */
   virtual int Solve( int &nev, HCL_LinearOp_d &A, HCL_Vector_d &x ) = 0;

   /** Access to eigenvalues and eigenvectors computed by Solve().
       @param i Which eigenpair to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @param eigi Returns the imaginary part of the ith eigenvalue.
       @param vr Returns the real part of the ith eigenvector.
       @param vi Returns the imaginary part of the ith eigenvector.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   virtual int EigPair(int i, double &eigr, double &eigi, HCL_Vector_d *vr,
	   HCL_Vector_d *vi) = 0;
   /** Access to (real part of) eigenpairs computed by Solve().
       @param i Which eigenvalue to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @param vr Returns the real part of the ith eigenvector.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   virtual int EigPair(int i, double &eigr, HCL_Vector_d *vr ) = 0;
   /** Access to eigenvalues computed by Solve().
       @param i Which eigenvalue to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @param eigi Returns the imaginary part of the ith eigenvalue.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   virtual int Eigenvalue( int i, double &eigr, double &eigi ) = 0;
   /** Access to (real part of) eigenvalues computed by Solve().
       @param i Which eigenvalue to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   virtual int Eigenvalue( int i, double &eigr ) = 0;
   //@}
};



// Implicitly Restarted Arnoldi method


// Based on the Fortran code in ARPACK written by:
//    Danny Sorensen               Phuong Vu
//    Richard Lehoucq              CRPC / Rice University
//    Dept. of Computational &     Houston, Texas
//    Applied Mathematics
//    Rice University
//    Houston, Texas

/** {\bf HCL\_IRArnoldi\_d} implements the implicitly restarted Arnoldi
algorithm for solving the (generalized) eigenvalue problem.
The algorithm computes approximations to a few eigenpairs of of a linear
operator.  Optionally, this can be done with respect to the semi-inner product
defined by a positive semi-definite linear operator B. 
This is an adaptation of the algorithm inplemented in ARPACK. \\
References: \\
   R.B. Lehoucq, D.C. Sorensen \& C. Yang, "ARPACK Users Guide" \\
*/
class HCL_IRArnoldi_d : public HCL_ItEigSolver_d
{
 private:

   enum TypeEnum {  // Problem modes
      Regular = 0,
      ShitfR = 1,
      ShiftI = 2,
      RealPt = 3,
      ImagPt = 4
   };
   
   Table * ParamTable;
   
   double *dr, *di;
   int vsize;
   HCL_Vector_d **V;


   double *ritzr, *ritzi, *bounds, *workl, *heigr, *heigi, *hbds, *workev;
   int *select;
   double rnorm;
   HCL_Vector_d *tmpvec, *tmpvec2;

   
   GeneralMatrix_d *Q;
   GeneralMatrix_d *R;
   GeneralMatrix_d *H;

   HCL_VectorSpace_d *dom;

   void *shiftproc;

   int nconv,ncv,nev,kplusp;
   int type, ishift;
   
   int TermCode;
   int DispFlag;
   int DispPrecision;
   int OldDispPrecision;
   int MaxIter;
   int EigVectors, SchurVectors;
   double Tol;
   char Which[32];


   // timing variables

   double OP_time, B_time, Arnoldi_time, Factor_time, HessEig_time, QR_time;
   double EigPair_time, Reorthog_time;

   char *OP_str, *B_str;

   int iter, OP_count, B_count, Reorthog_count; 

   
   // private methods

   int IRArnoldi(HCL_LinearOp_d * OP, HCL_LinearOp_d * B, 
		 HCL_Vector_d *resid);
   int ComputeEigPairs(double sigmar, HCL_Vector_d *resid,
		 GeneralMatrix_d *q);

   int Arnoldi(HCL_LinearOp_d & A, HCL_LinearOp_d * B,
         GeneralMatrix_d & H, HCL_Vector_d **V,
         int k, int np, int &npout,
         HCL_Vector_d & resid, double & rnorm,
         HCL_Vector_d * B_rnorm, int type);

   void Initialize();

   int _Solve( int &nev, HCL_Vector_d &resid, 
	       HCL_LinearOp_d *OP, HCL_LinearOp_d *M,
	       double sigmar, int type);


   // Free local storage.
   void Clean();

   /** Solve A*x = lambda*M*x with M symmetric positive semi-definite,
       and using shift-and-invert mode; NOTE: THERE IS A BUG IN THIS
       METHOD---IT DOESN'T WORK.  
       Eigenvalues near the real shift 'sigmar' will be emphasized during the 
       iteration.  Note that the user must supply the linear operator 'OP', 
       rather than 'A'.
       @param nev Number of eigenvalues wanted.  On output this contains
		  the number of eigenvalues actually computed.
       @param OP Linear operator Inv[A-sigmar*M]*M.
       @param sigmar Real shift to apply.
       @param x Vector to start Arnoldi iteration.  This is changed by
		the function, and does not have a meaningful value on return.
       @return The value of TermCode (See the Parameters section of 
	       this document.)
    */
   int Solve( int &nev, HCL_LinearOp_d &OP, HCL_LinearOp_d &M,
	      double sigmar, HCL_Vector_d &x);



 public:

   enum TermCodeEnum {
      Success = 0,
      TooManyIterations = 1,
      ZeroInitialVector = 2,
      LapackEigFailed = 3,
      InvariantSubspace = 4,
      // RefinementFailed = 5,
      NoShiftsAvailable = 6,
      ComputeSchurVectsFailed = 7,
      ComputeEigVectsFailed = 8
   };



   /**@name        Methods      */
   //@{

   /* Constructors and destructor */
   /** Constructor.
       @param fname Name of file to read parameters from or NULL.  */
      HCL_IRArnoldi_d(char *fname = NULL);
   // Destructor
      virtual ~HCL_IRArnoldi_d();

   /** Access to parameter table. Algorithmic parameters can be accessed
       with Parameters().GetValue("NAME",val) or changed with
       Parameters().PutValue("NAME",val). */
   virtual Table & Parameters() const { return *ParamTable; }


   /** Solve A*x = lambda*x.
       @param nev Number of eigenvalues wanted.  On output this contains
		  the number of eigenvalues actually computed.
       @param A Linear operator.
       @param x Vector to start Arnoldi iteration.  This is changed by
		the function, and does not have a meaningful value on return.
       @return The value of TermCode (See the Parameters section of 
	       this document.)
    */
   int Solve( int &nev, HCL_LinearOp_d &A, HCL_Vector_d &x );
   
   /** Solve A*x = lambda*x using shift-and-invert mode.  
       Eigenvalues near the real shift 'sigmar' will be emphasized during the 
       iteration.  Note that the user must supply the linear operator 'OP', 
       rather than 'A'.
       @param nev Number of eigenvalues wanted.  On output this contains
		  the number of eigenvalues actually computed.
       @param OP Linear operator Inv[A-sigmar*I].
       @param sigmar Real shift to apply.
       @param x Vector to start Arnoldi iteration.  This is changed by
		the function, and does not have a meaningful value on return.
       @return The value of TermCode (See the Parameters section of 
	       this document.)
    */
   int Solve( int &nev, HCL_LinearOp_d &OP, double sigmar, 
	      HCL_Vector_d &resid );

   /** Solve A*x = lambda*M*x, with M symmetric positive semi-definite.
       Note that the user must supply the linear operator 'OP', 
       rather than 'A'.
       @param nev Number of eigenvalues wanted.  On output this contains
		  the number of eigenvalues actually computed.
       @param OP Linear operator Inv[M]*A.
       @param x Vector to start Arnoldi iteration.  This is changed by
		the function, and does not have a meaningful value on return.
       @return The value of TermCode (See the Parameters section of 
	       this document.)
    */
   int Solve( int &nev, HCL_LinearOp_d &OP, HCL_LinearOp_d &M,
	      HCL_Vector_d &x );
   
   /** Access to eigenvalues computed by Solve().
       @param i Which eigenvalue to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @param eigi Returns the imaginary part of the ith eigenvalue.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   virtual int Eigenvalue( int i, double &eigr, double &eigi );
   /** Access to (real part of) eigenvalues computed by Solve().
       @param i Which eigenvalue to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   virtual int Eigenvalue( int i, double &eigr );
   /** Access to eigenvalues and eigenvectors computed by Solve().
       @param i Which eigenpair to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @param eigi Returns the imaginary part of the ith eigenvalue.
       @param vr Returns the real part of the ith eigenvector.
       @param vi Returns the imaginary part of the ith eigenvector.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   int EigPair(int i, double &eigr, double &eigi, HCL_Vector_d *vr,
	   HCL_Vector_d *vi);
   /** Access to real part of eigenpairs computed by Solve().
       @param i Which eigenpair to return.  (Must be between 1 and the number
		of eigenvalues computed in Solve())
       @param eigr Returns the real part of the ith eigenvalue.
       @param vr Returns the real part of the ith eigenvector.
       @return Nonzero if 'i' is not valid or if Solve() has not been
	       called.
    */
   int EigPair(int i, double &eigr, HCL_Vector_d *vr );


   //@}

      /** Prints description of the object. */
      virtual ostream & Write(ostream &) const;


#ifdef UNDEFINED  // for documentation purposes only

   /**@name         Input Parameters
       In a parameter file, these names may be prepended by "ItEigSolver::" or "IRArnoldi::" */
   //@{

     /** Dimension of Krylov subspace.  The algorithm will allocate 
	 roughly this many vectors in the domain of the problem. Increasing 
	 this value will generally improve convergence.  SubspaceDim must
	 satisfy 2*nev < SubspaceDim < Dim(A).
	 \\
	 Default = 2*nev + 2
      */
      int SubspaceDim;

     /** Eigenvalue selection criteria.  This determines which part of the
	 spectrum the algorithm will attempt to locate.  Valid values are: \\
	 "SM" - Smallest magnitude \\
	 "LM" - Largest magnitude \\
	 "SR" - Smallest real part \\
	 "LR" - Largest real part \\
	 "SI" - Smallest imaginary part \\
	 "LI" - Largest imaginary part \\
	 \\
	 Default = "LM"
      */
      char Which[2];

      /** Stopping criterion.  A ritz value is considered converged if the
	  relative accuracy is less than Tol.
	  \\
	  Default = machine-epsilon
	*/
      double Tol;

      /** Maximum number of Arnoldi restarts.
	  \\
	  Default = 100
       */
      int MaxItn;

      /** Flag to compute eigenvectors.  (0 - Don't compute; 1 - Compute)
	  \\
	  Default = 1
       */
      int EigVectors;

      /** Flag to compute Schur vectors.  (0 - Don't compute; 1 - Compute)
	  Note that setting this to 0 is a useful optimization in the case of 
	  symmetric operators, where Schur vectors are actually eigenvectors.
	  \\
	  Default = 1
       */
      int SchurVectors;

      /** Display level.  This determines how much information should be 
	  displayed during the execution of the algorithm.  Common values
	  are: \\
	  0 - No output \\
	  1 - One-line summary output \\
	  2 - Detailed header and footer with timing information \\
	  3 - Per-iteration statistics \\

	  Default = 0
       */
      int DispFlag;
   //@}

   /**@name         Output Parameters */
   //@{
      /** Termination code.  Value is one of: \\
	  
	  0 - Success \\
	  1 - Too many restart iterations taken \\
	  2 - Initial vector is zero \\
	  3 - LAPACK Eig failed (shouldn't happen) \\
	  4 - Algorithm stalled-out with an invariant subspace \\
	  6 - No shifts available for shifted QR algorithm (shouldn't happen) \\
	  7 - Computing the Schur vectors failed \\
	  8 - Computing the eigenvectors failed 
       */
      int TermCode;
   //@}

#endif
};

#endif
