/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef HCL_LNLOp_D_H_
#define HCL_LNLOp_D_H_

#include "HCL_ScaleLinearOp_d.h"
#include "HCL_Op_d.h"

/** {\bf HCL\_ZeroBiLinearOp\_d} implements a bilinear operator 
\TEX{$B:X \times Y  \rightarrow Z$} which is identically zero:
\TEX{$B(x,y) \equiv 0$}.
*/

class HCL_ZeroBiLinearOp_d: public HCL_BiLinearOp_d
{
   private:
      HCL_VectorSpace_d * dom1;
      HCL_VectorSpace_d * dom2;
      HCL_VectorSpace_d * range;
   public:

   // Default constructor---disabled.
      HCL_ZeroBiLinearOp_d();
   // Copy constructor---disabled.
      HCL_ZeroBiLinearOp_d( const HCL_ZeroBiLinearOp_d & );
   /// Usual constructor; just needs pointers to the domains and range.
      HCL_ZeroBiLinearOp_d( HCL_VectorSpace_d * d1,
                            HCL_VectorSpace_d * d2,
                            HCL_VectorSpace_d * r );
   // Destructor.
      virtual ~HCL_ZeroBiLinearOp_d();

   /** Access to the domain of the first argument.  Returns a reference
       to a {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain1() const;
   /** Access to the domain of the second argument.  Returns a reference
       to a {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain2() const;
   /** Access to the range.  Returns a reference to a
       {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;

   /// Image computes the action of the operator on the pair (x,y), giving z.
      virtual void Image( const HCL_Vector_d & x,const HCL_Vector_d & y,
                                HCL_Vector_d & z ) const;
   /// Access to Op1 as a linear operator.
      virtual HCL_LinearOp_d * Op1( const HCL_Vector_d & x ) const;
   /// Access to Op2 as a linear operator.
      virtual HCL_LinearOp_d * Op2( const HCL_Vector_d & y ) const;

   /** Access to the "partial adjoint" of the bilinear operator.
       Assume B:XxY --> Z is the bilinear operator.
       Then Op3 returns the operator x |---> (B(x))'z.
       Its adjoint is the operator y |---> (B(.)y)'z */
      virtual HCL_LinearOp_d * Op3( const HCL_Vector_d & z ) const;

  /** Debugging information. */
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_EvalLNLOp\_d} is the evaluation object for {\bf HCL\_LNLOp\_d}.
It creates a nonlinear operator (i.e. an instance of {\bf HCL\_Op\_d})
from a linear operator (i.e. an instance of {\bf HCL\_LinearOp\_d}).

For more details, see the documentation for the classes
{\bf HCL\_EvaluateFunctional\_d} and {\bf HCL\_LNLOp\_d}. */
class HCL_EvalLNLOp_d: public HCL_EvaluateOp_d
{
   private:
      HCL_LinearOp_d * L;
      HCL_Vector_d * x;
   public:

   // Default constructor---disabled.
      HCL_EvalLNLOp_d();
   // Copy constructor---disabled.
      HCL_EvalLNLOp_d( const HCL_EvalLNLOp_d & );
   /** Usual constructor; just needs a pointer to the linear operator and
       a reference to the vector x at which the operator is to be evaluated.*/
      HCL_EvalLNLOp_d( HCL_LinearOp_d * LL,const HCL_Vector_d & xx );
   // Destructor.
      virtual ~HCL_EvalLNLOp_d();

   /** Domain space access. Returns a reference to an
       {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access. Returns a reference to an
       {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;

   /// {\bf Image} computes the image of the operator.
      virtual void Image( HCL_Vector_d & y ) const;
   /** {\bf ImageAdd} computes \TEX{$y\leftarrow aF(x)+by$}. */
      virtual void ImageAdd( HCL_Vector_d & y,
                             double a=1.0,double b=1.0 ) const;
   /** {\bf ImageAdd} computes \TEX{$y\leftarrow aF(x)+bz$}. */
      virtual void ImageAdd( const HCL_Vector_d & z,HCL_Vector_d & y,
                             double a=1.0,double b=1.0 ) const;

   /** {\bf Deriv} computes the derivative, as a linear operator
       of the operator.  Note that this method constructs an
       {\bf HCL\_LinearOp\_d}, which the calling program is responsible
       for deleting with {\bf HCL\_delete}. */
      virtual HCL_LinearOp_d * Deriv() const;

   /** {\bf SecondDeriv} computes the second derivative of the operator.
       The second derivative is a bilinear operator which knows how to
       compute all relevant adjoints. */
      virtual HCL_BiLinearOp_d * SecondDeriv() const;

   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_LNLOp\_d} creates a nonlinear op from a linear one.
That is, it returns a nonlinear operator object (i.e. an instance of
{\bf HCL\_Op\_d}) which has precisely the same mapping behaviour as
a linear operator (i.e. an instance of
{\bf HCL\_LinearOp\_d}). This is useful for instance in creating 
formally nonlinear least squares functions using {\bf HCL_LS\_d} 
(which can be fed to various
nonlinear optimization algorithms),which
are actually linear least squares functions. */
class HCL_LNLOp_d: public HCL_Op_d
{
   private:
      HCL_LinearOp_d * L;
   public:

   // Default constructor---disabled.
      HCL_LNLOp_d();
   // Copy constructor---disabled.
      HCL_LNLOp_d( const HCL_LNLOp_d & );
   /// Usual constructor; just needs a pointer to the linear operator.
      HCL_LNLOp_d( const HCL_LinearOp_d * LL );
   // Destructor.
      virtual ~HCL_LNLOp_d();

   /** Domain space access. Returns a reference to an
       {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access. Returns a reference to an
       {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;

   /** Evaluate creates an evaluation object which knows how
       to compute the image, first derivative, and second derivative, along
       with all relevant adjoints, at a point.   The purpose of
       the evaluation object is to accommodate the common situation in which
       the computation of the image and derivatives involve some common
       intermediate calculations.  */
      virtual HCL_EvaluateOp_d * Evaluate( const HCL_Vector_d & x ) const;

   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};

#endif
