/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _LS_
#define _LS_

#include "HCL_Adjoint_d.h"
#include "HCL_LinCombLinearOp_d.h"
#include "HCL_ScaleLinearOp_d.h"
#include "HCL_FunctionalProductDomain_d.h"
#include "HCL_OpProductDomain_d.h"
#include "HCL_CompLinearOp_d.h"

/** {\bf HCL\_EvalLSFcnlGN\_d} is the evaluation object for the functional
class {\bf HCL\_LeastSquareFcnlGN\_d}, which implements a least-squares
functional based on a nonlinear operator.  The suffix "GN" indicates that
the Gauss-Newton approximation to the Hessian is used.  See the
documentation for {\bf HCL\_LeastSquaresFcnlGN\_d} and
{\bf HCL\_EvaluateFunctional\_d} for more information. */

class HCL_EvalLSFcnlGN_d: public HCL_EvaluateFunctional_d {

private:

  int opflag;                 // = 0 for linear problems, = 1 for nonlinear
  HCL_Op_d * N;
  HCL_EvaluateOp_d * Neval;
  HCL_LinearOp_d * L;
  HCL_Vector_d * d;
  HCL_Vector_d * x;
  HCL_Vector_d * r;
  // regularization
  HCL_LinearOp_d * R;
  HCL_Vector_d * xprior;
  HCL_Vector_d * p;
  double eps;

public:
  // Default constructor.
  HCL_EvalLSFcnlGN_d();
  // Copy constructor - disabled.
  HCL_EvalLSFcnlGN_d( HCL_EvalLSFcnlGN_d & F);
  /** Usual constructor for linear operator input.  The constructor requires
      (a reference to) the input vector, pointers to the operator and data,
      and, optionally, a regularization weight, operator, and target model.
      Note that an independent copy of the input vector is made.  See
      the documentation for {\bf HCL\_LeastSquaresFcnlGN\_d} for more
      information about regularization. */
  HCL_EvalLSFcnlGN_d(const HCL_Vector_d & Input,
		     HCL_LinearOp_d * Operator, 
		     HCL_Vector_d * Data,
		     double PenaltyParameter,
		     HCL_LinearOp_d * RegularizingOperator,
		     HCL_Vector_d * ModelPrior);

  /** Usual constructor for nonlinear operator input.  The constructor requires
      (a reference to) the input vector, pointers to the operator and data,
      and, optionally, a regularization weight, operator, and target model.
      Note that an independent copy of the input vector is made.  See
      the documentation for {\bf HCL\_LeastSquaresFcnlGN\_d} for more
      information about regularization. */
  HCL_EvalLSFcnlGN_d(const HCL_Vector_d & Input,
		     HCL_Op_d * Operator, 
		     HCL_Vector_d * Data,
		     double PenaltyParameter,
		     HCL_LinearOp_d * RegularizingOperator,
		     HCL_Vector_d * ModelPrior);
  // Destructor.
  virtual ~HCL_EvalLSFcnlGN_d();

  /** Domain space access. Returns a reference to the vector space which
      forms the domain of the functional. */
  virtual HCL_VectorSpace_d & Domain() const;

  /// {\bf Value} returns the values of the functional at the given point.
  virtual double Value() const;

  /// {\bf Gradient} assigns to {\bf g} the gradient of the functional.
  virtual void Gradient( HCL_Vector_d & g ) const;

  /// {\bf Hessian} creates the Hessian operator.
  virtual HCL_LinearOp_d * Hessian() const;

  /// {\bf Write} prints a description of the object.
  virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_LeastSquaresFcnlGN} creates least squares objective functions 
  from operators and data.   This is a so called {\em bridge class},
  i.e. it exists merely to make objects of some class behave like
  objects of another class. In this case, the input objects are a
  linear or nonlinear operator \TEX{$A$} (an instance of {\bf HCL\_LinearOp\_d}
  or {\bf HCL\_Op\_d}) and a data vector d. The interface to these 
  objects constructed by the class is the least squares functional
  \TEX{
  \[
  x \mapsto \frac{1}{2} \|A[x]-d\|^2 \equiv J[x]
  \]}
  its gradient
  \TEX{
  \[
  x \mapsto DA[x]^T(A[x]-d) = \nabla J[x]
  \]}
  and the Gauss-Newton approximation to its Hessian operator
  \TEX{
  \[
  x \mapsto DA[x]^TDA[x]
  \]}
  expressed as an {\bf HCL\_Functional\_d}, an input class for
  optimization methods.
  
  So this class can be used to solve linear least squares problems straight
  from the implementation of the underlying operator, without constructing by
  hand the least squares function, as follows:

  If you have a linear operator {\bf L} (an instance of
  {\bf HCL\_LinearOp\_d}), and a data vector {\bf d} (an instance of
  {\bf HCL\_Vector}), construct the least squares functional as follows:

  \begin{verbatim}
  HCL_LeastSquaresFcnlGN_d F(L, d);
  \end{verbatim}

  whereas if you have a nonlinear operator {\bf N} instead (an instance of 
  {\bf HCL\_Op\_d}), use

  \begin{verbatim}
  HCL_LeastSquaresFcnlGN_d F(N, d);
  \end{verbatim}

  That is, the constructors look the same---the code does appropriate 
  things according to the types of the arguments submitted to the 
  constructor.

  The {\bf HCL\_Functional\_d} F so constructed is suitable for submission to 
  an optimization algorithm using second derivative information, such
  as Steihaug-Toint ({\bf HCL\_UMinTR\_d}), or to any of the algorithms
  using first derivative information, such as BFGS ({\bf HCL\_UMin\_lbfgs\_d})
  or nonlinear conjugate gradients ({\bf HCL\_UMinNLCG\_d}).

  Many least squares problems of practical interest are ill-posed, 
  and must be regularized to yield stable solutions. This class offers
  a simple interface for defining a quadratic penalty term, to create
  a penalized least squares function
  \TEX{
  \[
  x \mapsto \frac{1}{2} [\|A[x]-d\|^2 + \epsilon^2 \|R(x-x_0)\|^2] \equiv 
  J_{\epsilon}[x]
  \]}
  from three optional parameters to the constructor:

  \begin{itemize}
  \item the penalty parameter $\epsilon$ (a scalar)
  \item the regularizing operator $R$ (an instance of {\tt HCL\_LinearOp}
  \item the model prior $x_0$ (an instance of {\tt HCL\_Vector}
  \end{itemize}

  These parameters appear in the order given here, and default to 
  natural values. If none of them appear, no regularization takes place
  and the constructor builds the function previously described. Otherwise
  several special cases are implemented efficiently:

  \begin{itemize}
  \item
  \begin{verbatim}
  HCL_LeastSquaresFcnlGN_d F(L, d, eps);
  \end{verbatim}
  constructs a penalized problem with $R=I$, $x_0=0$, and prescribed
  penalty parameter ("ordinary Tihonov regularization");
  \item
  \begin{verbatim}
  HCL_LeastSquaresFcnlGN_d F(L, d, eps, R);
  \end{verbatim}
  constructs a penalized problem with $x_0=0$, and prescribed penalty 
  parameter and regularizing operator;
  \item
  \begin{verbatim}
  HCL_LeastSquaresFcnlGN_d F(L, d, eps, NULL, xprior);
  \end{verbatim}
  constructs a penalized problem with $R=I$ and prescribed penalty
  parameter and model prior.
  \end{itemize}
  */

class HCL_LeastSquaresFcnlGN_d: public HCL_Functional_d {

private:

  int opflag;                 // = 0 for linear problems, = 1 for nonlinear
  HCL_Op_d * N;
  HCL_LinearOp_d * L;
  HCL_Vector_d * d;

  // regularization
  HCL_LinearOp_d * R;
  HCL_Vector_d * xprior;
  double eps;

public:  

  // Default constructor---disabled.
  HCL_LeastSquaresFcnlGN_d();
  // Copy constructor---disabled.
  HCL_LeastSquaresFcnlGN_d( const HCL_LeastSquaresFcnlGN_d & );
  /** Usual constructor for linear operator input.  The constructor
      needs pointers to the (linear) operator and data.  Optionally,
      a regularization weight, regularization operator, and target model
      can be given.  See the class documentation for the behavior of the
      class using regularization. */
  HCL_LeastSquaresFcnlGN_d(HCL_LinearOp_d * Operator, 
		           HCL_Vector_d * Data,
			   double PenaltyParameter = 0.0,
			   HCL_LinearOp_d * RegularizingOperator = NULL,
			   HCL_Vector_d * ModelPrior = NULL);

  /** Usual constructor for nonlinear operator input.  The constructor
      needs pointers to the (nonlinear) operator and data.  Optionally,
      a regularization weight, regularization operator, and target model
      can be given.  See the class documentation for the behavior of the
      class using regularization. */
  HCL_LeastSquaresFcnlGN_d(HCL_Op_d * Operator, 
		           HCL_Vector_d * Data,
			   double PenaltyParameter = 0.0,
			   HCL_LinearOp_d * RegularizingOperator = NULL,
			   HCL_Vector_d * ModelPrior = NULL);
  // Destructor.
  virtual ~HCL_LeastSquaresFcnlGN_d();

  /** Domain space access.  Returns a reference to the vector space defining
      the domain of the functional. */
  virtual HCL_VectorSpace_d & Domain() const;

   /** {\bf Evaluate} creates an "evaluation" object, which knows how
       to compute the function value, gradient,and Hessian at the given x. */
  virtual HCL_EvaluateFunctional_d *
              Evaluate( const HCL_Vector_d & x ) const;

  /// {\bf Write} prints out a description of the functional.
  virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_EvalLSFcnlGNProductDomain\_d} is the evaluation object for
the functional class {\bf HCL\_LeastSquareFcnlGNProductDomain\_d}, which
implements a least-squares functional based on a nonlinear operator.  The
suffix "GN" indicates that the Gauss-Newton approximation to the Hessian is
used, while the term "ProductDomain" indicates that the operator (and hence
the least-squares functional) is defined on a product space.  See the
documentation for {\bf HCL\_LeastSquaresFcnlGNProductDomain\_d} and
{\bf HCL\_EvalFunctionalProductDomain\_d} for more information. */

class HCL_EvalLSFcnlGNProductDomain_d: public HCL_EvalFunctionalProductDomain_d
{
   private:

     HCL_OpProductDomain_d * N;
     HCL_EvalOpProductDomain_d * Neval;
     HCL_Vector_d * d;
     HCL_Vector_d * r;

   public:

   // Default constructor---disabled.
      HCL_EvalLSFcnlGNProductDomain_d();
   // Copy constructor---disabled.
      HCL_EvalLSFcnlGNProductDomain_d( 
          const HCL_EvalLSFcnlGNProductDomain_d & F);
  /** Usual constructor.  The constructor requires
      (a reference to) the input vector and pointers to the operator and data.
      Note that an independent copy of the input vector is made. */
      HCL_EvalLSFcnlGNProductDomain_d( const HCL_ProductVector_d & Input,
                                             HCL_OpProductDomain_d * Operator, 
                                             HCL_Vector_d * Data);
   // Destructor.
      virtual ~HCL_EvalLSFcnlGNProductDomain_d();

   /** Domain space access; returns a reference to the vector space defining
       the domain of the functional. */
      virtual HCL_VectorSpace_d & Domain() const;

   /// {\bf Value} computes the value of the functional.
      virtual double Value() const;

   /** {\bf PartialGradient} computes the component of the gradient
       corresponding to the ith component of the independent variable. */
      virtual void PartialGradient( int i,HCL_Vector_d & g ) const;

   /** Hessian computes the Hessian of the function. This method is
       re-defined in this class because, when the entire Hessian is
       desired, it can be provided more efficiently as the composition
       of the derivative and its adjoint, rather than as a block operator. */
      virtual HCL_LinearOp_d * Hessian() const;
   /** {\bf HessianBlock} creates the (i,j) block of the Hessian, as a
       linear operator. */
      virtual HCL_LinearOp_d * HessianBlock( int i,int j ) const;

   /** {\bf Write} invokes the Write method for the underlying objects
       (operator and data). */
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_LeastSquaresFcnl\_} creates least squares objective functions 
  from operators and data.  The term "ProductDomain" indicates the operator
  (and therefore the functional) is defined on a product space.
  This is a so called {\em bridge class}, i.e. it exists merely to make
  objects of some class behave like objects of another class. In this case,
  the input objects are a nonlinear operator \TEX{$A$} (an instance of
  {\bf HCL\_OpProductDomain\_d}) and a data vector d. The interface to these 
  objects constructed by the class is the least squares functional
  \TEX{
  \[
  x \mapsto \frac{1}{2} \|A[x]-d\|^2 \equiv J[x]
  \]}
  its gradient
  \TEX{
  \[
  x \mapsto DA[x]^T(A[x]-d) = \nabla J[x]
  \]}
  and the Gauss-Newton approximation to its Hessian operator
  \TEX{
  \[
  x \mapsto DA[x]^TDA[x]
  \]}
  expressed as an {\bf HCL\_Functional\_d}, an input class for optimization
  methods.
  
  So this class can be used to solve nonlinear least squares problems straight
  from the implementation of the underlying operator, without constructing by
  hand the least squares function, as follows:
  If you have a nonlinear operator {\bf N} instead (an instance of 
  {\bf HCL\_OpProductDomain\_d}), use
  \begin{verbatim}
  HCL_LeastSquaresFcnl_d F(N, d);
  \end{verbatim}

  The {\bf HCL\_Functional\_d} F so constructed is suitable for submission to 
  an optimization algorithm using second derivative information, such
  as Steihaug-Toint ({\bf HCL\_UMinTR\_d}), or to any of the algorithms
  using first derivative information, such as BFGS ({\bf HCL\_UMin\_lbfgs\_d})
  or nonlinear conjugate gradients ({\bf HCL\_UMinNLCG\_d}), as well as
  those algorithms specifically recognizing the product space structure,
  such as {\bf HCL\_UMinAlt\_d} and {\bf HCL\_UMinSub\_d}.

  */

class HCL_LeastSquaresFcnlGNProductDomain_d:
                      public HCL_FunctionalProductDomain_d
{
   private:

      HCL_OpProductDomain_d * N;
      HCL_Vector_d * d;

   public:  

   // Default constructor---disabled.
      HCL_LeastSquaresFcnlGNProductDomain_d();
   // Copy constructor---disabled.
      HCL_LeastSquaresFcnlGNProductDomain_d(
          const HCL_LeastSquaresFcnlGNProductDomain_d & );
   /// Usual constructor.  Pointers to the operator and data are required.
      HCL_LeastSquaresFcnlGNProductDomain_d(
               HCL_OpProductDomain_d * Operator,HCL_Vector_d * Data);
   // Destructor.
      virtual ~HCL_LeastSquaresFcnlGNProductDomain_d();

   /** Domain space access. Returns a reference to the vector space defining
       the domain of the functional. */
      virtual HCL_VectorSpace_d & Domain() const;

   /** {\bf Evaluate} creates an "evaluation" object, which knows how
       to compute the function value, gradient,and Hessian at the given x. */
      virtual HCL_EvaluateFunctional_d *
                  Evaluate( const HCL_Vector_d & x ) const;

   /// {\bf Write} invokes the Write methods of the operator and data.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_EvalLSFcnl\_d} is the evaluation object for the functional
class {\bf HCL\_LeastSquareFcnl\_d}, which implements a least-squares
functional based on a nonlinear operator.  See the documentation for
{\bf HCL\_LeastSquaresFcnl\_d} and {\bf HCL\_EvaluateFunctional\_d} for more
information. */

class HCL_EvalLSFcnl_d: public HCL_EvaluateFunctional_d {

private:

  HCL_Op_d * N;
  HCL_EvaluateOp_d * Neval;
  HCL_Vector_d * d;
  HCL_Vector_d * x;
  HCL_Vector_d * r;
  // regularization
  HCL_LinearOp_d * R;
  HCL_Vector_d * xprior;
  HCL_Vector_d * p;
  double eps;

public:

  // Default constructor.
  HCL_EvalLSFcnl_d();
  // Copy constructor - disabled.
  HCL_EvalLSFcnl_d( HCL_EvalLSFcnl_d & F);
  /** Usual constructor.  The constructor requires
      (a reference to) the input vector and pointers to the operator and data,
      and, optionally, a regularization weight, operator, and target model.
      Note that an independent copy of the input vector is made.  See
      the documentation for {\bf HCL\_LeastSquaresFcnl\_d} for more
      information about regularization. */
  HCL_EvalLSFcnl_d(const HCL_Vector_d & Input,
		     HCL_Op_d * Operator, 
		     HCL_Vector_d * Data,
		     double PenaltyParameter,
		     HCL_LinearOp_d * RegularizingOperator,
		     HCL_Vector_d * ModelPrior);
  // Destructor.
  virtual ~HCL_EvalLSFcnl_d();

  /** Domain space access. Returns a reference to the vector space defining
      the domain of the functional. */
  virtual HCL_VectorSpace_d & Domain() const;

   /// {\bf Value} computes the value of the functional.
  virtual double Value() const;

  /// {\bf Gradient} assigns to {\bf g} the gradient of the functional.
  virtual void Gradient( HCL_Vector_d & g ) const;

  /// {\bf Hessian} creates the Hessian operator.
  virtual HCL_LinearOp_d * Hessian() const;

   /// {\bf Write} invokes the Write methods of the operator and data.
  virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_LeastSquaresFcnl\_d} creates least squares objective functions 
  from operators and data.  This is a so called {\em bridge class},
  i.e. it exists merely to make objects of some class behave like
  objects of another class. In this case, the input objects are
  a nonlinear operator {\bf A} (an instance of {\bf HCL\_Op\_d})
  and a data vector {\bf d}. The interface to these  objects constructed by
  the class is the least squares functional
  \TEX{
  \[
  x \mapsto \frac{1}{2} \|A[x]-d\|^2 \equiv J[x]
  \]}
  its gradient
  \TEX{
  \[
  x \mapsto DA[x]^T(A[x]-d) = \nabla J[x]
  \]}
  and its Hessian operator
  \TEX{
  \[
  x \mapsto DA[x]^TDA[x]+L_3[A[x]-d]
  \]}
  expressed as an {\bf HCL\_Functional\_d}, an input class for
  optimization methods.
  
  This class can be used to solve nonlinear least squares problems straight
  from the implementation of the underlying operator, without constructing by
  hand the least-squares function, as follows:

  If you have a nonlinear operator {\bf N} (an instance of {\bf HCL\_Op\_d}),
  and a data vector d (an instance of {\bf HCL\_Vector\_d}), construct the
  least-squares functional as follows:

  \begin{verbatim}
  HCL_LeastSquaresFcnl_d F(N, d);
  \end{verbatim}

  The {\bf HCL\_Functional\_d} {\bf F} so constructed is suitable for
  submission to  an optimization algorithm using second derivative
  information, such as Steihaug-Toint (HCL\_UMinTR), or to any of the
  algorithms using first derivative information, such as BFGS
  ({\bf HCL\_UMin\_lbfgs\_d}) or nonlinear conjugate gradients
  ({\bf HCL\_UMinNLCG\_d}).

  Many least-squares problems of practical interest are ill-posed, 
  and must be regularized to yield stable solutions. This class offers
  a simple interface for defining a quadratic penalty term, to create
  a penalized least squares function
  \TEX{
  \[
  x \mapsto \frac{1}{2} [\|A[x]-d\|^2 + \epsilon^2 \|R(x-x_0)\|^2] \equiv 
  J_{\epsilon}[x]
  \]}
  from three optional parameters to the constructor:

  \begin{itemize}
  \item the penalty parameter $\epsilon$ (a scalar)
  \item the regularizing operator $R$ (an instance of {\tt HCL\_LinearOp}
  \item the model prior $x_0$ (an instance of {\tt HCL\_Vector}
  \end{itemize}

  These parameters appear in the order given here, and default to 
  natural values. If none of them appear, no regularization takes place
  and the constructor builds the function previously described. Otherwise
  several special cases are implemented efficiently:

  \begin{itemize}
  \item
  \begin{verbatim}
  HCL_LeastSquaresFcnl_d F(L, d, eps);
  \end{verbatim}
  constructs a penalized problem with $R=I$, $x_0=0$, and prescribed
  penalty parameter ("ordinary Tihonov regularization");
  \item
  \begin{verbatim}
  HCL_LeastSquaresFcnl_d F(L, d, eps, R);
  \end{verbatim}
  constructs a penalized problem with $x_0=0$, and prescribed penalty 
  parameter and regularizing operator;
  \item
  \begin{verbatim}
  HCL_LeastSquaresFcnl_d F(L, d, eps, NULL, xprior);
  \end{verbatim}
  constructs a penalized problem with $R=I$ and prescribed penalty
  parameter and model prior.
  \end{itemize}
  */
class HCL_LeastSquaresFcnl_d: public HCL_Functional_d {

private:

  HCL_Op_d * N;
  HCL_Vector_d * d;

  // regularization
  HCL_LinearOp_d * R;
  HCL_Vector_d * xprior;
  double eps;

public:  

  // Default constructor---disabled.
  HCL_LeastSquaresFcnl_d();
  // Copy constructor---disabled.
  HCL_LeastSquaresFcnl_d( const HCL_LeastSquaresFcnl_d & );
  /** Usual constructor. The constructor needs pointers to the operator and
      data.  Optionally, a regularization weight, regularization operator,
      and target model can be given.  See the class documentation for the
      behavior of the class using regularization. */
  HCL_LeastSquaresFcnl_d(HCL_Op_d * Operator, 
		           HCL_Vector_d * Data,
			   double PenaltyParameter = 0.0,
			   HCL_LinearOp_d * RegularizingOperator = NULL,
			   HCL_Vector_d * ModelPrior = NULL);
  // Destructor.
  virtual ~HCL_LeastSquaresFcnl_d();

  /** Domain space access. Returns a reference to the vector space defining
      the domain of the functional. */
  virtual HCL_VectorSpace_d & Domain() const;

   /** {\bf Evaluate} creates an "evaluation" object, which knows how
       to compute the function value, gradient,and Hessian at the given x. */
  virtual HCL_EvaluateFunctional_d *
              Evaluate( const HCL_Vector_d & x ) const;

   /// {\bf Write} invokes the Write methods of the operator and data.
  virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_EvalLSFcnlProductDomain\_d} is the evaluation object for the
functional class {\bf HCL\_LeastSquareFcnlProductDomain\_d}, which
implements a least-squares functional based on a nonlinear operator.  The
term "ProductDomain" indicates that the operator (and hence the functional)
is defined on a product space.  See the documentation for
{\bf HCL\_LeastSquaresFcnlProductDomain\_d} and
{\bf HCL\_EvalFunctionalProductDomain\_d} for more information. */

class HCL_EvalLSFcnlProductDomain_d: public HCL_EvalFunctionalProductDomain_d
{
   private:

     HCL_OpProductDomain_d * N;
     HCL_EvalOpProductDomain_d * Neval;
     HCL_Vector_d * d;
     HCL_Vector_d * r;

   public:

   // Default constructor---disabled.
      HCL_EvalLSFcnlProductDomain_d();
   // Copy constructor---disabled.
      HCL_EvalLSFcnlProductDomain_d( 
          const HCL_EvalLSFcnlProductDomain_d & F);
  /** Usual constructor.  The constructor requires
      (a reference to) the input vector and pointers to the operator and data.
      Note that an independent copy of the input vector is made. */
      HCL_EvalLSFcnlProductDomain_d( const HCL_ProductVector_d & Input,
                                             HCL_OpProductDomain_d * Operator, 
                                             HCL_Vector_d * Data);
   // Destructor.
      virtual ~HCL_EvalLSFcnlProductDomain_d();

  /** Domain space access. Returns a reference to the vector space defining
      the domain of the functional. */
      virtual HCL_VectorSpace_d & Domain() const;

   /// {\bf Value} computes the value of the functional.
      virtual double Value() const;

   /** {\bf PartialGradient} computes the component of the gradient
       corresponding to the ith component of the independent variable. */
      virtual void PartialGradient( int i,HCL_Vector_d & g ) const;

   /** {\bf HessianBlock} creates the (i,j) block of the Hessian, as a
       linear operator. */
      virtual HCL_LinearOp_d * HessianBlock( int i,int j ) const;

   /** {\bf Write} invokes the Write method for the underlying objects
       (operator and data). */
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_LeastSquaresFcnlProductDomain\_d} creates least-squares
  objective functions from operators and data.  The term "ProductDomain"
  indicates that the operator (and hence the functional) is defined on
  a product space.  This is a so called {\em bridge class},
  i.e. it exists merely to make objects of some class behave like
  objects of another class. In this case, the input objects are
  a nonlinear operator {\bf A} (an instance of {\bf HCL\_OpProductDomain\_d})
  and a data vector {\bf d}. The interface to these  objects constructed by
  the class is the least-squares functional
  \TEX{
  \[
  x \mapsto \frac{1}{2} \|A[x]-d\|^2 \equiv J[x]
  \]}
  its gradient
  \TEX{
  \[
  x \mapsto DA[x]^T(A[x]-d) = \nabla J[x]
  \]}
  and its Hessian operator
  \TEX{
  \[
  x \mapsto DA[x]^TDA[x]+L_3[A[x]-d]
  \]}
  expressed as an {\bf HCL\_FunctionalProductDomain\_d}, an input class for
  optimization methods.
  
  This class can be used to solve nonlinear least squares problems straight
  from the implementation of the underlying operator, without constructing by
  hand the least-squares function, as follows:

  If you have a nonlinear operator {\bf N} (an instance of
  {\bf HCL\_OpProductDomain\_d}), and a data vector d (an instance of
  {\bf HCL\_Vector\_d}), construct the least-squares functional as follows:

  \begin{verbatim}
  HCL_LeastSquaresFcnlProductDomain_d F(N, d);
  \end{verbatim}

  The {\bf HCL\_Functional\_d} {\bf F} so constructed is suitable for
  submission to  an optimization algorithm using second derivative
  information, such as Steihaug-Toint (HCL\_UMinTR), or to any of the
  algorithms using first derivative information, such as BFGS
  ({\bf HCL\_UMin\_lbfgs\_d}) or nonlinear conjugate gradients
  ({\bf HCL\_UMinNLCG\_d}), as well as those algorithms specifically
  recognizing the product space structure, such as {\bf HCL\_UMinAlt\_d}
  and {\bf HCL\_UMinSub\_d}.
  */
class HCL_LeastSquaresFcnlProductDomain_d:
                      public HCL_FunctionalProductDomain_d
{
   private:

      HCL_OpProductDomain_d * N;
      HCL_Vector_d * d;

   public:  

   // Default constructor---disabled.
      HCL_LeastSquaresFcnlProductDomain_d();
   // Copy constructor---disabled.
      HCL_LeastSquaresFcnlProductDomain_d(
          const HCL_LeastSquaresFcnlProductDomain_d & );
   /// Usual constructor.  Pointers to the operator and data are required.
      HCL_LeastSquaresFcnlProductDomain_d(
               HCL_OpProductDomain_d * Operator,HCL_Vector_d * Data);
   // Destructor.
      virtual ~HCL_LeastSquaresFcnlProductDomain_d();

  /** Domain space access. Returns a reference to the vector space defining
      the domain of the functional. */
      virtual HCL_VectorSpace_d & Domain() const;

   /** {\bf Evaluate} creates an "evaluation" object, which knows how
       to compute the function value, gradient,and Hessian at the given x. */
      virtual HCL_EvaluateFunctional_d *
                  Evaluate( const HCL_Vector_d & x ) const;

   /// {\bf Write} invokes the Write methods of the operator and data.
      virtual ostream & Write( ostream & str ) const;
};

#endif
