/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_LINESEARCH_H_
#define HCL_LINESEARCH_H_

#include <iostream.h>
#include <stdlib.h>
#include "Table.h"
#include "HCL_ScaleLinearOp_d.h"
#include "HCL_LinearOp_d.h"
#include "HCL_LinearSolver_d.h"
#include "HCL_Functional_d.h"
#include "HCL_Arch_d.h"
#include "ProtectedDivision_d.h"

/** Abstract base class for line search routines.  The purpose of
a line search algorithm is to locate an approximate minimizer
of a function \TEX{$\phi:[0,\infty) \rightarrow R$}.  Such problems
arise in certain algorithms for unconstrained and constrained
optimization, where \TEX{$\phi$} is a 1D "slice" of the objective or
merit function.  The is, the original problem is to minimizer
some function \TEX{$f:X\rightarrow R$}, where \TEX{$X$} is a Hilbert space,
and \TEX{$\phi$} is defined by \TEX{$\phi(a) = f(x+ay)$}, where \TEX{$x$}
and \TEX{$y$} are vectors in \TEX{$X$}.

The main method in a {\bf HCL\_LineSearch\_d} class is {\bf Search},
which is invoked as follows:
\begin{verbatim}
   MyFcnl f;  // MyFcnl must be derived from HCL_Functional_d
   MyVector x( "x0" ); // Assume MyVector has a constructor which
                       //reads from a file
   MyVector y( "direction" );
   MyVector * xnext = new MyVector( x );  // Assume MyVector has a copy
                                          // constructor.

   HCL_LineSearch_DS_d lsearch( "lsearch.dat" );  // Dennis-Schnabel
                                                  // line search
   HCL_EvaluateFunctional_d * eval = f.EvaluateGrad( x );
   HCL_EvaluateFunctional_d * evalnext 
                                  lsearch.Search( f,x,*xnext,y,eval );
\end{verbatim}
Note that it is necessary to evaluate the functional at the
initial point \TEX{$x$} and pass this evaluation object to the {\bf Search}
method.  The method attempts to locate a suitable point and
return it in the vector pointed to by the pointer xnext.
Also, the evaluation object of \TEX{$f$} is the return value of the method.

Standard line search algorithms are usually based on one or
two goals.  Essentially all algorithms attempt to reduce
the value of the function \TEX{$\phi$} from \TEX{$\phi(0)$}, usually by
satisfying a condition
\TEX{$$
   \phi(a) \le \phi(0) + tol a\phi'(0)
$$}
(the assumption is made that \TEX{$\phi'(0)$} is negative).  The number
\TEX{$tol$} in the above inequality is a standard input parameter
to a line search algorithm; in HCL this parameter is called
{\bf FcnDecreaseTol}.  A second goal which is often sought is
the reduction of the slope from \TEX{$\phi'(0)$}:
\TEX{$$
   |\phi'(a)| \le tol |\phi'(0)|.
$$}
The tolerance in this condition is called {\bf SlopeDecreaseTol}
in HCL.

Other important parameters appearing in line search algorithms
are {\bf MaxSample} (the maximum number of times that the
function can be sampled before the algorithm gives up), and
{\bf MinStep} and {\bf MaxStep} (the smallest and largest
allowable steps).  For a complete list of algorithmic parameters,
see the {\bf Input parameters} documentation for the base class,
and for each concrete derived class (note that a specific algorithm
may use a parameter not mentioned in the base class).

Algorithmic parameters can be accessed or modified through the
{\bf Parameters} method of the class, which returns a reference to
the parameter table (see the {\bf Table} class for more details
about parameter tables):
\begin{verbatim}
   lsearch.Parameters().GetValue( "MinStep",m ); // Gets the value of
                                                 // the parameter MinStep
                                                 // and puts it in m.
   lsearch.Parameters().PutValue( "MinStep",2*m ); // Sets the value of
                                                   // the parameter MinStep
                                                   // to 2*m.
\end{verbatim}
In addition, derived line search classes should have constructors
which take a file name and read parameters from that file
(using the {\bf Table} file format).  All input parameters have
default values which are given in the (derived) class documentation.

All line searchs are required to set two output parameters in the
paramter table, {\bf TermCode} and {\bf MaxTkn}.  The possible
values of {\bf TermCode} are given by the enum variable {\bf ErrorCodes}.
The documentation of {\bf ErrorCodes} should be consulted for a complete
list of codes, but the following convention is followed: 0 means
a succesful completion, a positive values means that the algorithm
succeeded in reducing the function value, although the termination
criteria were not all satisfied, and a negative value indicates
that the search was unable to reduce the function value.
{\bf MaxTkn} is a boolean integer variable; if it is set, this
indicates that the step taken was of length {\bf MaxStep}.  */

class HCL_LineSearch_d: public HCL_Base
{
   private:


// The following are for documentation purposes only
#ifdef NOT_DEFINED

   // required parameters found in ParamTable for all derived line searches

   // Input parameters
   /**@name Input parameters */
   //@{
   /** Tolerance for decreasing function value (Armijo-Goldstein
       condition). */
      double FcnDecreaseTol; 
   /** Slope at approximate minimizer must be reduced by a factor
       of SlopeDecreaseTol (exact interpretation varies). */
      double SlopeDecreaseTol;
   /// Minimum step length.
      double MinStep;
   /// Maximum step length.
      double MaxStep;
   /// Maximum number of function evaluations to perform
      int MaxSample;
   /// Tolerance for divide-by-zero
      double ZeroDivideTol;
   /** DispFlag controls the amount of output sent to the screen during
       execution of the Search method.  The usual choices are
               0 --- no output
               1 --- a summary at the end of execution
               2 --- a summary at each iteration
      In addition, values of DispFlag greater than 2 may send increasing
      amount of detail to the screen (this is intended primarily for
      development use, i.e. for debugging). */
      int DispFlag;
   /** DumpFlag has exactly the same purpose as DispFlag, except the
       information is sent to a file, whose name is defined by DumpFile. */
      int DumpFlag;
   /** DumpFile is the name of the file to which results are written (if
       DumpFlag is positive). */
      char DumpFile[81];
   /// DispPrecision gives the number of digits in numbers sent to the screen.
      int DispPrecision;
   /// DumpPrecision gives the number of digits in numbers sent to the file.
      int DumpPrecision;
   //@}


   /**@name Required output parameters */
   //@{
   ///exit status  (<0 = no reduction, 0 = success, >0 = reduction, but failed)
      int TermCode;
   ///maximum step size taken?  (1 = taken, 0 = not taken)
      int MaxTkn;   
   //@}
#endif

   public:
	  
	  /**@name Term codes */
	  //@{
	  enum {
		 NoErr = 0,
		 Succeed = 0,
                 NoReduction = -2,
		 InitialSlopePositive = -1,
		 TooManyFunctionEvals = 2,
		 Fail = 3
	  } ErrorCodes;
	  //@}

   /// Access to the parameter table

      virtual Table & Parameters() const = 0;

   /** {\bf SetScaling} allows the specification of a different inner
       product (and hence norm).  Instead of \TEX{$(x,y)$}, the scaled inner
       product \TEX{$(x,Sy)$} will be used, where \TEX{$S$} is a symmetric
       positive definite operator.  Through this method, a line
       search can be made consistent with a minimization algorithm which
       scales the norm. */
      virtual void SetScaling( const HCL_LinearOp_d * S ) = 0;
   /** {\bf UnSetScaling} causes the line search to use the default inner
       product. */
      virtual void UnSetScaling() = 0;

   /** LineSearch algorithm.
       @param f function
       @param xcur starting point
       @param dir starting direction
       @param xinit_eval evaluation object for starting point
    */
      virtual HCL_EvaluateFunctional_d * Search( 
                    HCL_Functional_d & f,
                    const HCL_Vector_d & xcur,HCL_Vector_d & xnext,
                    HCL_Vector_d & dir,
                    HCL_EvaluateFunctional_d * xinit_eval) = 0;

};


/** An implementation of a backtracking
line search algorithm using cubic interpolation
(See Dennis and Schnabel, "Numerical Methods for Unconstrained
Optimization and Nonlinear Equations", Prentice-Hall (1983)).

The description of this class differs from the description of the
base class {\bf HCL\_LineSearch\_d} in only one regard: the second
criteria for defining an acceptable step is
\TEX{$$
   \phi'(a) \ge tol \phi'(0).
$$}
This prevents short steps by requiring that the slope be increased
from its initial (negative) value, although it does not require that
the step reach an approximate minimizer.  This criterion is sufficient
to allow a positive definite update in a BFGS algorithm.
*/
class HCL_LineSearch_DS_d: public HCL_LineSearch_d
{
   private:

      Table * ParamTable;

   // parameters found in ParamTable

      // Input Parameters
   /**@name Input parameters */
   //@{
   /** (1e-4) Tolerance for decreasing function value (Armijo-Goldstein
       condition). */
      double FcnDecreaseTol; 
   /** (9e-1) Slope at approximate minimizer must be positive or be reduced
       by a factor of SlopeDecreaseTol. */
      double SlopeDecreaseTol;
   /// (1e-20) minimum step length 
      double MinStep;
   /// (1e20) maximum step length 
      double MaxStep;
   /// (8) Maximum number of function evaluations to perform
      int MaxSample;
   /// (1e-40) Tolerance for divide-by-zero
      double ZeroDivideTol;
   /** (0) DispFlag controls the amount of output sent to the screen during
       execution of the Search method.  The usual choices are
               0 --- no output
               1 --- a summary at the end of execution
               2 --- a summary at each iteration
      In addition, values of DispFlag greater than 2 may send increasing
      amount of detail to the screen (this is intended primarily for
      development use, i.e. for debugging). */
      int DispFlag;
   /** (0) DumpFlag has exactly the same purpose as DispFlag, except the
       information is sent to a file, whose name is defined by DumpFile. */
      int DumpFlag;
   /** (HCL\_LineSearch\_DS.DumpFile) DumpFile is the name of the file to
       which results are written (if DumpFlag is positive). */
      char DumpFile[81];
   /** (6) DispPrecision gives the number of digits in numbers sent to the
       screen. */
      int DispPrecision;
   /** (DispPrecision) DumpPrecision gives the number of digits in numbers
       sent to the file. */
      int DumpPrecision;
   /// (2) factor by which to increase step length if initial step is too small
      double BracketIncrease;
   //@}


   /**@name Output parameters */
   //@{
   ///exit status  (<0 = no reduction, 0 = success, >0 = reduction, but failed)
      int TermCode;
   ///maximum step size taken?  (1 = taken, 0 = not taken)
      int MaxTkn;   
   //@}

   // local variables
      double f_x,f_xnext;           // f(xcur)  and  f(xnext)
      int Itn;                      // Backtracking iteration counter
      int OldDispPrec;

      ofstream * dump;

   // Some temporary vectors
      HCL_LinearOp_d * S;           // preconditioner Op
      HCL_Vector_d * tmpvec;    



   /// Read parameters from parameter table and allocate temporary vectors.

      virtual void Initialize( HCL_VectorSpace_d & dom);

   /// Write output parameters out to parameter table and deallocate vectors.

      virtual void Clean();

   public:

   // Copy constructor --- disabled
      HCL_LineSearch_DS_d( const HCL_LineSearch_d & x );
   /** Usual constructor.
       @param fname Parameter file name or NULL
   */
      HCL_LineSearch_DS_d( char *fname = NULL );

   // destructor
      virtual ~HCL_LineSearch_DS_d();

   /// Access to the parameter table
      virtual Table & Parameters() const { return *ParamTable; }

   /** {\bf SetScaling} allows the specification of a different inner
       product (and hence norm).  Instead of \TEX{$(x,y)$}, the scaled inner
       product \TEX{$(x,Sy)$} will be used, where \TEX{$S$} is a symmetric
       positive definite operator.  Through this method, a line
       search can be made consistent with a minimization algorithm which
       scales the norm. */
      virtual void SetScaling( const HCL_LinearOp_d * S );
   /** {\bf UnSetScaling} causes the line search to use the default inner
       product. */
      virtual void UnSetScaling();
   /** LineSearch algorithm.  Returns an evaluation object for the final
       point or NULL, depending on whether the line search was
       successful or not.  The Parameter TermCode contains the exit status.

       @param fptr function
       @param xcur starting point, on successful completion, this will hold
              the computed minimum
       @param dir starting direction
       @param xinit_eval evaluation object for starting point
    */

      virtual HCL_EvaluateFunctional_d * Search( 
                    HCL_Functional_d & fptr,
                    const HCL_Vector_d & xcur,HCL_Vector_d & xnext,
                    HCL_Vector_d & dir,
                    HCL_EvaluateFunctional_d * xinit_eval);

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

/** An implementation of the More' and Thuente line search algorithm
(See More' and Thuente, "Line Search Algorithms with Guaranteed
Sufficient Decrease", ACM TOMS, Vol. 20, No. 3, 286--307 (1994)).
The description of this class follows closely that of the base
class {\bf HCL\_LineSearch\_d}, which should be consulted for details.
*/
class HCL_LineSearch_MT_d: public HCL_LineSearch_d
{
   private:

      Table * ParamTable;

   // parameters found in ParamTable

   /**@name Input parameters */
   //@{
   /** (1e-4) Tolerance for decreasing function value (Armijo-Goldstein
       condition). */
      double FcnDecreaseTol; 
   /** (9e-1) Slope at minimum must be smaller than SlopeDecreaseTol times the
        initial slope. */
      double SlopeDecreaseTol; 
   /// (1e-20) Minimum step length 
      double MinStep;
   /// (1e20) Maximum step length 
      double MaxStep;
   /// (8) Maximum number of function evaluations to perform
      int MaxSample;
   /// (1e-40) Tolerance for divide-by-zero
      double ZeroDivideTol;
   /** (0) DispFlag controls the amount of output sent to the screen during
       execution of the Search method.  The usual choices are
               0 --- no output
               1 --- a summary at the end of execution
               2 --- a summary at each iteration
      In addition, values of DispFlag greater than 2 may send increasing
      amount of detail to the screen (this is intended primarily for
      development use, i.e. for debugging). */
      int DispFlag;
   /** (0) DumpFlag has exactly the same purpose as DispFlag, except the
       information is sent to a file, whose name is defined by DumpFile. */
      int DumpFlag;
   /** (HCL\_LineSearch\_MT.DumpFile) DumpFile is the name of the file to
       which results are written (if DumpFlag is positive). */
      char DumpFile[81];
   /** (6) DispPrecision gives the number of digits in numbers sent to
       the screen. */
      int DispPrecision;
   /** (DispPrecision) DumpPrecision gives the number of digits in numbers
       sent to the file. */
      int DumpPrecision;
   /** (sqrt(macheps)) Line search halts if the length of the interval of
       of uncertainty falls below IntervalTol. */
      double IntervalTol;
   /// (4) factor to change mu during backtracking
      double BracketIncrease;
   //@}


   /**@name Output parameters */
   //@{
   ///exit status (<0 = no reduction, 0 = success, >0 = reduction, but failed)
      int TermCode;
   /// Maximum step size taken?  (1 = taken, 0 = not taken)
      int MaxTkn;   
   //@}

   // local variables
      int OldDispPrec;
      double f_x,f_xnext;           // f(xcur)  and  f(xnext)
      int Itn;                      // Backtracking iteration counter

      HCL_LinearOp_d * S;           // preconditioner Op
      HCL_Vector_d * tmpvec;    
      ofstream * dump;

   /// Read parameters from parameter table and allocate temporary vectors.
      virtual void Initialize( HCL_VectorSpace_d & dom);

   /// Write output parameters out to parameter table and deallocate vectors.
      virtual void Clean();

   /// Display results
      void DisplayResults( int );

   /// Update the interval of uncertainty and compute the next step.
      int cstep( double & mux,double & fx,double & dgx,double & muy,
                 double & fy,double & dgy,double & mu,double & fv,
                 double & dg,int & bracket,double & mumin,double & mumax );

   public:

   // Default constructor --- disabled
      HCL_LineSearch_MT_d( const HCL_LineSearch_d & x );
   
   /** Usual constructor.
       @param fname Parameter file name or NULL
   */
      HCL_LineSearch_MT_d( char *fname = NULL );

   // destructor
      virtual ~HCL_LineSearch_MT_d();

   /// Access to the parameter table
      virtual Table & Parameters() const { return *ParamTable; }

   /** {\bf SetScaling} allows the specification of a different inner
       product (and hence norm).  Instead of \TEX{$(x,y)$}, the scaled inner
       product \TEX{$(x,Sy)$} will be used, where \TEX{$S$} is a symmetric
       positive definite operator.  Through this method, a line
       search can be made consistent with a minimization algorithm which
       scales the norm. */
      virtual void SetScaling( const HCL_LinearOp_d * S );
   /** {\bf UnSetScaling} causes the line search to use the default inner
       product. */
      virtual void UnSetScaling();

   /** LineSearch algorithm.  Returns an evaluation object for the final
       point or NULL, depending on whether the line search was
       successful or not.  The Parameter TermCode contains the exit status.

       @param fptr function
       @param xcur starting point, on successful completion, this will hold
              the computed minimum
       @param dir starting direction
       @param xinit_eval evaluation object for starting point
    */
      virtual HCL_EvaluateFunctional_d * Search( 
                    HCL_Functional_d & fptr,
                    const HCL_Vector_d & xcur,HCL_Vector_d & xnext,
                    HCL_Vector_d & dir,
                    HCL_EvaluateFunctional_d * xinit_eval);
    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};


/** An implementation of a backtracking
line search algorithm using cubic  and quadratic interpolation
(See Fletcher, "Practical Methods of Optimization" (2nd edition),
Wiley, (1987)).  The description of this class follows closely that of the
base class {\bf HCL\_LineSearch\_d}, which should be consulted for details. */
class HCL_LineSearch_Fl_d: public HCL_LineSearch_d
{
   private:

      Table * ParamTable;

   // parameters found in ParamTable

   /**@name Input parameters */
   //@{
   /** (1e-4) Tolerance for decreasing function value (Armijo-Goldstein
       condition). */
      double FcnDecreaseTol; 
   /** (1e-1) Slope at minimum must be smaller than SlopeDecreaseTol times the
        initial slope. */
      double SlopeDecreaseTol; 
   /// (1e-20) Minimum step length.
      double MinStep;
   /// (1e20) Maximum step length.
      double MaxStep;
   /// (8) Maximum number of function evaluations to perform
      int MaxSample;
   /// (1e-40) Tolerance for divide-by-zero
      double ZeroDivideTol;
   /** (0) DispFlag controls the amount of output sent to the screen during
       execution of the Search method.  The usual choices are
               0 --- no output
               1 --- a summary at the end of execution
               2 --- a summary at each iteration
      In addition, values of DispFlag greater than 2 may send increasing
      amount of detail to the screen (this is intended primarily for
      development use, i.e. for debugging). */
      int DispFlag;
   /** (0) DumpFlag has exactly the same purpose as DispFlag, except the
       information is sent to a file, whose name is defined by DumpFile. */
      int DumpFlag;
   /** (HCL\_LineSearch\_Fl.DumpFile) DumpFile is the name of the file to
       which results are written (if DumpFlag is positive). */
      char DumpFile[81];
   /** (6) DispPrecision gives the number of digits in numbers sent to the
       screen. */
      int DispPrecision;
   /** (DispPrecision) DumpPrecision gives the number of digits in numbers
       sent to the file. */
      int DumpPrecision;
   /** (0) Minimum value of function.  The algorithm will halt if the function
       value falls below this value. */
      double MinValue; 
   /** (sqrt(macheps)) Line search halts if the length of the interval of
       of uncertainty falls below IntervalTol. */
      double IntervalTol;
   /// (9) Factor by which to increase the width of the bracket.
      double BracketIncrease; 
   /** (1e-1)tau2 and tau3 determine the size of the interval in which the
       minimum value of the interpolating polynomial is allowed to lie during
       the sectioning phase.  For example, if the interval for sectioning is
       [0,1], then the reduced interval will be [0, alpha] or [alpha, 1]
       with alpha in [tau2, 1-tau3]. */
      double tau2; 
   /// (5e-1)
      double tau3; 
   /** (0) Step length flag; use last mu if set; otherwise try step
       length one first. */
      int LineStepFlag;
   //@}


   /**@name Output parameters */
   //@{
   /** exit status (negative means no reduction, 0 means success, 
                    positive means reduction, but failed) */
      int TermCode;
   ///maximum step size taken?  (1 = taken, 0 = not taken)
      int MaxTkn;   
   //@}

   // local variables

      int OldDispPrec;      
      ofstream * dump;
      double LastMu;

   // Some temporary vectors
      HCL_LinearOp_d * S;           // preconditioner Op
      HCL_Vector_d * g;             // starting gradient
      HCL_Vector_d * tmpvec;    



   /// Read parameters from parameter table and allocate temporary vectors.

      virtual void Initialize( HCL_VectorSpace_d & dom);

   /// Write output parameters out to parameter table and deallocate vectors.

      virtual void Clean();


      double CubicInterpMin(double ap, double apf_x, double apslope, 
	     double bp, double bpf_x, double bpslope, double cp, double dp);

      double QuadInterpMin(double ap, double apf_x, double apslope, 
	     double bp, double bpf_x, double cp, double dp);

   public:

   // Copy constructor --- disabled
      HCL_LineSearch_Fl_d( const HCL_LineSearch_d & x );
   
   /** Usual constructor.
       @param fname Parameter file name or NULL
   */
      HCL_LineSearch_Fl_d( char *fname = NULL );

   // destructor
      virtual ~HCL_LineSearch_Fl_d();

   /// Access to the parameter table
      virtual Table & Parameters() const { return *ParamTable; }

   /** {\bf SetScaling} allows the specification of a different inner
       product (and hence norm).  Instead of \TEX{$(x,y)$}, the scaled inner
       product \TEX{$(x,Sy)$} will be used, where \TEX{$S$} is a symmetric
       positive definite operator.  Through this method, a line
       search can be made consistent with a minimization algorithm which
       scales the norm. */
      virtual void SetScaling( const HCL_LinearOp_d * S );
   /** {\bf UnSetScaling} causes the line search to use the default inner
       product. */
      virtual void UnSetScaling();

   /** LineSearch algorithm.  Returns an evaluation object for the final
       point or NULL, depending on whether the line search was
       successful or not.  The Parameter TermCode contains the exit status.

       @param fptr function
       @param xcur starting point, on successful completion, this will hold
              the computed minimum
       @param dir starting direction
       @param xinit_eval evaluation object for starting point (or NULL)
    */
      virtual HCL_EvaluateFunctional_d * Search( 
                    HCL_Functional_d & fptr,
                    const HCL_Vector_d & xcur,HCL_Vector_d & xnext,
                    HCL_Vector_d & dir,
                    HCL_EvaluateFunctional_d * xinit_eval);

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
