/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef _HCL_OP_D_H
#define _HCL_OP_D_H

#include "HCL_BiLinearOp_d.h"
#include "HCL_Arch_d.h"
#include "maxmin_d.h"

// "Look-ahead" declarations.
class HCL_OpDefaultDeriv_d;
class HCL_OpDefaultSecondDeriv_d;

/** This class represents an "evaluation" object that will be created
by an {\bf HCL\_Op\_d} object to represent the image and derivatives
of the operator at a point.  Thus, if \TEX{$F$} is a nonlinear operator and
\TEX{$x$} is a vector in the domain, then the evaluation object represents
the triple
\TEX{
\[
(F(x),DF(x),D^2F(x)).
\]
}
The purpose of having an evaluation object is to deal with the common
situation in which the operator image and derivatives share some intermediate
calculations.  A typical implementation of such an operator involves defining
an evaluation object whose constructor performs these intermediate
calculations.  The {\bf Image} method then completes the computation
of the operator image, while the {\bf Deriv} and {\bf SecondDeriv}
methods complete the calculation of the derivatives.

The operator image and derivatives should then be accessed through
the {\bf ImageRef}, {\bf DerivRef}, and {\bf SecondDerivRef} methods.
These are methods implemented in the base class (i.e. they do not need to
be redefined in a derived class).  The {\bf ImageRef} method invokes
{\bf Image} to get the image and returns a reference to it.  On subsequent
calls, {\bf ImageRef} merely references the stored value.  The other
"reference" methods work in a similar fashion.

Here is a summary of the primary methods of this class:
\begin{itemize}
\item {\bf Domain} Returns a reference to the domain of the underlying
operator.
\item {\bf Range} Returns a reference to the range of the underlying
operator.
\item {\bf Image} Computes the image of the operator at the point
at which this object was created.  Must be implemented by the creator
of a concrete class.
\item {\bf ImageRef} Calls {\bf Image}, if necessary, to compute
the operator image, saves it, and returns it.  Implemented in the
base class.
\item {\bf ImageAdd} Adds the image of the operator to the output
(as opposed to {\bf Image}, which overwrites the output with the image).
\item {\bf Deriv} Computes the derivative of the operator, at the point
at which this object was created, as an {\bf HCL\_LinearOp\_d}.
Must be implemented by the creator of a concrete class.
\item {\bf DerivRef} Calls {\bf Deriv}, if necessary, to compute
the derivatives, saves it, and returns a reference to it.
Implemented in the base class.
\item {\bf SecondDeriv} Computes the second derivative of the operator,
at the point at which this object was created, as a
{\bf HCL\_BiLinearOp\_d}.  Must be implemented by the creator of a
concrete class.
\item {\bf SecondDerivRef} Calls {\bf SecondDeriv}, if necessary, to
compute the second derivative, saves it, and returns a reference to it.
Implemented in the base class.
\end{itemize}

In addition to the above methods, {\bf HCL\_EvaluateOp\_d} has a number of
protected virtual functions that allow the user to avoid writing
distinct derivative classes if desired.  If these methods, which
include {\bf DerivImage}, {\bf SecondDerivImage}, and others, are over-ridden
in a derived class, then the default {\bf Deriv} and {\bf SecondDeriv}
methods can be used, which create instances of {\bf HCL\_OpDefaultDeriv\_d}
and {\bf HCL\_OpDefaultSecondDeriv\_d}.

See the following report for more details and concrete examples.

   "Implementing operators in HCL", Technical Report 99-22, Department of
   Computational and Applied Mathematics, Rice University, Houston, TX
   77251-1892.

downloadable from the HCL home page.
*/

class HCL_EvaluateOp_d: public HCL_Base
{
  friend class HCL_OpDefaultDeriv_d;
  friend class HCL_OpDefaultSecondDeriv_d;

protected:
  HCL_Vector_d * SavedImage;
  HCL_LinearOp_d * SavedDeriv;
  HCL_BiLinearOp_d * Saved2ndDeriv;

  /** {\bf DerivImage} computes the image of the derivative on dx,
    yielding dy.  Default implementation returns an error message and
    exits. To use the default {\bf Deriv} method in a derived class, you
    must override the default implementation of this method with one that
    actually implements the derivative. In that case the linear operator
    ({\bf HCL\_LinearOp\_d}) returned by {\bf Deriv} will be an instance
    of {\bf HCL\_OpDefaultDeriv\_d}. */

  virtual void DerivImage( const HCL_Vector_d & dx, 
			   HCL_Vector_d & dy ) const;

   /** {\bf DerivAdjImage} computes the image of the adjoint of the derivative
     on dy, yielding dx.  Default implementation returns an error message and
    exits. To use the default {\bf Deriv} method in a derived class, you
    must override the default implementation of this method with one that
    actually implements the derivative. In that case the linear operator
    ({\bf HCL\_LinearOp\_d}) returned by {\bf Deriv} will be an instance
    of {\bf HCL\_OpDefaultDeriv\_d}. */

  virtual void DerivAdjImage( const HCL_Vector_d & dy,
			      HCL_Vector_d & dx ) const;

   /** {\bf DerivInvImage} computes the image of the inverse of the derivative
     on dy, yielding dx. Default implementation returns an error message and
    exits. To use the default {\bf Deriv} method in a derived class, you
    must override the default implementation of this method with one that
    actually implements the derivative. In that case the linear operator
    ({\bf HCL\_LinearOp\_d}) returned by {\bf Deriv} will be an instance
    of {\bf HCL\_OpDefaultDeriv\_d}. */

  virtual void DerivInvImage( const HCL_Vector_d & dy,
			      HCL_Vector_d & dx ) const;

  /** {\bf DerivInvAdjImage} computes the image of the inverse of the
    adjoint of the derivative on dx, yielding dy. Default
    implementation returns an error message and exits. To use the
    default {\bf Deriv} method in a derived class, you must override
    the default implementation of this method with one that actually
    implements the derivative. In that case the linear operator ({\bf
    HCL\_LinearOp\_d}) returned by {\bf Deriv} will be an instance of
    {\bf HCL\_OpDefaultDeriv\_d}.  */

  virtual void DerivInvAdjImage( const HCL_Vector_d & dx,
                                           HCL_Vector_d & dy ) const;

   /** {\bf SecondDerivImage} computes the image of the second
    derivative on dx1 and dx2, yielding dy. Default implementation
    returns an error message and exits. To use the default {\bf
    SecondDeriv} method in a derived class, you must override the
    default implementation of this method with one that actually
    implements the derivative. In that case the bilinear operator
    ({\bf HCL\_BiLinearOp\_d}) returned by {\bf SeondDeriv} will be an
    instance of {\bf HCL\_OpDefaultDeriv\_d}. */


  virtual void SecondDerivImage( const HCL_Vector_d & dx1,
				 const HCL_Vector_d & dx2,
				 HCL_Vector_d & dy ) const;

   /** {\bf SecondDerivPartialAdjImage} computes the image of the
    so-called "partial adjoint" of the second derivative (which is a
    bilinear operator) on dx1 and dy, yielding dx2.  See the
    documentation for {\bf HCL\_BiLinearOp\_d} for more details about
    bilinear operators and partial adjoints.  Note that since the
    second derivative of a (sufficiently smooth) operator is
    symmetric, there is no ambiguity about which partial adjoint is
    intended. Default implementation returns an error message and
    exits. To use the default {\bf SecondDeriv} method in a derived
    class, you must override the default implementation of this method
    with one that actually implements the derivative. In that case the
    bilinear operator ({\bf HCL\_BiLinearOp\_d}) returned by {\bf
    SeondDeriv} will be an instance of {\bf
    HCL\_OpDefaultDeriv\_d}. */
  
  virtual void SecondDerivPartialAdjImage(const HCL_Vector_d & dx1,
					  const HCL_Vector_d & dy,
					  HCL_Vector_d & dx2 ) const;
public:

  // Default constructor
       HCL_EvaluateOp_d();
  // Copy constructor
       HCL_EvaluateOp_d( const HCL_EvaluateOp_d & );
  // Destructor
       virtual ~HCL_EvaluateOp_d();

  /** {\bf Domain} returns a reference to the domain of the underlying
    operator. Pure virtual: must be implemented in any derived
    instantiable class. */

  virtual HCL_VectorSpace_d & Domain() const = 0;

  /** {\bf Range} returns a reference to the range of the underlying
      operator. Pure virtual: must be implemented in any derived
      instantiable class.  */

  virtual HCL_VectorSpace_d & Range() const = 0;

  /** {\bf Image} computes the image of the operator and assigns it to
      y. Pure virtual: must be implemented in any derived instantiable
      class. */

  virtual void Image( HCL_Vector_d & y ) const = 0;

   /** {\bf ImageAdd} computes \TEX{$y\leftarrow aF(x)+by$}.  This method
     is given a default implementation in the base class, and need not
     be re-implemented in a derived class unless it is desired to
     obtain the gain in efficiency from "loop-fusion".  Note that the
     scalars a and b are optional and default to 1.0. */

  virtual void ImageAdd( HCL_Vector_d & y,
			 double a=1.0,double b=1.0 ) const;

  /** {\bf ImageAdd} computes \TEX{$y\leftarrow aF(x)+bz$}.  This method
    is given a default implementation in the base class, and need not
    be re-implemented in a derived class unless it is desired to
    obtain the gain in efficiency from "loop-fusion".  Note that the
    scalars a and b are optional and default to 1.0. */

  virtual void ImageAdd( const HCL_Vector_d & z,HCL_Vector_d & y,
                             double a=1.0,double b=1.0 ) const;

  /** {\bf ImageRef} returns a reference to the image of the operator at this
    point.  The user has a choice of calling {\bf Image}, in which case the
    storage for the result is managed by the calling routine, or
    calling {\bf ImageRef}, in which the Evaluation object manages the
    storage. This method is implemented in this class. */
      
  virtual const HCL_Vector_d & ImageRef() const;
  
  /** {\bf Deriv} computes the derivative of the operator.  Note that
    this method constructs an {\bf HCL\_LinearOp\_d}, which the
    calling program is responsible for deleting with {\bf
    HCL\_delete}. The default implementation requires that default
    implementations of {\bf DerivImage} and related methods be
    overridden with code that really does the job - see documentation
    on these classes, above. */
    
  virtual HCL_LinearOp_d * Deriv() const;

  /** {\bf DerivRef} returns a reference to the derivative at this point.
    The programmer has a choice of calling {\bf Deriv}, in which case the
    calling routine is responsible for managing the storage,
    or calling {\bf DerivRef}, in which case the Evaluation object
    does. This method is implemented in this class. */

  virtual const HCL_LinearOp_d & DerivRef() const;
  
  /** {\bf SecondDeriv} computes the second derivative of the
    operator.  The second derivative is a bilinear operator which
    knows how to compute all relevant adjoints. The default
    implementation requires that default implementations of {\bf
    SecondDerivImage} and related methods be overridden with code that
    really does the job - see documentation on these classes,
    above. */
  
  virtual HCL_BiLinearOp_d * SecondDeriv() const;
  
  /** {\bf SecondDerivRef} returns a reference to the second
    derivative at this point. The programmer has a choice of calling
    {\bf SecondDeriv}, in which case the calling routine is
    responsible for managing the storage, or calling {\bf
    SecondDerivRef}, in which case the Evaluation object does. This
    method is implemented in this class. */
  
  virtual const HCL_BiLinearOp_d & SecondDerivRef() const;

};

// "Look-ahead" declaration.
class HCL_OpDefaultEval_d;

/** {\bf HCL\_Op\_d} is the base class for (presumably nonlinear) operators.
Such an operator maps one vector space into another, say 
\TEX{
\[
F:X\rightarrow Y.
\]
}

The primary methods of this class are:
\begin{itemize}
\item {\bf Domain} Returns a reference to the domain of the operator.
This is used primarily for error-checking.
\item {\bf Range} Returns a reference to the range of the operator.
This is used primarily for error-checking.
\item {\bf Image} Computes the action of the operator on a vector.
Here is an example illustrating the use of {\bf Domain}, {\bf Range},
and {\bf Image}:
\begin{verbatim}
   void fcn( HCL_Op_d & F,HCL_Vector_d & x,HCL_Vector_d & y )
   {
      if( x.Space() != F.Domain() || y.Space() != F.Range() )
         // Error condition
      .
      .
      .
      F.Image( x,y ); // y <-- F(x)
      .
      .
      .
   }
\end{verbatim}
\item {\bf ImageAdd} Adds the image of the operator to the output vector.
\item {\bf Deriv} Creates the derivative of the operator at a point.
This derivative is a linear operator
\TEX{
\[
DF(x):X\rightarrow Y,
\]}
and this method constructs it as an {\bf HCL\_LinearOp\_d}.
\item {\bf SecondDeriv} Creates the second derivative of the operator at a
point.  This derivative is a bilinear operator
\TEX{
\[
D^2F(x):X\times X\rightarrow Y,
\]}
and this method constructs it as an {\bf HCL\_BiLinearOp\_d}.
\item {\bf Evaluate} Creates an "evaluation object" which knows the
image and derivatives of the operator at a given point.  The purpose of
the evaluation object is to accommodate the common situation in which
the computation of the image and derivatives involve some common
intermediate calculations.  The implementor of the operator can
put these intermediate calculations in the evaluation object, so
that they need only be done once at each point.  The implementor of
optimization algorithms should respect this by using {\bf Evaluate}
to get the image and/or derivatives instead of using {\bf Image},
{\bf Deriv}, and/or {\bf SecondDeriv}.
\end{itemize}

In addition to the above methods, {\bf HCL\_Op\_d} has a number of
protected virtual functions that allow the user to avoid writing
a distinct evaluation class if desired.  If these methods, which
include {\bf DerivImage}, {\bf SecondDerivImage}, and others, are over-ridden
in a derived class, then the default {\bf Evaluate} method can be used,
which creates an instance of {\bf HCL\_OpDefaultEval\_d}.

Defining operators is expected to be one of the main activities of users
of HCL, since the many optimization problems involve operators in the 
definition of the objective function or constraints.  As the last paragraph
suggests, there are a number of ways to implement operators.  These are
described in detail in the report

   "Implementing operators in HCL", Technical Report 99-22, Department of
   Computational and Applied Mathematics, Rice University, Houston, TX
   77251-1892.

downloadable from the HCL home page.
*/

class HCL_Op_d: public HCL_Base
{
   friend class HCL_OpDefaultEval_d;

   protected:

   /** {\bf DerivImage} computes the image of the derivative on dx,
     yielding dy.  Default implementation returns an error message and
     exits. To use the default {\bf Evaluate} method in a derived
     class, you must either (1) override the default implementation of
     {\bf Deriv1}, or, failing that, (2) override the default
     implementation of this method and its relatives with code that
     actually implements the linear map defined by the derivative. In
     either case the evaluation object will then be an instance of
     {\bf HCL\_OpDefaultEval\_d}. */

   virtual void DerivImage( const HCL_Vector_d & x,
			    const HCL_Vector_d & dx,
			    HCL_Vector_d & dy ) const;

   /** {\bf DerivAdjImage} computes the image of the adjoint of the
     derivative on dy, yielding dx. Default implementation returns an
     error message and exits. To use the default {\bf Evaluate} method
     in a derived class, you must either (1) override the default
     implementation of {\bf Deriv1}, or, failing that, (2) override
     the default implementation of this method and its relatives with
     code that actually implements the linear map defined by the
     derivative. In either case the evaluation object will then be an
     instance of {\bf HCL\_OpDefaultEval\_d}. */

   virtual void DerivAdjImage( const HCL_Vector_d & x,
			       const HCL_Vector_d & dy,
			       HCL_Vector_d & dx ) const;

   /** {\bf DerivInvImage} computes the image of the inverse of the
     derivative on dy, yielding dx. Default implementation returns an
     error message and exits. To use the default {\bf Evaluate} method
     in a derived class, you must either (1) override the default
     implementation of {\bf Deriv1}, or, failing that, (2) override
     the default implementation of this method and its relatives with
     code that actually implements the linear map defined by the
     derivative. In either case the evaluation object will then be an
     instance of {\bf HCL\_OpDefaultEval\_d}. */

   virtual void DerivInvImage( const HCL_Vector_d & x,
			       const HCL_Vector_d & dy,
			       HCL_Vector_d & dx ) const;

   /** {\bf DerivInvAdjImage} computes the image of the inverse of the
     adjoint of the derivative on dx, yielding dy. Default implementation
     returns an error message and exits. To use the default {\bf Evaluate}
     method in a derived class, you must either (1) override the default
     implementation of {\bf Deriv1}, or, failing that, (2) override the
     default implementation of this method and its relatives with code that
     actually implements the linear map defined by the derivative. In
     either case the evaluation object will then be an instance of {\bf
     HCL\_OpDefaultEval\_d}. */

   virtual void DerivInvAdjImage( const HCL_Vector_d & x,
				  const HCL_Vector_d & dx,
				  HCL_Vector_d & dy ) const;

   /** {\bf Image1} computes the image of the operator on x, yielding
     y.  Default implementation returns an error message and
     exits. The default implementation of {\bf Evaluate} uses this
     method, so you must either override it or override {\bf Evaluate}
     to obtain a working child class.  */

   virtual void Image1( const HCL_Vector_d & x,HCL_Vector_d & y ) const;

   /** {\bf Deriv1} computes the derivative of the operator at x.
     Default implementation returns a null value. The default
     implementation of {\bf Evaluate} calls this method, and returns
     its value if not null. If the return value is null, it uses the
     {\bf HCL\_OpDefaultDeriv} class to construct a derivative
     implementation; this construction in turn uses the {\bf
     DerivImage}, {\bf DerivAdjImage}, etc. methods. Therefore to
     obtain a working operator class, with an implemented
     derivative, you must either (1) override {\bf Evaluate}, (2)
     override {\bf Deriv1} (this method), or (3) override {\bf
     DerivImage}, {\bf DerivAdjImage}, etc. */

   virtual HCL_LinearOp_d * Deriv1( const HCL_Vector_d & x ) const;

   /** Writing $B = D^2F(x)$, this method computes $dy=B(dx1,dx2)$.
     If this method and the related protected image methods are
     implemented, then the default {\bf Evaluate} method can be used.
     The evaluation method will then be an instance of {\bf
     HCL\_OpDefaultEval\_d}.  Note that either the pair {\bf
     SecondDerivImage}, {\bf SecondDerivPartialAdjImage} or the method
     {\bf SecondDeriv1} must be implemented to enable the use of the
     default second derivative construction. */

   virtual void SecondDerivImage( const HCL_Vector_d & x,
				  const HCL_Vector_d & dx1,
				  const HCL_Vector_d & dx2,
				  HCL_Vector_d & dy ) const;

   /** This method computes $dx2=B(dx1,\cdot)^*dy$.  If this method
     and the related protected image methods are implemented, then the
     default {\bf Evaluate} method can be used.  The evaluation method
     will then be an instance of {\bf HCL\_OpDefaultEval\_d}.  Note
     that either the pair {\bf SecondDerivImage}, {\bf
     SecondDerivPartialAdjImage} or the method {\bf SecondDeriv1} must
     be implemented to enable the use of the default second derivative
     construction.*/

   virtual void SecondDerivPartialAdjImage(const HCL_Vector_d & x,
					   const HCL_Vector_d & dx1,
					   const HCL_Vector_d & dy,
					   HCL_Vector_d & dx2 ) const;

   /** {\bf SecondDeriv1} computes the second derivative of the
     operator at x.  Its default implementation returns NULL. If the
     default implementation is overridden to return a legal {\bf
     HCL\_BiLinearOp\_d} object, then that object is returned by the
     {\bf SecondDeriv} method of the default evaluation class {\bf
     HCL\_OpDefaultEval\_d}. Otherwise, this method returns a second
     derivative object that calls the methods {\bf SecondDerivImage},
     {\bf SecondDerivPartialAdjImage}, which must then be implemented
     to provide a working second derivative. */
     
   virtual HCL_BiLinearOp_d * SecondDeriv1( const HCL_Vector_d & x ) const;

   public:  

   // Default constructor.
      HCL_Op_d();
   // Copy constructor.
      HCL_Op_d( const HCL_Op_d & );
   // Destructor.
      virtual ~HCL_Op_d();

   /** Domain space access.  Returns a reference to the {\bf
     HCL\_VectorSpace\_d} which forms the domain of the operator.
     Note that, in the current design, the domain of an operator must
     be an entire vector space, although many nonlinear operators are
     only defined on an open subset of a vector space.  This is a flaw
     in the design of HCL which may be corrected in some future
     version of the software.  A partial work-around to this problem
     is provided by the {\bf MaxStep} method. Pure virtual: must be
     implemented in any derived instantiable class. */

   virtual HCL_VectorSpace_d & Domain() const = 0;

  /** Range space access. Returns a reference to the range of the
    operator, which is an {\bf HCL\_VectorSpace\_d}. Pure virtual:
    must be implemented in any derived instantiable class. */
   
   virtual HCL_VectorSpace_d & Range() const = 0;

  /** {\bf MaxStep} computes the greatest a such that x + a*dir lies
      in the domain.  By default (i.e. unless this virtual function is
      overridden in a derived class), the domain is assumed to be the
      whole space, and MaxStep returns the largest floating point
      number. This method is used in some of the optimization
      algorithms and provides a partial solution to the problem that
      many operators are not defined on an entire vector space, but
      rather only on a subset. */

   virtual double MaxStep( const HCL_Vector_d & x, 
			   const HCL_Vector_d &dir) const;

   /** {\bf Image} computes the action of the operator on x, giving y.
     It does this by creating an evaluation object via
     {\bf Evaluate}, extracting the image, and deleting the
     evaluation object.  This method is defined in this (the base)
     class, so it need not be redefined in concrete derived classes. */

   virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;

   /** {\bf ImageAdd} computes \TEX{$y\leftarrow aF(x)+by$}.  (Note
     that the scalars a and b are optional and default to 1.0.)  It
     does this by creating an evaluation object via {\bf Evaluate},
     extracting the image, and deleting the evaluation object.  This
     method is defined in this (the base) class, so it need not be
     redefined in concrete derived classes. However greater efficiency
     may be obtained in some cases by overriding the default
     implementation, through ``loop fusion''. */

   virtual void ImageAdd( const HCL_Vector_d & x,HCL_Vector_d & y,
			  double a=1.0,double b=1.0 ) const;

   /** {\bf ImageAdd} computes \TEX{$y\leftarrow aF(x)+bz$}.  (Note
       that the scalars a and b are optional and default to 1.0.)  It
       does this by creating an evaluation object via {\bf Evaluate},
       extracting the image, and deleting the evaluation object.  This
       method is defined in this (the base) class, so it need not be
       redefined in concrete derived classes. However greater
       efficiency may be obtained in some cases by overriding the
       default implementation, through ``loop fusion''. */

   virtual void ImageAdd( const HCL_Vector_d & x,const HCL_Vector_d & z,
			  HCL_Vector_d & y,
			  double a=1.0,double b=1.0 ) const;

   /** {\bf Deriv} computes the derivative, as an {\bf
       HCL\_LinearOp\_d}, of the operator at x.  It does this by
       creating an evaluation object via {\bf Evaluate}, extracting
       the derivative, and deleting the evaluation object.  This
       method is defined in this (the base) class, so it need not be
       redefined in concrete derived classes. */

   virtual HCL_LinearOp_d * Deriv( const HCL_Vector_d & x ) const;

   /** {\bf SecondDeriv} computes the second derivative of the operator at a
       point.  It does this by creating an evaluation object, using 
       {\bf Evaluate}, getting the second derivative, and deleting
       the evaluation object.  The method is implemented in this (the base)
       class, so it need not be redefined for derived classes. */

   virtual HCL_BiLinearOp_d * SecondDeriv( const HCL_Vector_d & x ) const;

   /** {\bf Evaluate} creates an evaluation object which knows how to
       compute the image, first derivative, and second derivative,
       along with all relevant adjoints, at a point.  The purpose of
       the evaluation object is to accommodate the common situation in
       which the computation of the image and derivatives involve some
       common intermediate calculations.  See the class documentation
       for more information on evaluation objects. The default
       implementation uses the {\bf Image1} method and either the {\bf
       Deriv1} and {\bf SecondDeriv1} methods, or {\bf DerivImage},
       {\bf DerivAdjImage},...,{\bf SecondDerivImage},...These return
       null values or exit with an error message in their default
       implementations, so for each level of functionality (image
       alone, image plus first derivative,...) some combination of
       these methods must be overriden in any derived
       class. Alternatively, the user may override the default {\bf
       Evaluate} method and provide standalone classes for the
       derivatives. */

   virtual HCL_EvaluateOp_d * Evaluate( const HCL_Vector_d & x ) const;

   /** Check analytic derivative.  This method computes the relative difference
       between the directional derivative as computed from the analytic
       derivative and the same estimated by a central difference formula.
       This is done at n equally-spaced points on the line segment
       [x+hmin*y,x+hmax*y].  The parameters hmin, hmax, and n are optional
       and default to 0.1, 1.0, and 10, respectively.
       If the analytic derivative is correctly implemented and the function
       is sufficiently smooth, then these errors will lie on a parabola when
       plotted against the step size. The results are sent to the output
       stream str. This method is implemente in this class. */

   int CheckDeriv( const HCL_Vector_d &,const HCL_Vector_d &,
		   ostream & str,int n=10,double hmin=0.1,double hmax=1.0 ) const;
   /** {\bf CheckDerivAdj} calls {\bf CheckDeriv} to test the
     correctness of the derivative, and then calls the {\bf
     HCL\_LinearOp\_d} method {\bf CheckAdj} to test the correctness
     of its adjoint. This method is implemented in this class. */

   int CheckDerivAdj( const HCL_Vector_d &,const HCL_Vector_d &,
		      ostream & str,int n=10,double hmin=0.1,double hmax=1.0 ) const;

   /** Check analytic second derivative.  This method computes the
       relative difference between the second directional derivative
       as computed from the analytic derivative and the same estimated
       by a central difference formula. This is done at n
       equally-spaced points on the line segment [x+hmin*y,x+hmax*y].
       The parameters hmin, hmax, and n are optional and default to
       0.1, 1.0, and 10, respectively.  If the analytic derivative is
       correctly implemented and the function is sufficiently smooth,
       then these errors will lie on a parabola when plotted against
       the step size. The results are sent to the output stream
       str. This method is implemented in this class. */

   int CheckSecondDeriv( const HCL_Vector_d &,const HCL_Vector_d &,
			 ostream & str,int n=10,double hmin=0.1,double hmax=1.0 ) const;

   /** {\bf CheckSecondDerivAdj} calls {\bf CheckSecondDeriv} to test the
     correctness of the second derivative, and then calls the
     {\bf HCL\_LinearOp\_d} method {\bf CheckAdj} to test the correctness of
     the three related adjoints. This method is implemented in this class. */

   int CheckSecondDerivAdj( const HCL_Vector_d &,const HCL_Vector_d &,
			    ostream & str,int n=10,double hmin=0.1,double hmax=1.0 ) const;

};

/** {\bf HCL\_OpDefaultEval\_d} is the default "evaluation" class for
nonlinear operators.  If
\TEX{
\[
F:X\rightarrow Y
\]}
is a nonlinear operator represented by an instance of {\bf HCL\_Op\_d},
then the corresponding evaluation class represents the triple
\TEX{
\[
(F(x),DF(x),D^2F(x)).
\]}
In many cases, there is a significant efficiency advantage to creating
a specific evaluation class for a given operator, since then intermediate
calculations that are common to the image, derivative, and second derivative
can be re-used.  However, if there is no such efficiency advantage to be
gained, then the default evaluation class {\bf HCL\_OpDefaultEval\_d} can
be used.

The various mechanisms for implementing evaluation classes, including
the use of this default class, are explained in detail, with concrete
examples, in the report

   "Implementing operators in HCL", Technical Report 99-22, Department of
   Computational and Applied Mathematics, Rice University, Houston, TX
   77251-1892.

   available on the HCL home page.

   This class is concrete: all of its methods are implemented, and 
   are not intended to be overridden.

*/
class HCL_OpDefaultEval_d: public HCL_EvaluateOp_d
{
   private:
      HCL_Vector_d * x;
      const HCL_Op_d * op;
   protected:

   /** {\bf DerivImage} computes the image of the derivative on dx,
       yielding dy.  This method calls the corresponding method from
       the underlying operator class. */
      virtual void DerivImage( const HCL_Vector_d & dx,
                                     HCL_Vector_d & dy ) const
      {
         op->DerivImage( *x,dx,dy );
      }
   /** {\bf DerivAdjImage} computes the image of the adjoint of the
       derivative on dy, yielding dx.  This method calls the corresponding
       method from the underlying operator class. */
      virtual void DerivAdjImage( const HCL_Vector_d & dy,
                                        HCL_Vector_d & dx ) const
      {
         op->DerivAdjImage( *x,dy,dx );
      }
   /** {\bf DerivInvImage} computes the image of the inverse of the
       derivative on dy, yielding dx.  This method calls the corresponding
       method from the underlying operator class. */
      virtual void DerivInvImage( const HCL_Vector_d & dy,
                                        HCL_Vector_d & dx ) const
      {
         op->DerivInvImage( *x,dy,dx );
      }
   /** {\bf DerivInvAdjImage} computes the image of the inverse of the adjoint
       of the derivative on dx, yielding dy.  This method calls the
       corresponding method from the underlying operator class. */
      virtual void DerivInvAdjImage( const HCL_Vector_d & dx,
                                           HCL_Vector_d & dy ) const
      {
         op->DerivInvAdjImage( *x,dx,dy );
      }
   /** {\bf SecondDerivImage} computes the image of the second derivative
       on (dx1,dx2), yielding dy.  This method calls the corresponding
       method from the underlying operator class. */
      virtual void SecondDerivImage( const HCL_Vector_d & dx1,
                                     const HCL_Vector_d & dx2,
                                           HCL_Vector_d & dy ) const
      {
         op->SecondDerivImage( *x,dx1,dx2,dy );
      }
   /** {\bf SecondDerivPartialAdjImage} computes the image of the so-called
       "partial adjoint" of the second derivative on (dx1,dy), yielding dx1.
       This method calls the corresponding method from the underlying operator
       class. */
      virtual void SecondDerivPartialAdjImage(const HCL_Vector_d & dx1,
                                              const HCL_Vector_d & dy,
                                                    HCL_Vector_d & dx2 ) const
      {
         op->SecondDerivPartialAdjImage( *x,dx1,dy,dx2 );
      }
   public:

   // Default constructor---disabled.
      HCL_OpDefaultEval_d();
   // Copy constructor---disabled.
      HCL_OpDefaultEval_d( const HCL_OpDefaultEval_d & );
   /** Usual constructor; requires a pointer to an operator and a reference
       to the input vector.  Note that an independent copy of the input
       vector is made. */
      HCL_OpDefaultEval_d( const HCL_Op_d * op,const HCL_Vector_d & in );
   // Destructor
      virtual ~HCL_OpDefaultEval_d();

   /** {\bf Domain} returns a reference to the domain of the underlying
       operator. */
      virtual HCL_VectorSpace_d & Domain() const { return op->Domain(); }
   /// {\bf Range} returns a reference to the range of the underlying operator.
      virtual HCL_VectorSpace_d & Range() const { return op->Range(); }

   /// {\bf Image} computes the image of the operator.
      virtual void Image( HCL_Vector_d & y ) const { op->Image1( *x,y ); }
   /** {\bf Deriv} computes the derivative of the operator.  Note that
       this method constructs an {\bf HCL\_LinearOp\_d}, which the calling
       program is responsible for deleting with {\bf HCL\_delete}. */
      virtual HCL_LinearOp_d * Deriv() const;
   /** {\bf SecondDeriv} computes the second derivative of the operator.
       The second derivative is a bilinear operator which knows how to
       compute all relevant adjoints. */
      virtual HCL_BiLinearOp_d * SecondDeriv() const;
 
   /** {\bf Write} prints a description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_OpDefaultDeriv\_d} is a linear operator class implementing
the derivative of a nonlinear operator.  It exists to relieve the implementor
of an operator class from the necessity of creating a distinct class to
represent the derivative if there is no (efficiency) advantage gained by
creating such a class.  In order to use this default derivative class,
the implementor of the operator class must implement the protected virtual
functions {\bf DerivImage}, {\bf DerivAdjImage}, etc.

The various mechanisms for implementing operator classes, including
the use of this default class, are explained in detail, with concrete
examples, in the report

   "Implementing operators in HCL", Technical Report 99-22, Department of
   Computational and Applied Mathematics, Rice University, Houston, TX
   77251-1892.

downloadable from the HCL home page.

   This class is concrete: all of its methods are implemented, and 
   are not intended to be overridden.

*/
class HCL_OpDefaultDeriv_d: public HCL_LinearOp_d
{
   private:
       const HCL_EvaluateOp_d * eval;
   public:  

  // Default constructor---disabled.
      HCL_OpDefaultDeriv_d();
  // Copy constructor---disabled.
      HCL_OpDefaultDeriv_d( const HCL_OpDefaultDeriv_d & L );
  /** Usual constructor; just needs an operator's evaluation object. */
      HCL_OpDefaultDeriv_d( const HCL_EvaluateOp_d * ev );
  // destructor
      virtual ~HCL_OpDefaultDeriv_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;
  /// {\bf Image} computes the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\bf Write} prints a description of the operator. */
      virtual ostream & Write( ostream & str ) const;

};

/** {\bf HCL\_OpDefaultSecondDeriv\_d} is a bilinear operator class
implementing the second derivative of a nonlinear operator.  It exists
to relieve the implementor of an operator class from the necessity of
creating a distinct class to represent the derivative if there is no
(efficiency) advantage gained by creating such a class.  In order to
use this default derivative class, the implementor of the operator
class must implement the protected virtual functions {\bf
SecondDerivImage}, {\bf SecondDerivPartialAdjImage}.

The various mechanisms for implementing operator classes, including
the use of this default class, are explained in detail, with concrete
examples, in the report

   "Implementing operators in HCL", Technical Report 99-22, Department of
   Computational and Applied Mathematics, Rice University, Houston, TX
   77251-1892.

downloadable from the HCL home page.

   This class is concrete: all of its methods are implemented, and 
   are not intended to be overridden.

*/
class HCL_OpDefaultSecondDeriv_d: public HCL_BiLinearOp_d
{
   private:
      const HCL_EvaluateOp_d * eval;
   protected:

   /** The method {\bf PartialAdjImage} implements the mappings
       $(x,z)\mapsto B(x,\cdot)^*z$ ($i=1$) and $(y,z)\mapsto B(\cdot,y)^*z$
       ($i=2$). It is implemented to enable the use of the default linear
       operator class {\bf HCL\_RestrictedBiLinearOp\_d} to represent the
       operators {\bf Op1}, {\bf Op2}, {\bf Op3}. */
      virtual void PartialAdjImage( int i,const HCL_Vector_d & xy,
                                              const HCL_Vector_d & z,
                                                    HCL_Vector_d & yx ) const;
   public:

  // Default constructor---disabled.
      HCL_OpDefaultSecondDeriv_d();
  // Copy constructor---disabled.
      HCL_OpDefaultSecondDeriv_d( const HCL_OpDefaultSecondDeriv_d & L );
  /** Usual constructor; just needs an operator's evaluation object. */
      HCL_OpDefaultSecondDeriv_d( const HCL_EvaluateOp_d * ev );
  // destructor
      virtual ~HCL_OpDefaultSecondDeriv_d();

   /** Access to the domain of the first argument.  Returns a reference
       to an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain1() const { return eval->Domain(); }
   /** Access to the domain of the second argument.  Returns a reference
       to an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain2() const { return eval->Domain(); }
   /** Access to the range.  Returns a reference to a
       {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const { return eval->Range(); }

   /// Image computes the action of the operator on the pair (x,y), giving z.
      virtual void Image( const HCL_Vector_d & x,const HCL_Vector_d & y,
                                HCL_Vector_d & z ) const;

  /** {\bf Write} prints a description of the object. */
      virtual ostream & Write( ostream & str ) const;
};



#endif

