/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef __HCL_PartitionedRnVector_d
#define __HCL_PartitionedRnVector_d

#include "HCL_Rn_d.h"
#include "HCL_ProductVector_d.h"

class HCL_PartitionedRnVector_d;

/** {\bf HCL\_PartitionedRnSpace\_d} represents a product vector space,
in which each factor is an instance of {\bf HCL\_RnSpace\_d}.  The
corresponding vector class, {\bf HCL\_PartitionedRnVector\_d},
implements the product vector as a "long vector"; that is, the components
are stored in a single array. 

See {\bf HCL\_RnSpace\_d} and {\bf HCL\_ProductSpace\_d} for more details. */
class HCL_PartitionedRnSpace_d: public HCL_ProductSpace_d
{

   private:

      static char Type;
      int n,numinit;
      int IdenticalFlag;
      HCL_RnSpace_d ** rnsp;

   public:

   // Default constructor---disabled.
      HCL_PartitionedRnSpace_d();
   /// Copy constructor.
      HCL_PartitionedRnSpace_d(const HCL_PartitionedRnSpace_d & x);

   /** This constructor creates a space of n identical factors, each
       of dimension m. */
      HCL_PartitionedRnSpace_d( int m, int n );
   /** This constructor creates a space of n factors.  The dimension of each
       factor must be specified with the {\bf SetNext} method. */
      HCL_PartitionedRnSpace_d( int n );
   // Destructor.
      virtual ~HCL_PartitionedRnSpace_d();

   /** Get type of vector space for comparison.
       {\bf GetType} makes it possible to check that two vector spaces
       are of the same derived class before comparing them.  {\bf GetType} just
       returns the address of the static member {\bf Type}, cast to void*.
       This is a primitive form of Run Time Type Information. */
      virtual void * GetType() const ;

   /** Virtual constructor.  This method constructs a vector belonging to
       this vector space and returns a pointer to it.  Note that the
       calling routine is responsible for deleting the vector when it
       is no longer needed.  {\bf Member} is used to create temporary vectors
       "abstractly", i.e. without knowing the exact type of the vector
       desired.  */
      virtual HCL_Vector_d * Member() const ;

   /** Virtual copy constructor.  This method constructs a copy of this
       vector space and returns a pointer to it. */
      virtual HCL_VectorSpace_d * Replicate() const ;

   /// {\bf Number} returns the number of factors in the product.
      virtual int Number() const ;

   /// The () operator returns a reference to the ith factor in the product.
      virtual HCL_VectorSpace_d & operator()( int i ) const ;

   /** {\bf SetNext} is used to specify the dimension of the factors in
       the product.  (Note that if all of the factors have the same
       dimension, then a special constructor can be used, in which case
       {\bf SetNext} need not be called.) */
      void SetNext( int m );
   /** {\bf SetNext} is used to specify the next factor in
       the product.  (Note that if all of the factors have the same
       dimension, then a special constructor can be used, in which case
       {\bf SetNext} need not be called.) */
      void SetNext( HCL_RnSpace_d * s );
};

/** {\bf HCL\_PartitionedRnVector\_d} represents a vector from a product space
in which each factor is an instance of {\bf HCL\_RnSpace\_d}.  The product
vector is implemented as a "long vector"; that is, the components
are stored in a single array.  The purpose of the class is to make it
convenient to view the vector as a product (and hence convenient to
manipulate the factors), but also to enable the use of direct linear algebra
techniques on the long vector.

See {\bf HCL\_RnVector\_d} and {\bf HCL\_ProductVector\_d} for more details. */
class HCL_PartitionedRnVector_d: public HCL_ProductVector_d
{
   private:
      int n,dim;
      HCL_RnVector_d ** rnvec;
      HCL_PartitionedRnSpace_d * sp;
      double * data;

   public:

   // Default constructor---disabled.
      HCL_PartitionedRnVector_d();
   // Copy constructor---disabled.
      HCL_PartitionedRnVector_d(const HCL_PartitionedRnVector_d & p);
   /** Special constructor for the case of n identical factors, each with
       dimension m. */
      HCL_PartitionedRnVector_d(int m, int n);
   /** Special constructor; reads the vector from an ASCII file.  The format of
       the file is as follows:  The first entry is an integer n, the
       number of factors.  Subsequent entries consist of n blocks, each
       having the dimension of the factor followed by its components. */
      HCL_PartitionedRnVector_d( char * fname );
   /** General constructor; constructs a vector based on an instance of
       {\bf HCL\_PartitionedRnVector\_d}. */
      HCL_PartitionedRnVector_d( HCL_PartitionedRnSpace_d * s );
   // Destructor.
      virtual ~HCL_PartitionedRnVector_d();

   /** Return the dimension of the vector space. */
      virtual int Dim() const { return dim; }
   /** The () operator returns a reference to the ith factor.  The result
       can be safely cast to type {\bf HCL\_RnVector\_d}. */
      virtual HCL_Vector_d & operator()(int i) const ; 

   /** {\bf Data} returns a pointer to the (long) array of components. */
      virtual double * Data() { return data; }

   /// {\bf Space} returns a reference to the underlying vector space.
      virtual HCL_VectorSpace_d & Space() const ;

   /** {\bf Save} stores the vector in an ASCII file, in the following format:
       The first entry is an integer n, the number of factors.  Subsequent
       entries consist of n blocks, each having the dimension of the factor
       followed by its components. */
      void Save( char * fname ) const;
};

#endif
