/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_ProductVector_d_H
#define HCL_ProductVector_d_H


#include "HCL_Vector_d.h"
#include "maxmin_d.h"

/** Index arrays for indirect addressing into product spaces and vectors. */

class iArray: public HCL_Base {

private:

  int len;
  int * ind;

public:

  //@name Constructors, destructor
  //@{
  /// Default constructor - disallowed
  iArray(){cerr<<"Error: no default constr for iArray"<<endl;exit(1);}
  /// Copy constructor - disallowed
  iArray(iArray & I){cerr<<"Error: no copy constr for iArray"<<endl;exit(1);}
  /// Standard constructor. Creates index array of length {\tt n}
  iArray(int n){len=n; ind=new int[len];}
  /// Destructor
  ~iArray(){delete [] ind;}
  //@}

  /// returns length of index array
  int Number() const {return len;}

  /// alters ith component
  void Set(int i, int j) { ind[i-1]=j; }

  /// returns ith component
  int Index(int i) const { return ind[i-1]; }

  /// lists out contents of array
  ostream & Write(ostream & str) const {
    for (int i=0; i<len; i++) {
      str<<i+1<<" --> "<<ind[i]<<endl;
    }
    return str;
  }
};

class HCL_ProductVector_d;

/** {\bf HCL\_ProductSpace\_d} is the base class for product vector space
classes.  Such a class represents a vector space of the form
\TEX{\[
X=X_1\times X_2\times\cdots\times X_n,
\]}
where each of \TEX{$X_1,X_2,\ldots,X_n$} is a vector space represented by
an instance of {\bf HCL\_VectorSpace\_d}.

In addition to the usual {\bf HCL\_VectorSpace\_d} methods, this class
can return the number of factors in the product (n) with the {\bf Number}
method, and can return a reference to any one of the factors with the
() operator. */

class HCL_ProductSpace_d : public HCL_VectorSpace_d
{
   public:  

   // Default constructor---disabled.
      HCL_ProductSpace_d(){}
   // Copy constructor.
      HCL_ProductSpace_d(const HCL_ProductSpace_d & x){}
   // Destructor.
      virtual ~HCL_ProductSpace_d(){}

   /** Operator ==; true (nonzero) if the input space agrees with this
       vector space, otherwise false (zero). */
      virtual int operator == (const HCL_VectorSpace_d & xx) const;

   /* Get type of vector space for comparison.
       {\bf GetType} makes it possible to check that two vector spaces
       are of the same derived class before comparing them. */
      virtual void * GetType() const = 0;

   /** Virtual constructor.  This method constructs a vector belonging to
       this vector space and returns a pointer to it.  Note that the
       calling routine is responsible for deleting the vector when it
       is no longer needed. */
      virtual HCL_Vector_d * Member() const = 0;
   /** Virtual copy constructor.  {\bf Replicate} constructs a copy of this
       vector space and returns a pointer to it. */
      virtual HCL_VectorSpace_d * Replicate() const = 0;

   /** Return the dimension of the vector space. */
      virtual int Dim() const;

   /// {\bf Number} returns the number of factors.
      virtual int Number() const = 0;
   /// This operator returns a reference to the ith factor.
      virtual HCL_VectorSpace_d & operator()( int i ) const = 0;

   /// Write prints out the factors in the product.
      virtual ostream & Write(ostream & str) const;
};

/** {\bf HCL\_ProductVector\_d} represents a vector from a product space,
that is, from a space of the form
\TEX{\[
X=X_1\times X_2\times\cdots\times X_n,
\]}
where each of \TEX{$X_1,X_2,\ldots,X_n$} is a vector space represented by
an instance of {\bf HCL\_VectorSpace\_d}.  This is an abstract base class;
it serves to define the properties of every concrete product vector class.

In addition to the usual {\bf HCL\_Vector\_d} methods, an instance of this
class can return the number of factors with the {\bf Number} method,
can identify the space to which it belongs, explicitly as an instance of
{\bf HCL\_ProductVectorSpace\_d}, with the {\bf ProductSpace} method,
and can return a reference to the ith factor with the () operator.
*/

class HCL_ProductVector_d: public HCL_Vector_d
{
   public:

   /* Default constructor. */
      HCL_ProductVector_d(){}
   // Copy constructor.
      HCL_ProductVector_d(const HCL_ProductVector_d & p){}
   // Destructor.
      virtual ~HCL_ProductVector_d(){}

   /// Operator() returns a reference to the ith factor.
      virtual HCL_Vector_d & operator()(int i) const = 0; 
   /// Space returns a reference to the space to which this vector belongs.
      virtual HCL_VectorSpace_d & Space() const = 0;
   /** ProductSpace returns a reference to the space to which this vector belongs,
       explicitly as a product space. */
      virtual HCL_ProductSpace_d & ProductSpace() const 
        { return (HCL_ProductSpace_d &)Space();}
   /// Number returns the number of factors in the product.
      virtual int Number() const 
        { return ProductSpace().Number();}

   /**@name Vector operations (z is the object invoking the method). */
   //@{
   /// z <-- x
      virtual void Copy( const HCL_Vector_d & x );
   /// z <-- (-z)
      virtual void Neg();
   /// z <-- a*z
      virtual void Mul( const double & a );
   /// z <-- a*x
      virtual void Mul( const double & a,const HCL_Vector_d & x );
   /// z <-- z + x
      virtual void Add( const HCL_Vector_d & x );
   /// z <-- x + y
      virtual void Add( const HCL_Vector_d & x,
                        const HCL_Vector_d & y );
   /// z <-- z - x
      virtual void Sub( const HCL_Vector_d & x );
   /// z <-- x - y
      virtual void Sub(const HCL_Vector_d & x,
                       const HCL_Vector_d & y );
   /// z <-- a*z + x
      virtual void ScaleAdd(const double & a,
                            const HCL_Vector_d & x );
   /// z <-- z + a*x
      virtual void AddScale(const double & a,
                            const HCL_Vector_d & x );
   /// z <-- x + a*y
      virtual void AddScale(const double & a, const HCL_Vector_d & x,
                            const HCL_Vector_d & y );

   /// Inner product of z with x.
      virtual double Inner( const HCL_Vector_d & x ) const;
   /// Norm of z.
      virtual double Norm() const;
   /// Norm squared of z.
      virtual double Norm2() const;
   /// z <-- 0
      virtual void Zero();
   /** Returns a "random" vector.  The meaning of random is left up to
       the implementor of each derived class. */
     virtual void Random();
   //@}

      /**@name Array operations - ``Matlab'' methods. 
	In describing these,
       \TEX{$c_i$} denotes the ith component of the current array, i.e. the
       data array of the HCL\_Vector object */
      //@{
	/// \TEX{$c_i = a,\,\,i=1,n$}
     virtual void Fill( double a );
     /** \TEX{$c_i = c_i + a$} - overload of Add with scalar argument
         {\em a la} Matlab. */
     virtual void Add( double a );
     /** \TEX{$c_i = x_i + a$} - overload of Add with scalar argument
         {\em a la} Matlab. */
     virtual void Add( const HCL_Vector_d & x, double a );
     /// \TEX{$c_i = c_i*x_i$} - matrix multiply by diag(x)
     virtual void DiagScale( const HCL_Vector_d & x );
     /// \TEX{$c_i = y_i*x_i$}
      virtual void DiagScale( const HCL_Vector_d & x,
                              const HCL_Vector_d & y );
   /** \TEX{$c_i = c_i/{\rm maxmod}(x_i,tol)$} - matrix division by diag(x). 
     Safeguarded against overflow - if tol is less
     that the smallest safe divisor sfmin as defined by LAPACK::[SD]LAMCH, then
     tol is replaced by sfmin - thus the default is tol=0.0, which effectively
     means tol=sfmin. */
      virtual void DiagRecipScale( const HCL_Vector_d & x,
				   double tol=0.0);
   /** \TEX{$c_i = y_i/{\rm maxmod}(x_i,tol)$} - matrix division by diag(x).
     Safeguarded against overflow - if tol is less
     that the smallest safe divisor sfmin as defined by LAPACK::[SD]LAMCH, then
     tol is replaced by sfmin - thus the default is tol=0.0, which effectively
     means tol=sfmin. */
      virtual void DiagRecipScale( const HCL_Vector_d & x,
				   const HCL_Vector_d & y,
				   double tol=0.0);

      /// returns largest of components or -FLT\_MAX 
      virtual double Max() const;
      /// \TEX{$c_i = {\rm max}(c_i,x_i) $}
      virtual void Max( const HCL_Vector_d & x );
      /// \TEX{$c_i = {\rm max}(y_i,x_i) $}
      virtual void Max( const HCL_Vector_d & x,
                        const HCL_Vector_d & y );
      /// returns smallest of components or FLT\_MAX
      virtual double Min() const;
      /// \TEX{$c_i = {\rm min}(c_i,x_i) $}
      virtual void Min( const HCL_Vector_d & x );
      /// \TEX{$c_i = {\rm min}(y_i,x_i) $}
      virtual void Min( const HCL_Vector_d & x,
                        const HCL_Vector_d & y );
      /// returns \TEX{$\sum_{i=1}^n\,c_i$}
      virtual double Sum() const;
      /// \TEX{$c_i = |c_i|$}
      virtual void Abs();
      /// \TEX{$c_i = |x_i|$}
      virtual void Abs( const HCL_Vector_d & x );
      /** \TEX{$c_i = {\rm sign}\,(c_i)$}. Sign returns 1 for positive
           arguments, -1 for negative arguments, and 0 for zero arguments. */
      virtual void Sign();
      /// \TEX{$c_i = {\rm sign}\,(c_i - t)$}
      virtual void Sign( double t );
      /// \TEX{$c_i = {\rm sign}\,(c_i-x_i)$}
      virtual void Sign( const HCL_Vector_d & x );
      /// \TEX{$c_i = {\rm sign}\,(x_i-y_i)$}
      virtual void Sign( const HCL_Vector_d & x,
                         const HCL_Vector_d & y );
      /// \TEX{$c_i = {\rm sign}\,(x_i - t)$}
      virtual void Sign( const HCL_Vector_d & x,double t );
      /// \TEX{$c_i = {\rm sign}\,(t - x_i)$}
      virtual void Sign( double t, const HCL_Vector_d & y );
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$c_i$} > \TEX{$t$}, else 0
      virtual void Greater( double t );
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$c_i$} > \TEX{$x_i$}, else 0
      virtual void Greater( const HCL_Vector_d & x );
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$x_i$} > \TEX{$y_i$}, else 0
      virtual void Greater( const HCL_Vector_d & x,
                         const HCL_Vector_d & y );
      /// Mask function: \TEX{$c_i = 1 $} if \TEX{$x_i$} > \TEX{$t$}, else 0
      virtual void Greater( const HCL_Vector_d & x, double t );
      /// Mask function: \TEX{$c_i = 1 $} if \TEX{$t$} > \TEX{$x_i$}, else 0
      virtual void Greater( double t, const HCL_Vector_d & x );
      /** \TEX{$c_i = c_i^{p}$} - arbitrary real power.
	Should only be invoked for nonintegral \TEX{$p$} when
	\TEX{$c_i \ge 0$}, but have not found a good way to enforce 
	this constraint, so rely on runtime */
      virtual void Power( double p );
      /** \TEX{$c_i = x_i^{p}$} - arbitrary real power.
	should only be invoked for nonintegral \TEX{$p$} when
	\TEX{$x_i \ge 0$}, but have not found a good way to enforce 
	this constraint, so rely on runtime */
      virtual void Power( double p, const HCL_Vector_d & x );
      /** \TEX{$c_i=\sqrt{c_i}$} - componentwise square root.
      exits with error condition if any component is negative. */
      virtual void Sqrt();
      /** \TEX{$c_i=\sqrt{x_i}$} - componentwise square root.
      exits with error condition if any component is negative. */
      virtual void Sqrt( const HCL_Vector_d & x );
      /// \TEX{$c_i=\exp{c_i}$} 
      virtual void Exp();
      /// \TEX{$c_i=\exp{x_i}$} 
      virtual void Exp( const HCL_Vector_d & x );
      /** \TEX{$c_i=\log{c_i}$} - componentwise natural log.
	exits with error condition if any component is nonpositive. */
      virtual void Log();
      /** \TEX{$c_i=\log{x_i}$} - componentwise natural log.
	exits with error condition if any component is negative. */
      virtual void Log( const HCL_Vector_d & x );
      //@}

   /// Write prints out the factors in the product.
      virtual ostream & Write(ostream &) const;

};
  
#endif




