/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_QUADFCNL_D_H_
#define HCL_QUADFCNL_D_H_

#include "Std_includes.h"
#include "HCL_Functional_d.h"

/** {\bf HCL\_EvalQuadFunctional\_d} is the evaluation object for
    {\bf HCL\_QuadFunctional\_d}, which implements a functional of the
    form
       $f(x) = \frac{1}{2}(x,Ax)+(b,x)+\alpha$.
    Here $A$ is assumed to be a self-adjoint linear operator, $b$ is a
    fixed vector, and $\alpha$ is a constant scalar.

    For more information, see {\bf HCL\_EvaluateFunctional\_d} and
    {\bf HCL\_QuadFunctional\_d}.
 */
class HCL_EvalQuadFunctional_d: public HCL_EvaluateFunctional_d
{
   private:
      HCL_Vector_d * input;
      HCL_LinearOp_d * A;
      HCL_Vector_d * b;
      HCL_Vector_d * Ax;
      double alpha;
   public:

   // Default constructor---disabled.
      HCL_EvalQuadFunctional_d();
   /// Copy constructor---disabled.
      HCL_EvalQuadFunctional_d( const HCL_EvalQuadFunctional_d & );
   /** Usual constructor; takes the input vector, a pointer to the
       linear operator, a pointer to the constant vector, and the constant
       scalar. */
      HCL_EvalQuadFunctional_d( const HCL_Vector_d & x,HCL_LinearOp_d * a,
                                      HCL_Vector_d * bb,double aalpha );
   // Destructor.
      virtual ~HCL_EvalQuadFunctional_d();

   /// Domain space access.
      virtual HCL_VectorSpace_d & Domain() const { return A->Domain(); }

   /// Value computes the value of the function.
      virtual double Value() const;
   /// Gradient computes the gradient of the function.
      virtual void Gradient( HCL_Vector_d & g ) const;
   /// Hessian computes the Hessian of the function.
      virtual HCL_LinearOp_d * Hessian() const;

   /// Write prints out some useful information about the object.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_QuadFunctional\_d} implements a quadratic functional defined
    as follows:
              $x \mapsto \frac{1}{2}(x,Ax) + (b,x) + \alpha$.
    The operator $A$ is assumed to be self-adjoint.  The vector $b$ and the
    scalar $\alpha$ are optional; if they are not provided to the constructor,
    they are assumed to be zero. */
class HCL_QuadFunctional_d: public HCL_Functional_d
{
   private:
      HCL_LinearOp_d * A;
      HCL_Vector_d * b;
      double alpha;

   public:  

   // Default constructor---disabled.
      HCL_QuadFunctional_d();
   // Copy constructor---disabled.
      HCL_QuadFunctional_d( const HCL_QuadFunctional_d & );
   /** Usual constructor; takes a pointer to the linear operator, a pointer
       to the constant vector, and the constant scalar. */
      HCL_QuadFunctional_d( HCL_LinearOp_d * a,HCL_Vector_d * bb = NULL,
                            double aalpha=0.0 );
   // Destructor.
      virtual ~HCL_QuadFunctional_d();

   /// Domain space access.
      virtual HCL_VectorSpace_d & Domain() const { return A->Domain(); }

   /** {\bf Evaluate} returns an evaluation object that can compute the
       function value, gradient, and Hessian. */
      HCL_EvaluateFunctional_d * Evaluate( const HCL_Vector_d & x ) const;

   /// Write prints out some useful information about the object.
      virtual ostream & Write( ostream & str ) const;
};

#endif
