/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_RNSPACE_D_H
#define HCL_RNSPACE_D_H

#include "Std_includes.h"
#include "HCL_Arch_d.h"
#include "maxmin_d.h"
#include "lapack.h"
#include "HCL_Vector_d.h"

extern int rnd_seed[];

class HCL_RnVector_d;

/** {\bf HCL\_RnSpace\_d} is a vector space class representing the simplest
vector space, namely, Euclidean n-space.  The corresponding vector class
is {\bf HCL\_Rn\_d}, which stores the components of a vector in a 1D
Fortran-style array.  The space is defined simply by a positive integer,
the dimension of the space. */

class HCL_RnSpace_d: public HCL_VectorSpace_d
{
   protected:
      static char Type;
      int dimension;
   public:

   // Default constructor---disabled.
      HCL_RnSpace_d();
   /// Usual constructor; creates a space of dimension n.
      HCL_RnSpace_d( int d ) { dimension = d; }

   /// Creates a copy of an existing Rn space.
      HCL_RnSpace_d( const HCL_RnSpace_d & x ) { dimension = x.Dim(); }

   // Destructor.
      virtual ~HCL_RnSpace_d(){}

   /// {\bf Dim} returns the dimension of the space.
      int Dim() const { return dimension; }

   /** Comparison; returns 1 if the input space is the same as this space;
       0 otherwise. */
      virtual int operator == ( const HCL_VectorSpace_d & xx ) const;
   /** Comparison; returns 1 if the input space differs from this space;
       0 otherwise. */
      virtual int operator != ( const HCL_VectorSpace_d & x ) const;

   /** {\bf GetType} returns the address of the static field {\bf Type}.
       This provides a primitive form of runtime type information. */
      virtual void * GetType() const { return (void*)&Type; }

   /** {\bf Member} is a virtual constructor; it creates a new vector that
       belongs to this vector space. */
      HCL_Vector_d * Member() const;

   /** {\bf Replicate} creates a new vector space identical to this space.
       This method is a virtual copy constructor for this class. */
      virtual HCL_VectorSpace_d * Replicate() const;

   /** {\bf Write} prints a description of this class. */
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_RnVector\_d} is the simplest vector class.  The components of
a vector are stored in a 1D Fortran-style array.  Operations on the vectors
are implemented using explicit loops. */

class HCL_RnVector_d: public HCL_Vector_d
{
   protected:
      HCL_RnSpace_d * sp;
      double * data;
      char * savename;
      int newdata;

   public:

   // Copy constructor---disabled.
      HCL_RnVector_d();
   /// Copy constructor
      HCL_RnVector_d( const HCL_RnVector_d & x );
   /// Construct a vector belonging to the {\bf HCL\_RnSpace\_d} s.
      HCL_RnVector_d( HCL_RnSpace_d * s );
   /** Construct a vector belonging to n-space, and optionally initialize with
       the contents of the "data" array by COPYING a pointer to double -
       note that in this case the destructor will NOT DELETE the data 
       pointer, as it presumably belongs to some other object.*/
      HCL_RnVector_d( int n, double *data = NULL );
   /** Construct a ``new'' vector belonging to the {\bf HCL\_RnSpace\_d} s,
       and optionally specify an ASCII file to which the data will be saved.
       The format for the file is an integer specifying the dimension,
       followed by whitespace or newline-separated values. */
      HCL_RnVector_d( const HCL_RnSpace_d & s, char * fname = NULL );
   /** Construct a ``old'' vector by reading the contents of the 
       specified ASCII file. The format for the file is an integer 
       specifying the dimension, followed by whitespace or newline-
       separated values. If the save flag is set the contents are
       saved back to the file on destruction - this is the default
       behaviour. */
      HCL_RnVector_d( char * fname, int save = 1 );

   // Destructor.
      virtual ~HCL_RnVector_d();

   /// {\bf Dim} returns the dimension of the vector space.
      virtual int Dim() const { return sp->Dim(); }

   /** {\bf Space} returns a reference to the vector space to which this
       vector belongs. */
      virtual HCL_VectorSpace_d & Space() const { return *sp;}


   /** The () operator is a function for accessing the components of the
       vector. This version allows the value to be changed. */
      double & operator()( int i ) { return data[i-1]; }
   /** The () operator is a function for accessing the components of the
       vector.  This version allows access only to the value (it cannot be
       changed). */
      double operator()( int i ) const { return data[i-1]; }

   /** {\bf Data} returns a pointer to an array containing the component
       values of the vector.  This is the actual storage array in the vector,
       so changing it will change the components of the vector.  This pointer
       belongs to the {\bf HCL\_RnVector\_d} object, and may not be deleted. */
      virtual double * Data() { return data; }

   /**@name Vector operations (z is *this, which is C++-ese for the instance
       of the class invoking the method). */
   //@{
   /// z <-- x
      virtual void Copy( const HCL_Vector_d & x );
   /// z <-- -z
      virtual void Neg();
   /// z <-- a*z
      virtual void Mul( const double & a );
   /// z <-- a*x
      virtual void Mul( const double & a,const HCL_Vector_d & x );
   /// z <-- z + x
      virtual void Add( const HCL_Vector_d & x );
   /// z <-- x + y
      virtual void Add( const HCL_Vector_d & x,
                        const HCL_Vector_d & y );
   /// Inner product of z with x.
      virtual double Inner( const HCL_Vector_d & x ) const;

   /// z <-- z - x
      virtual void Sub( const HCL_Vector_d & x );
   /// z <-- x - y
      virtual void Sub( const HCL_Vector_d & x,
                        const HCL_Vector_d & y );
   /// z <-- a*z + x
      virtual void ScaleAdd( const double  & a,
                             const HCL_Vector_d & x );
   /// z <-- z + a*x
      virtual void AddScale( const double  & a,
                             const HCL_Vector_d & x );
   /// z <-- x + a*y
      virtual void AddScale( const double  & a,
                             const HCL_Vector_d & x,
                             const HCL_Vector_d & y );
   /// Norm of z.
      double Norm() const;
   /// Norm squared of z.
      double Norm2() const;
   //@}

   /**@name Array operations - ``Matlab'' methods. In describing these,
      \TEX{$c_i$} denotes the ith component of the current array, i.e. the
      data array of the HCL\_Vector object */
   //@{
   /** \TEX{$c_i = c_i + a$} - overload of Add with scalar argument
         {\em a la} Matlab. */
      virtual void Add( double a );
   /** \TEX{$c_i = x_i + a$} - overload of Add with scalar argument
       {\em a la} Matlab. */
      virtual void Add( const HCL_Vector_d & x, double a );
   /// \TEX{$c_i = a,\,\,i=1,n$}
      virtual void Fill( double a );
   /// \TEX{$c_i = 0,\,\,i=1,n$}
      virtual void Zero();
   /** Returns a vector with (pseudo-)random components.  The components
       are uniformly distributed in \TEX{$(-1,1)$}. */
      virtual void Random();

   /// \TEX{$c_i = c_i*x_i$} - matrix multiply by diag(x)
      virtual void DiagScale( const HCL_Vector_d & x );
   /// \TEX{$c_i = y_i*x_i$}
      virtual void DiagScale( const HCL_Vector_d & x,
                              const HCL_Vector_d & y );
   /** \TEX{$c_i = c_i/{\rm maxmod}(x_i,tol)$} - matrix division by diag(x). 
       Safeguarded against overflow - if tol is less
       that the smallest safe divisor sfmin as defined by LAPACK::[SD]LAMCH,
       then tol is replaced by sfmin - thus the default is tol=0.0, which
       effectively means tol=sfmin. */
      virtual void DiagRecipScale( const HCL_Vector_d & x,
				   double tol=0.0);
   /** \TEX{$c_i = y_i/{\rm maxmod}(x_i,tol)$} - matrix division by diag(x). 
       Safeguarded against overflow - if tol is less
       that the smallest safe divisor sfmin as defined by LAPACK::[SD]LAMCH,
       then tol is replaced by sfmin - thus the default is tol=0.0, which
       effectively means tol=sfmin. */
      virtual void DiagRecipScale( const HCL_Vector_d & x,
				   const HCL_Vector_d & y,
				   double tol=0.0);

   // "MATLAB" methods

      /// returns largest of components or -FLT\_MAX 
      virtual double Max() const;
      /// returns smallest of components or FLT\_MAX
      virtual double Min() const;
      /// returns \TEX{$\sum_{i=1}^n\,c_i$}
      virtual double Sum() const;
      /// \TEX{$c_i = {\rm max}(c_i,x_i) $}
      virtual void Max( const HCL_Vector_d & x );
      /// \TEX{$c_i = {\rm max}(y_i,x_i) $}
      virtual void Max( const HCL_Vector_d & x,
                        const HCL_Vector_d & y );
      /// \TEX{$c_i = {\rm min}(c_i,x_i) $}
      virtual void Min( const HCL_Vector_d & x );
      /// \TEX{$c_i = {\rm min}(y_i,x_i) $}
      virtual void Min( const HCL_Vector_d & x,
                        const HCL_Vector_d & y );
      /// \TEX{$c_i = |c_i|$}
      virtual void Abs();
      /// \TEX{$c_i = |x_i|$}
      virtual void Abs( const HCL_Vector_d & x );
     /** \TEX{$c_i = {\rm sign}\,(c_i)$}. Sign returns 1 for positive
         arguments, -1 for negative arguments, and 0 for zero arguments. */
      virtual void Sign();
      /// \TEX{$c_i = {\rm sign}\,(c_i - t)$}
      virtual void Sign( double t );
      /// \TEX{$c_i = {\rm sign}\,(c_i-x_i)$}
      virtual void Sign( const HCL_Vector_d & x );
      /// \TEX{$c_i = {\rm sign}\,(x_i-y_i)$}
      virtual void Sign( const HCL_Vector_d & x,
                         const HCL_Vector_d & y );
      /// \TEX{$c_i = {\rm sign}\,(x_i - t)$}
      virtual void Sign( const HCL_Vector_d & y,double t );
      /// \TEX{$c_i = {\rm sign}\,(t - x_i)$}
      virtual void Sign( double t,const HCL_Vector_d & y );
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$c_i$} > \TEX{$t$}, else 0
      virtual void Greater( double t );
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$c_i$} > \TEX{$x_i$}, else 0
      virtual void Greater( const HCL_Vector_d & x );
      /// Mask function: \TEX{$c_i = 1$} if \TEX{$x_i$} > \TEX{$y_i$}, else 0
      virtual void Greater( const HCL_Vector_d & x,
                         const HCL_Vector_d & y );
      /// Mask function: \TEX{$c_i = 1 $} if \TEX{$x_i$} > \TEX{$t$}, else 0
      virtual void Greater( const HCL_Vector_d & y,double t );
      /// Mask function: \TEX{$c_i = 1 $} if \TEX{$t$} > \TEX{$x_i$}, else 0
      virtual void Greater( double,const HCL_Vector_d & y );
      /** \TEX{$c_i = c_i^{p}$} - arbitrary real power.
          Should only be invoked for nonintegral \TEX{$p$} when
          \TEX{$c_i \ge 0$}, but have not found a good way to enforce 
          this constraint, so rely on runtime. */
      virtual void Power( double );
      /** \TEX{$c_i = x_i^{p}$} - arbitrary real power.
          should only be invoked for nonintegral \TEX{$p$} when
          \TEX{$x_i \ge 0$}, but have not found a good way to enforce 
          this constraint, so rely on runtime */
      virtual void Power( double p, const HCL_Vector_d & x );
      /** \TEX{$c_i=\sqrt{c_i}$} - componentwise square root.
          exits with error condition if any component is negative. */
      virtual void Sqrt();
      /** \TEX{$c_i=\sqrt{x_i}$} - componentwise square root.
      exits with error condition if any component is negative. */
      virtual void Sqrt( const HCL_Vector_d & x );
      /// \TEX{$c_i=\exp{c_i}$} 
      virtual void Exp();
      /// \TEX{$c_i=\exp{x_i}$} 
      virtual void Exp( const HCL_Vector_d & x );
      /** \TEX{$c_i=\log{c_i}$} - componentwise natural log.
	exits with error condition if any component is nonpositive. */
      virtual void Log();
      /** \TEX{$c_i=\log{x_i}$} - componentwise natural log.
	exits with error condition if any component is nonpositive. */
      virtual void Log( const HCL_Vector_d & x );

   //@}

   /** Save to a file.  The file format is an integer specifying the 
       dimension, followed by the data values, one per line. */
      void Save( char * fname );

   /** {\bf Write} prints the dimension of the vector, followed by
       the first 10 components of the vector (or all of the componenents
       if the dimension is less than 10).  Note that this method is intended
       for routine debugging, i.e. so that the vector can be examined when
       an error condition occurs.  To save all of the vector to a file, use
       the {\bf Save} method. */
      virtual ostream & Write( ostream & str ) const;
};

#endif

