/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_SCALELINEAROP_D_H
#define HCL_SCALELINEAROP_D_H

#include "HCL_LinearOp_d.h"
#include "HCL_Vector_d.h"
#include "Table.h"
#include "HCL_Arch_d.h"

/** {\bf HCL\_ScaleLinearOp\_d} implementing the linear operator
\TEX{\[
x\mapsto ax,
\]}
where a is a scalar.  In addition to the {\bf HCL\_LinearOp\_d} methods,
this class has a {\bf Parameters} method, which returns a reference to
a {\bf Table} object.  The table contains the single parameter "Scale",
which is the scalar \TEX{$a$}.  This scalar can be accessed using
\begin{verbatim}
L.Parameters().GetValue( "Scale",a );
\end{verbatim}
and modified using
\begin{verbatim}
L.Parameters().PutValue( "Scale",a );
\end{verbatim}
*/
class HCL_ScaleLinearOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * sp;
      Table * ParamTable;
   public:

   // Default constructor---disabled.
      HCL_ScaleLinearOp_d();
   // Copy constructor---disabled.
      HCL_ScaleLinearOp_d( const HCL_ScaleLinearOp_d & L );
   /// Usual constructor; needs the domain and the scalar.
      HCL_ScaleLinearOp_d( HCL_VectorSpace_d * SP,double mu );
   // Destructor
      virtual ~HCL_ScaleLinearOp_d();

   /** Domain space access. Returns a reference to the domain of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access. Returns a reference to the range of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;
   /** Access to the Parameter table. The scalar defining the linear operator
       can be accessed through Parameters().GetValue( "Scale",val ) or
       changed by Parameters().PutValue( "Scale",val ). */
      Table & Parameters();

   /// {\bf Image} computes y = a*x.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} calls {\bf Image} (the operator is self-adjoint).
      virtual void AdjImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & y ) const;
   /// {\bf NormalImage} computes $y = a^2x$.
      virtual void NormalImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & y ) const;
   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b z$}.  */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b z$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b z$}.  */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b y$}. */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b y$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b y$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /// Write prints out the domain and scale.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_DiagScaleLinearOp\_d} is a class implementing the linear operator
mapping x to a*x, where x and a are vectors and 
the multiplication is done componentwise (in "Matlab notation", this
operation is a.*x).
*/
class HCL_DiagScaleLinearOp_d: public HCL_LinearOp_d
{
   private:
      HCL_Vector_d * diag;
   public:

   // Default constructor---disabled.
      HCL_DiagScaleLinearOp_d();
   // Copy constructor---disabled.
      HCL_DiagScaleLinearOp_d( const HCL_DiagScaleLinearOp_d & L );
   /** Usual constructor; just needs the scaling vector (which also defines
       the domain).  This version makes an independent copy of the scaling
       vector. */
      HCL_DiagScaleLinearOp_d( const HCL_Vector_d & d );
   /** Usual constructor; just needs the scaling vector (which also defines
       the domain).  This version copies a pointer to the scaling
       vector. */
      HCL_DiagScaleLinearOp_d( HCL_Vector_d * d );
   // Destructor.
      virtual ~HCL_DiagScaleLinearOp_d();

   /** Domain space access. Returns a reference to the domain of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access. Returns a reference to the range of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;

   /// {\bf Image} computes y = a.*x.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} calls {\bf Image} (the operator is self-adjoint).
      virtual void AdjImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & y ) const;
   /// {\bf NormalImage} computes y = a.*a.*x.
      virtual void NormalImage( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf InvImage} computes the action of the inverse on y, giving x.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /// Write prints out the domain and scale.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_ZeroLinearOp\_d} is a class implementing the zero linear
   operator mapping one vector space to another. */
class HCL_ZeroLinearOp_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * dom;
      HCL_VectorSpace_d * range;
   public:

   // Default constructor---disabled.
      HCL_ZeroLinearOp_d();
   // Copy constructor---disabled.
      HCL_ZeroLinearOp_d( const HCL_ZeroLinearOp_d & L );
   /// Usual constructor; needs the domain and range
      HCL_ZeroLinearOp_d( HCL_VectorSpace_d * d,HCL_VectorSpace_d * r );
   // Destructor
      virtual ~HCL_ZeroLinearOp_d();

   /** Domain space access. Returns a reference to the domain of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access. Returns a reference to the range of
       the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;

   /// {\bf Image} computes $y = 0$.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// {\bf AdjImage} computes $x = 0$.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;
   /// {\bf NormalImage} computes $y = 0$.
      virtual void NormalImage( const HCL_Vector_d & x,
                                   HCL_Vector_d & y ) const;
   /** {\bf InvImage} returns an error message, since the zero operator
       is obviously not invertible. */
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} returns an error message, since the zero operator
       is obviously not invertible. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /** {\tt ImageAdd} implements \TEX{$y = a A x + b z$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b z$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b z$}.  */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b z$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b z$}.  */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
                             const HCL_Vector_d & z,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt ImageAdd} implements \TEX{$y = a A x + b y$}. */
       virtual void ImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt AdjImageAdd} implements \TEX{$y = a A^* x + b y$}. */
       virtual void AdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt NormalImageAdd} implements \TEX{$y = a A^*A x + b y$}. */
       virtual void NormalImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvImageAdd} implements \TEX{$y = a A^{-1} x + b y$}. */
       virtual void InvImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;
   /** {\tt InvAdjImageAdd} implements \TEX{$y = a A^{-*} x + b y$}. */
       virtual void InvAdjImageAdd(const HCL_Vector_d & x,
			     HCL_Vector_d & y,
                             double a=1.0,
                             double b=1.0) const;

   /// Write prints out the domain and range
      virtual ostream & Write( ostream & str ) const;
};

#endif
