/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_SMOOTHL1FCNL_H_
#define HCL_SMOOTHL1FCNL_H_

#include "HCL_Functional_d.h"
#include "HCL_CompLinearOp_d.h"
#include "HCL_Adjoint_d.h"
#include "HCL_ScaleLinearOp_d.h"

/** {\bf HCL\_EvalSmoothL1Fcnl\_d} is the evaluation object for
{\bf HCL\_SmoothL1Fcnl\_d}, which implements a smoothed version of the
$\ell^1$-norm of $Ax-b$, where $A$ is a linear operator and $b$ is a
data vector:

    $f(x)=\sum_{i=1}^m\sqrt{(Ax-b)_i^2+\epsilon^2}$.

For more information, see the documentation for the base class
{\bf HCL\_EvaluateFunctional\_d} and for the class {\bf HCL\_SmoothL1Fcnl\_d}.
*/
class HCL_EvalSmoothL1Fcnl_d: public HCL_EvaluateFunctional_d
{
   private:
     HCL_VectorSpace_d * dom;
     HCL_Vector_d * fr;
     HCL_Vector_d * dfr;
     HCL_Vector_d * d2fr;
     HCL_Vector_d * r;
     HCL_Vector_d * t1;
     HCL_LinearOp_d * A;
     double eps;
   public:

   // Default constructor---disabled.
      HCL_EvalSmoothL1Fcnl_d();
   // Copy constructor---disabled.
      HCL_EvalSmoothL1Fcnl_d( const HCL_EvalSmoothL1Fcnl_d & x );
   /** Usual constructor; takes (a reference to) the input vector,
       a pointer to the linear operator, a pointer to the data vector,
       and the smoothing parameter. */
      HCL_EvalSmoothL1Fcnl_d( const HCL_Vector_d & x,HCL_LinearOp_d * a,
                                    HCL_Vector_d * d,double e );
   // Destructor.
      virtual ~HCL_EvalSmoothL1Fcnl_d();

   /// Domain space access.
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }

   /// {\bf Value} computes the value of the functional.
      virtual double Value() const;

   /// {\bf Gradient} computes the gradient of the functional.
      virtual void Gradient(HCL_Vector_d & g ) const;

   /// {\bf Hessian} computes the Hessian of the functional.
      virtual HCL_LinearOp_d * Hessian() const;

   /// {\bf Write} outputs a description of the evaluation object.
      virtual ostream & Write( ostream & str ) const;
};

/** {\bf HCL\_SmoothL1Fcnl\_d} implements a smoothed version of the
$\ell^1$-norm of $Ax-b$, where $A$ is a linear operator and $b$ is a
data vector:

    $f(x)=\sum_{i=1}^m\sqrt{(Ax-b)_i^2+\epsilon^2}$.

For a description of the methods of {\bf HCL\_SmoothL1Fcnl\_d}, see the
documentation for the base class {\bf HCL\_Functional\_d}.
*/

class HCL_SmoothL1Fcnl_d: public HCL_Functional_d
{
   private:
      HCL_LinearOp_d * A;
      HCL_Vector_d * data;
      double eps;

   public:
  
   // Default constructor---disabled.
      HCL_SmoothL1Fcnl_d();
   // Copy constructor---disabled.
      HCL_SmoothL1Fcnl_d( const HCL_SmoothL1Fcnl_d & x );
   /** Usual constructor; takes a pointer to the linear operator,
       a pointer to the data vector, and the smoothing parameter epsilon. */
      HCL_SmoothL1Fcnl_d(HCL_LinearOp_d * a,HCL_Vector_d * d,double e );

   // destructor
      virtual ~HCL_SmoothL1Fcnl_d();

   /// Domain space access.
      virtual HCL_VectorSpace_d & Domain() const { return A->Domain(); }

   /** {\bf Evaluate} creates an "evaluation" object, which knows how to
       compute the function value, gradient, and Hessian at the given x. */
      virtual HCL_EvaluateFunctional_d * 
                  Evaluate( const HCL_Vector_d & x ) const;

   /// {\bf Write} outputs a description of the functional.
      virtual ostream & Write( ostream & str ) const;
};

#endif
