#ifndef HCL_SUBPRODUCTPROJECTOR_D_H_
#define HCL_SUBPRODUCTPROJECTOR_D_H_

#include "HCL_ProductVector_d.h"
#include "HCL_GenericProductVector_d.h"
#include "HCL_LinearOp_d.h"

/** The class {\bf HCL\_SubProductProjector\_d} implements a linear operator
which maps a product space
\TEX{\[
X = X_1\times X_2\times\cdots X_n
\]}
into the "sub-product"
\TEX{\[
Y = X_{i_1}\times X_{i_2}\times\cdots X_{i_k}.
\]}
The operator is defined to be simple projection, i.e.
\TEX{\[
x \mapsto (x_{i_1},x_{i_2},\ldots,x_{i_k}).
\]}
The integers \TEX{$i_1,i_2,\ldots,i_k$} are defined by an iArray object.
The special case of
\TEX{\[
x \mapsto x_j
\]}
is handled by a special constructor which takes a single integer in place
of an iArray object.

Thus, an object of this type is constructed via
\begin{verbatim}
HCL_SubProductProjector_d L( &dom,&indices );
\end{verbatim}
where {\tt dom} is an instance of {\bf HCL\_ProductSpace\_d} and indices is
an instance of {\bf iArray}, or
\begin{verbatim}
HCL_SubProductProjector_d L( &dom,j );
\end{verbatim}
where {\tt j} is a positive integer.  In the first case, the range of the
operator is an instance of {\bf HCL\_GenericProductSpace\_d}, while in the
second case, it is an instance of {\bf HCL\_VectorSpace\_d}.
*/

class HCL_SubProductProjector_d: public HCL_LinearOp_d
{
   private:
      HCL_ProductSpace_d * dom;
      HCL_GenericProductSpace_d * rng;
      HCL_VectorSpace_d * rng1;
      iArray * ind;
      int j;
   public:

   // Default constructor---disabled.
      HCL_SubProductProjector_d();
   // Copy constructor---disabled.
      HCL_SubProductProjector_d( const HCL_SubProductProjector_d & );
   /** This constructor creates a projector whose range is an instance
       of {\bf HCL\_GenericProductSpace\_d}.  The arguments are the domain
       and an {\bf iArray} object specifying the factors of the range. */
      HCL_SubProductProjector_d( HCL_ProductSpace_d * d,iArray * indices );
   /** This constructor creates a projector whose range is an instance
       of {\bf HCL\_VectorSpace\_d}.  The arguments are the domain
       and an integer specifying the range. */
      HCL_SubProductProjector_d( HCL_ProductSpace_d * d,int jj );
   // Destructor.
      ~HCL_SubProductProjector_d();

  /** Domain space access. Returns a reference to the domain of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const { return *dom; }
  /** Range space access. Returns a reference to the range of
      the operator, which is an {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const
      {
         if( rng ) return *rng;
         else return *rng1;
      }

  /// {\bf Image} compute the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
  /** {\bf AdjImage} computes the action of the adjoint on y, giving x.
      This method is implemented in the base class to return an 
       error message. */
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;
   /** Debugging information. */
      virtual ostream & Write( ostream & str ) const;
};

#endif


