/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef HCL_TRUSTREGION_SS_H_
#define HCL_TRUSTREGION_SS_H_

#include "HCL_TrustRegionSolvers_d.h"
#include "HCL_ItEigSolvers_d.h"
#include "HCL_GenericProductVector_d.h"
#include "HCL_LinearSolver_d.h"

/** The class {\bf HCL\_TRSSOp\_d} defines the bordered operator used
by {\bf HCL\_TrustRegion\_SS\_d}.  This is a linear operator, and hence
an instance of {\bf HCL\_LinearOp\_d}. */

class HCL_TRSSOp_d: public HCL_LinearOp_d
{
   private:
      HCL_GenericProductSpace_d * dom;
      HCL_Vector_d * g;
      HCL_LinearOp_d * A;
      HCL_LinearOp_d * C;
      HCL_LinearSolver_d * lsolver;
      Table * ParamTable;
   public:

   // Default constructor---disabled.
      HCL_TRSSOp_d();
   // Copy constructor---disabled.
      HCL_TRSSOp_d( const HCL_TRSSOp_d & );
   /** Constructor; takes the operator A, the vector g, the scalar alpha,
       and, optionally, a preconditioner and corresponding linear solver. */
      HCL_TRSSOp_d( const HCL_LinearOp_d * AA, const HCL_Vector_d * gg,
                    double a,HCL_LinearOp_d * CC=NULL,
                    HCL_LinearSolver_d * ls=NULL );
   // Destructor.
      ~HCL_TRSSOp_d();

   /// Access to parameters, specifically, the scalar alpha.
      Table & Parameters() const;

   /** Domain space access. Returns a reference to the domain of
       the operator, which is a {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Domain() const;
   /** Range space access. Returns a reference to the range of
       the operator, which is a {\bf HCL\_VectorSpace\_d}. */
      virtual HCL_VectorSpace_d & Range() const;
   /// Image compute the action of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// AdjImage computes the action of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;

   /// {\bf InvImage} returns an error message and stops the program.
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} returns an error message and stops the program. */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /// Debugging information.
      virtual ostream & Write( ostream & str ) const;
};


/* Developer's notes:

   Algorithm does not set return value "Multiplier" (this is important!).
   Use of scaled inner product must be checked; it probably doesn't work.

*/

/** {\bf NOTE: This class is experimental and has not been carefully
debugged.  It is known to fail on occasions when it should not.}

The class {\bf HCL\_TrustRegion\_SS\_d} is an implementation of the
Santosa-Sorensen algorithm for solving the trust region subproblem.
This algorithm involves transforming the subproblem into a family of
parameterized eigenvalues problems, which are solving using the k-step
Arnoldi method.  The parameter in the eigenvalue problem is adjusted to
produce a solution to the trust region subproblem, along with its Lagrange
multiplier, simultaneously. For more information about the algorithm, see

    "A new matrix-free algorithm for the large-scale trust-region
     subproblem" by Sandra A. Santos and Danny C. Sorensen
     Technical report TR95-20, Department of Computational and Applied
     Mathematics, Rice University, Houston, TX 77251-1892.

In the initial implementation, the trust-region constraint is defined
by the unscaled Euclidean norm.
*/

class HCL_TrustRegion_SS_d: public HCL_TrustRegionSolver_d
{
   private:

      Table * ParamTable;

   // Input parameters (default value)
   /**@name Input parameters */
   //@{
   /// Initial trust region radius (1.0).
      double Delta;
   /** Tolerance for the step length.  The iteration is halted when
       the step is within this relative tolerance of the boundary (1.0e-2). */
      double Tol;
   /// Maximum number of iterations (20).
      int MaxItn;
   /** Display Flag (0). This flag controls how much detail about the
       progress of the iteration is sent to the screen.  Possible values
       are 0 (no display), 1 (summary statement at the end),
       2 (summary statement every {\bf DispFreq} iterations),
       3 (output of algorithmic parameters at every iterations---used for
       debugging purposes). */
      int DispFlag;
   /// Display precision; controls how many digits are sent to the screen (6).
      int DispPrecision;
   /** Display frequency; controls how often details of iteration are
       displayed on screen (MaxItn/5). */
      int DispFreq;
   /// minimum step length, that is, minimum Delta (1.0e-6).
      double MinStep;
   /// maximum step length, that is, maximum Delta (1.0e6).
      double MaxStep;
   /// Tolerance for the hard case (1.0e-2). */
      double HardCaseTol;
   //@}

   // The following is for documentation purposes only.
#ifdef NOT_DEFINED
   /**@name Required output parameters */
   //@{
   /** Termination code.  Possible values are
          BOUNDARY_SOLUTION 101
          INTERIOR_SOLUTION 102
          FAILED            103 */
      int TermCode;
   /// Lagrange multiplier.
      double Multiplier;
   /// Reduction in the quadratic model.
      double Reduction;
   /// Length of step taken.
      double StepLength;
   //@}
#endif

   // Some temporary objects

      HCL_IRArnoldi_d * eigsolver;
      HCL_TRSSOp_d * Balpha;
      HCL_RnVector_d * gammaVec;
      HCL_Vector_d * xtemp;
      HCL_GenericProductVector_d * y;

      HCL_LinearOp_d * C;
      HCL_LinearSolver_d * lsolver;

   // Some other local variables

      int OldDispPrec;
      double gnorm;           // Norm of g (used in detecting hard case)
      double alpha;           // The parameter in the bordered matrix
      double gamma;           // The first component of the eigenvector
      double lam1_prev;       // The smallest eigenvalue of Balpha from the
                              // previous iteration
      double lam1,lam2;       // The two smallest eigenvalues of Balpha
      double phi,phi_prev;    // The current and previous values of phi
      double dphi,dphi_prev;  // The current and previous derivatives of phi
      double alpha_l,alpha_u; // Lower and upper bounds on alpha (safeguard)
      double lam_l,lam_u;     // Lower and upper bounds on lambda (safeguard)
      double xnorm;           // Norm of approximate solution
      double xnorm_prev;      // Norm of approximate solution (previous itn)
      double ZeroTol;         // Tolerance for deciding when lam1 >= 0.0
                              // (use lam1 >= -ZeroTol)

   // Compute (scaled) vector norm

      double GetxNorm( const HCL_Vector_d & ) const;

   // Apply CG in case an interior solution is detected.

      int GetInteriorSolution( const HCL_LinearOp_d & A,
                               const HCL_Vector_d & g,HCL_Vector_d & p );

   // Initialize (allocate temporary vectors, etc.)

      void Initialize( HCL_VectorSpace_d & dom );

   // De-allocate temporaries, etc.

      void Clean();
      void DumpProblem( const HCL_LinearOp_d & A,const HCL_Vector_d & g,
                                                       HCL_Vector_d & x );
   public:

   // Copy constructor---disabled.
      HCL_TrustRegion_SS_d( const HCL_TrustRegion_SS_d & x );
   /** Usual constructor; optionally takes the name of a file from
       which to read algorithmic parameters.  If such a file is
       provided, it must be a plain text file with entries of the
       form

                 ParameterName = value

       or

                 TRSolver::ParameterName = value

       The second form allows the same file to contain parameters for
       more than one algorithm to be given in the same file.  For example,
       one can entries such as

                 UMin::DispFlag = 2
                 TRSolver::DispFlag = 1

       which set the display flags for the minimization algorithm and
       the trust region solver to different values. */
      HCL_TrustRegion_SS_d( char * fname = NULL );

   // Destructor.
      virtual ~HCL_TrustRegion_SS_d();

      /** The method {\bf Parameters} returns the parameter table of
          the algorithm.  Algorithmic parameters can be accessed
          via {\bf Parameters().GetValue( "name",val )} and changed
          via {\bf Parameters().PutValue( "name",val )}. */
      virtual Table & Parameters() const;

      /** {\bf SetScaling} defines a new inner product in terms of
          a symmetric, positive definite operator S: <x,y> = (x,Sy). */
      virtual void SetScaling( HCL_LinearOp_d * S,
                               HCL_LinearSolver_d * lsolver=NULL );
      /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();

      /** The method {\bf Solve} attempts to compute an approximate
          solution of the trust region subproblem defined by
          the inputs of the method.  Note that the trust region
          radius {\bf Delta} must defined through the {\bf Parameter}
          method. */
      virtual void Solve( const HCL_LinearOp_d & A,const HCL_Vector_d & g,
                                HCL_Vector_d & p );

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
