/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_UMINALTERNATION_D_H_
#define HCL_UMINALTERNATION_D_H_

#include "HCL_UMinProductDomain_d.h"
#include "HCL_UMin_d.h"

/** {\bf HCL\_UMinAlternation\_d} is the simplest concrete class
derived from {\bf HCL\_UMinProductDomain\_d}.  This class attempts to
find an unconstrained minimum of a function defined on a product space
by repeated minimizing the function over one component.  This algorithm
is sometimes referred to as coordinate-search.

The algorithm(s) used to minimize the functional in the component directions
must be specified by the user; they can be any instances of {\bf HCL\_UMin\_d}.
Note the algorithm need not be the same for each component. */

class HCL_UMinAlternation_d: public HCL_UMinProductDomain_d
{
   private:
      int numinit,M;
      HCL_UMin_d ** umin;
      Table * ParamTable;

   /**@name Input Parameters
      In a parameter file, the names may be prepended by "UMinAlternation::"
      or "UMinProductDomain::". */
   //@{
      /// (100) Maximum number of iterations.
      int MaxItn;
      /** (1.0) Typical value of the functional near the solution.  Setting
          this makes the gradient stopping tolerance more meaningful. */
      double Typf;
      /** (1.0) Typical norm of the unknown vector x near the solution.
          Setting this makes the gradient stopping tolerance more
          meaningful. */
      double TypxNorm;
      /** (1.0e-2) Gradient tolerance.  The algorithm attempts to locate
          a point where the relative gradient norm (i.e. the norm of the
          gradient scaled by the size of the vector x and by f(x))
          is less than this value. */
      double GradTol;
      /** (1.0e-20) Minimum allowable step.  If the algorithm takes a
          (relative) step less than this value in norm, it halts and reports
          "Possible convergence". */
      double MinStep;
      /** (1.0e+20) Maximum allowable step.  If the algorithm takes too many
          consecutive steps of length MaxStep, it assumes that the iterates
          are diverging and reports "Possible divergence". */
      double MaxStep;
      /** (5) Maximum number of steps of length MaxStep allowed before
          the algorithm decides that the iterates are diverging. */
      int CscMaxLimit;
      /** (0) Display level.  This determines how much information should be
          displayed during the execution of the algorithm.  Possible values
          are: 0 - No output; 1 - Function value and gradient norm after
          final iteration; 2 - Function value and gradient norm after every
          iteration. */
      int DispFlag;
      /** (0) Dump level.  This determines how much information should be
          sent to the dump file during the execution of the algorithm.
          Possible values are: 0 - No output; 1 - Function value and
          gradient norm after final iteration; 2 - Function value and
          gradient norm after every iteration. */
      int DumpFlag;
      /** (HCL\_UMinAlternation.DumpFile) Dump file name. */
      char DumpFile[81];
      /** (6) Display precision---the number of digits sent to the screen. */
      int DispPrecision;
      /** (6) Dump precision---the number of digits sent to the file. */
      int DumpPrecision;
      /** (0) If nonzero, the iterates are sent to a file using the
          {\bf Write} method from the vector class. */
      int TraceSteps;
      /** (HCL\_UMinAlternation.StepFile) File name for recording iterates. */
      char StepFile[81];
   //@}

      // Local variables

      int Itn,CscMax;
      double fx;
      double gnorm;
      double xnorm;
      double stepnorm;
      int OldDispPrec;
      int TermCode;
      ofstream * dump;
      ofstream * Steps;
      HCL_EvalFunctionalProductDomain_d * eval;
      HCL_LinearOp_d * S;
      HCL_LinearSolver_d * lsolver;
      HCL_ProductVector_d * tmpvec;
      HCL_ProductVector_d * step;
      HCL_ProductVector_d * xnext;

      void Initialize( HCL_VectorSpace_d & dom );
      int StoppingTest( HCL_ProductVector_d & x,
                        HCL_ProductVector_d * x1=NULL );
      void Display( const HCL_Functional_d & f,const HCL_Vector_d & x );
      void Clean();

   public:

   // Default constructor---disabled.
      HCL_UMinAlternation_d();
   // Copy constructor---disabled.
      HCL_UMinAlternation_d( const HCL_UMinAlternation_d & );
   /** Usual constructor.  The arguments are the number of factors in
       the domain of the functional to be minimized, and an optional
       parameter file.  If a parameter file is provided, it must be a plain
       text file with entries of the form

                 ParameterName = value

       or

                 UMinProductDomain::ParameterName = value

       or

                 UMinAlternation::ParameterName = value

       The second and third forms allow the same file to contain parameters for
       more than one algorithm to be given in the same file.  For example,
       one can entries such as

                 UMinAlternation::DispFlag = 2
                 TRSolver::DispFlag = 1

       which set the display flags for the minimization algorithm and
       the trust region solver to different values. */
      HCL_UMinAlternation_d( int n,char * fname=NULL );
   // Destructor.
      ~HCL_UMinAlternation_d();

   /** Access to parameter table. Algorithmic parameters can be accessed
       with Parameters().GetValue( "NAME",val ) or changed with
       Parameters().PutValue( "NAME",val ). */
      virtual Table & Parameters() const;

   /** LastEval returns a reference to the functional's evaluation object
       at the most recent point.  Thus, after the minimization has completed,
       the calling routine has access to the last function value and
       gradient. */
      virtual HCL_EvalFunctionalProductDomain_d & LastEval() const;

   /** {\bf SetNext} sets the next minimization algorithm to be set.
       The program creating the {\bf HCL\_UMinAlternation\_d} object
       is responsible for defining a minimization algorithm for each
       coordinate direction using this method. */
      void SetNext( HCL_UMin_d * um );

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy). */
      virtual void SetScaling( HCL_LinearOp_d * S,
                               HCL_LinearSolver_d * lsolver );
   /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();
   /** Minimize attempts to find a local minimizer of the functional f,
       using x as a starting guess. The return value is the termination
       code, as given above. */
      virtual int Minimize( HCL_FunctionalProductDomain_d & f,
                            HCL_ProductVector_d & x );
    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
