/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_NONLINNLCG_D_H
#define HCL_NONLINNLCG_D_H

#include "HCL_Functional_d.h"
#include "HCL_UMin_d.h"
#include "HCL_LineSearch_d.h"
#include "Table.h"

/** {\bf HCL\_UMinNLCG\_d} implements the nonlinear conjugate gradient
algorithm (Polak-Ribiere form) for unconstrained minimization.
For background on the algorithm, see, for example, "Practical Methods
of Optimization" (2nd edition) by R. Fletcher, Wiley, 1987.

This algorithm depends on a fairly accurate line search.
The line search is represented by an abstract base class,
{\bf HCL\_LineSearch\_d}; this allows the algorithm to be executed
with different choices of the line search algorithm.
The default is {\bf HCL\_LineSearch\_Fl\_d}, which implements
Fletcher's line search (see the above reference).

The purpose of the algorithm is to take a functional with gradient
and a starting guess in the domain of the functional, and to produce
a local minimizer via a descent algorithm.

The primary methods of this class are:
\begin{itemize}
\item {\bf Minimize}, which performs the minimization;
\item {\bf Parameters},which provides a means of accessing or changing
      algorithmic parameters which control the iteration.  These
      parameters are typically set initially by passing a file name to
      the constructor; this file contains parameter values.
      Any parameters not specified for the constructor are assigned
      default values.  See the documentation of the parameters for more
      information.
\item {\bf LastEval}, which provides access to the "evaluation" object
      at the final point.
\end{itemize}

Use of this minimizer typically involves the following steps:
\begin{itemize}
\item Construct functional with gradient, i.e. an object of a class
      derived from {\bf HCL\_Functional\_d};
\item Construct a starting vector in the domain of the functional;
\item Construct a line search object, i.e. an object of a class
      derived from {\bf HCL\_LineSearch\_d} (this step is optional,
      since there is a default line search);
\item Construct the minimizer by passing the constructor the line
      search object and a parameter file;
\item Call the {\bf Minimize} method;
\item Manipulate the value and/or gradient at the approximate minimizer
      through the {\bf LastEval} method.
\end{itemize}

Here's an example:
\begin{verbatim}
   MyFcnl f;  // MyFcnl must be derived from HCL\_Functional\_d
   MyVector x( "x0" ); // Assume MyVector has a constructor which
                       //reads from a file
   HCL_LineSearch_MT_d lsearch( "lsearch.dat" );  // More'-Thuente
                                                  // line search
   HCL_UMinNLCG_d umin( "umin.dat",&lsearch );    // Parameters read
                                                  // from "umin.dat"
   umin.Minimize( f,x );  // Search for local minimum
   cout << "Final value of f: " << umin.LastEval().ValueRef() << endl;
   cout << "Final gradient norm: "
        << umin.LastEval().GradientRef().Norm() << endl;
\end{verbatim}
(If the display flag is turned on, the final function value and gradient
norm will be displayed anyway, so the last two lines are not necessary.
But they illustrate that the last function value and gradient are stored
in case they are needed.) */
class HCL_UMinNLCG_d: public HCL_UMin_d
{
   private:

      Table * ParamTable;

   /**@name Input parameters
      In a parameter file, the names may be prepended by "UMinNLCG::"
      or simply "UMin::". */
   //@{
      ///(100) maximum number of iterations
      int MaxItn;
      /** (1.0) Typical value of the functional near the solution.  Setting
          this makes the gradient stopping tolerance more meaningful. */
      double Typf;
      /** (1.0) Typical norm of the unknown vector x near the solution.
          Setting this makes the gradient stopping tolerance more
          meaningful. */
      double TypxNorm;
      /** (1.0e-2) Gradient tolerance.  The algorithm attempts to locate
          a point where the relative gradient norm (i.e. the norm of the
          gradient scaled by the size of the vector x and by f(x))
          is less than this value. */
      double GradTol;
      /** (1e-20) Minimum allowable step.  If the algorithm takes a (relative)
          step less than this value in norm, it halts and reports
          "Possible convergence". */
      double MinStep;
      /** (1e+20) Maximum allowable step.  If the algorithm takes too many
          consecutive steps of length MaxStep, it assumes that the iterates
          are diverging and reports "Possible divergence". */
      double MaxStep;
      /** (5) Maximum number of steps of length MaxStep allowed before
          the algorithm decides that the iterates are diverging. */
      int CscMaxLimit;
      int DispFlag;
      /** (0) Dump level.  This determines how much information should be
          sent to the dump file during the execution of the algorithm.
          Possible values are: 0 - No output; 1 - Function value and
          gradient norm after final iteration; 2 - Function value and
          gradient norm after every iteration. */
      int DumpFlag;
      /** (HCL\_UMin\_lbfgs.DumpFile) Dump file name. */
      char DumpFile[81];
      /** (6) Display precision---the number of digits sent to the screen. */
      int DispPrecision;
      /** (6) Dump precision---the number of digits sent to the file. */
      int DumpPrecision;
      /** (0) If nonzero, the iterates are sent to a file using the
          {\bf Write} method from the vector class. */
      int TraceSteps;
      /** (HCL\_UMinNLCG.StepFile) File name for recording iterates. */
      char StepFile[81];
   //@}

   /**@name Output parameters */
   //@{
   /// Termination code.
      int TermCode;
   //@}


      // local variables

      int Itn,CscMax;
      double f_xcur,f_xnext,gnorm2,gnorm,gnorm_next,snorm,xnorm;

      ofstream * dump;
      ofstream * Steps;
      int fevalFlag,OldDispPrec;
      int StDescDir;

      HCL_LineSearch_d *lsearch;

      HCL_LinearOp_d * S;
      HCL_LinearSolver_d * lsolver;

      HCL_EvaluateFunctional_d * eval;

      // temporary vectors, etc.

      HCL_Functional_d * fptr;
      HCL_Vector_d *xcur,*xnext,*tmpvec,*tmpvec1,*dir,*step;

   /// Read parameters from parameter table and allocate temporary vectors.

      void Initialize( HCL_VectorSpace_d & dom );

   /// Write output parameters out to parameter table and deallocate vectors.

      void Clean();

   // print current status

      void Display( const HCL_Functional_d & f,const HCL_Vector_d & x );

   // check stopping criterion

      int StoppingTest( const HCL_Vector_d & x,
                        HCL_EvaluateFunctional_d & eval,
                        int ls_result, int MaxTkn);

   public:
   /**@name return codes (TermCode) */
   //@{
      enum {
             /// succeeded
             Succeeded = 0,
             /// method converged away from a minimum
             Converged = 1,
             /// method exceded MaxItn iterations
             Too_Many_Iterations = 2,
             /// method appeared to diverge
             Diverged = 3,
             /// misc. failure
             Failed = 4,
             /// for internal use
             Continue = 5
      } TermCodes;
   //@}

   /** Usual constructor.
       @param linesearch Linesearch class (optional). If no line search
              is specified, then Fletcher's line search
              ({\bf HCL\_LineSearch\_Fl\_d}) will be used.
       @param fname Parameter file name or NULL.  If given, this file
              contains algorithmic parameters (such as stopping tolerances),
              a full list of which is given elsewhere in the documentation
              of this class.  For example, to set the maximum number of
              iterations to 100, the file should contain a line of the
              form UMinNLCG::MaxItn = 100 or, alternately, UMin::MaxItn = 100.
              If the line seach is not provided, then this file will also
              be passed to the line search constructor, and so can contain
              entries such as LineSearch::DispFlag = 1.
    */
      HCL_UMinNLCG_d( HCL_LineSearch_d * linesearch = NULL,
                      char *fname = NULL );
      
   // Copy constructor---disabled.
      HCL_UMinNLCG_d( const HCL_UMinNLCG_d & um );

   // Destructor.

      virtual ~HCL_UMinNLCG_d();

   /** Access to parameter table. Algorithmic parameters can be accessed
       with Parameters().GetValue( "NAME",val ) or changed with
       Parameters().PutValue( "NAME",val ). */
      Table & Parameters() const { return *ParamTable; }

   /// Returns a reference to the last evaluation object.
      virtual HCL_EvaluateFunctional_d & LastEval() const;

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy). */
      virtual void SetScaling( HCL_LinearOp_d * S,
                               HCL_LinearSolver_d * lsolver=NULL );
   /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();

   /** Conjugate gradient minimizer.  This algorithm uses the 
       Polak and Ribiere formula to update the search direction.
       (See Fletcher (1987).
       @param fptr function
       @param x0 starting point, on successful completion, this will hold
              the computed minimum
    */
      int Minimize( HCL_Functional_d & f,HCL_Vector_d & x0);

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
