/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_UMINTR_D_H
#define HCL_UMINTR_D_H

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include "HCL_Functional_d.h"
#include "HCL_TrustRegionSolvers_d.h"
#include "HCL_UMin_d.h"

/** The class {\bf HCL\_UMinTR\_d} implements a generic algorithm for solving
unconstrained minimization problems using a trust region method.
A {\bf HCL\_TrustRegionSolver\_d}, which implements an algorithm for
solving the trust region subproblem, is required to construct the
minimizer.  If the user does not provide a trust region solver,
the {\bf HCL\_TRCG\_d}, which implements the Steihaug-Toint algorithm,
is used by default.  Note that the Steihaug-Toint algorithm is intended
for large-scale problems.

The three important methods of this class are:
\begin{itemize}
\item {\bf Minimize} This method takes a {\bf HCL\_Functional\_d}
      and a starting point, and attempts to locate a local minimizer
      of the functional.  The return value of the method is the
      termination code, which indicates why the algorithm terminated.
      See the documentation for the termination codes for details.
\item {\bf Parameters} This method returns a reference to the
      parameter table, and allows the programmer to access or
      change the parameters that control the iteration.
      See the documentation on {\bf input parameters} for details
      about algorithmic parameters.
\item {\bf LastEval} This method returns a reference to the
      evaluation object at the best point found by the algorithm,
      thus giving the calling routine access to the function value,
      gradient, and Hessian at the final point.  It is an error to call
      {\bf LastEval} before {\bf Minimize}.
\end{itemize}
*/
class HCL_UMinTR_d: public HCL_UMin_d
{
   private:

      HCL_TrustRegionSolver_d * trsolver;
      HCL_EvaluateFunctional_d * feval;
      int fevalFlag;
      HCL_LinearOp_d * C;
      HCL_LinearSolver_d * lsolver;

      Table * ParamTable;

      // parameters found in ParamTable

   /**@name Input Parameters
      In a parameter file, the names may be prepended by "UMinTR::"
      or simply "UMin::". */
   //@{
      ///(100) maximum number of iterations
      int MaxItn;
      /** (1.0) Typical value of the functional near the solution.  Setting
          this makes the gradient stopping tolerance more meaningful. */
      double Typf;
      /** (1.0) Typical norm of the unknown vector x near the solution.
          Setting this makes the gradient stopping tolerance more
          meaningful. */
      double TypxNorm;
      /** (1.0e-2) Gradient tolerance.  The algorithm attempts to locate
          a point where the relative gradient norm (i.e. the norm of the
          gradient scaled by the size of the vector x and by f(x))
          is less than this value. */
      double GradTol;
      /** (1e-20) Minimum allowable step.  If the algorithm takes a (relative)
          step less than this value in norm, it halts and reports
          "Possible convergence". */
      double MinStep;
      /** (1e+20) Maximum allowable step.  If the algorithm takes too many
          consecutive steps of length MaxStep, it assumes that the iterates
          are diverging and reports "Possible divergence". */
      double MaxStep;
      /** (5) Maximum number of steps of length MaxStep allowed before
          the algorithm decides that the iterates are diverging. */
      int CscMaxLimit;
      /** (0) Display level.  This determines how much information should be
          displayed during the execution of the algorithm.  Possible values
          are: 0 - No output; 1 - Function value and gradient norm after
          final iteration; 2 - Function value and gradient norm after every
          iteration. */
      int DispFlag;
      /** (0) Dump level.  This determines how much information should be
          sent to the dump file during the execution of the algorithm.
          Possible values are: 0 - No output; 1 - Function value and
          gradient norm after final iteration; 2 - Function value and
          gradient norm after every iteration. */
      int DumpFlag;
      /** (HCL\_UMin\_lbfgs.DumpFile) Dump file name. */
      char DumpFile[81];
      /** (6) Display precision---the number of digits sent to the screen. */
      int DispPrecision;
      /** (6) Dump precision---the number of digits sent to the dump file. */
      int DumpPrecision;
      /** (0) If nonzero, the iterates are sent to a file using the
          {\bf Write} method from the vector class. */
      int TraceSteps;
      /** (HCL\_UMinTR.StepFile) File name for recording iterates. */
      char StepFile[81];
      /** (1e-1) Tolerance for reducing the trust region radius.  If the ratio
          of the actual to the predicted reduction falls below LowerTol,
          the radius is cut in half. */
      double LowerTol;
      /** (9e-1) Tolerance for increasing the trust region radius.  If the
          ratio of the actual to the predicted reduction is greater than
          UpperTol, the radius is doubled. */
      double UpperTol;
      /** (1.0) Initial trust region radius. */
      double Delta;
   //@}

      // Local variables
      int OldDispPrec;
      int TermCode;
      double StepLength;
      ofstream * dump;
      ofstream * Steps;
      int Itn,CscMax;
      double fcur,fnext,gnorm,xnorm;
      double ared,pred,ratio;
      
      // Some temporary vectors

      HCL_Vector_d * tmpvec;
      HCL_Vector_d * step;
      HCL_Vector_d * xnext;

      // Initialize (allocate temporary vectors, etc.)

      void Initialize( HCL_VectorSpace_d & dom );

      // De-allocate temporaries, etc.

      void Clean();

      // Display intermediate results

      void Display( const HCL_Vector_d & x );
      void Display0( const HCL_Functional_d & f,const HCL_Vector_d & x );

      // Stopping test

      int StoppingTest( const HCL_Vector_d & x );

   public:

   // Default constructor---disabled.
      HCL_UMinTR_d();
   // Copy constructor---disabled.
      HCL_UMinTR_d( const HCL_UMinTR_d & xx );
   /** Usual constructor.
       @param trs TrustRegionSolver class (optional). If no trust region solver
              is specified, then the Steihaug-Toint algorithm
              ({\bf HCL\_TRCG\_d}) will be used.
       @param fname Parameter file name or NULL.  If given, this file
              contains algorithmic parameters (such as stopping tolerances),
              a full list of which is given elsewhere in the documentation
              of this class.  For example, to set the maximum number of
              iterations to 100, the file should contain a line of the
              form UMinTR::MaxItn = 100 or, alternately, UMin::MaxItn = 100.
              If the trust region solver is not provided, then this file will
              also be passed to the {\bf HCL\_TRCG\_d} constructor, and so
              can contain entries such as TRSolver::DispFlag = 1.
    */
      HCL_UMinTR_d( HCL_TrustRegionSolver_d * trs=NULL,char * fname=NULL );
   // Destructor.
      virtual ~HCL_UMinTR_d();

   /** Access to parameter table. Algorithmic parameters can be accessed
       with Parameters().GetValue( "NAME",val ) or changed with
       Parameters().PutValue( "NAME",val ). */
      Table & Parameters() const { return *ParamTable; }

   /** LastEval returns a reference to the functional's evaluation object
       at the most recent point.  Thus, after the minimization has completed,
       the calling routine has access to the last function value, gradient,
       and Hessian. */
      virtual HCL_EvaluateFunctional_d & LastEval() const;

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy). */
      virtual void SetScaling( HCL_LinearOp_d * S,
                               HCL_LinearSolver_d * lsolver=NULL );
   /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling();

   /** Minimize attempts to find a local minimizer of the functional f,
       using x as a starting guess. */
      int Minimize( HCL_Functional_d & f,HCL_Vector_d & x );

    /** Prints description of the object. */
      virtual ostream & Write( ostream & str ) const;
};

#endif
