/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_UMIN_D_H
#define HCL_UMIN_D_H

#include "HCL_Functional_d.h"
#include "HCL_LinearSolver_d.h"
#include "Table.h"

/** {\bf HCL\_UMin\_d} is the abstract base class for unconstrained
minimizers. In the context of HCL, this means that an object of this type
will minimize an {\bf HCL\_Functional\_d}.

All objects derived from {\bf HCL\_UMin\_d} have five methods:
\begin{itemize}
\item {\bf Minimize} This method takes a {\bf HCL\_Functional\_d}
      and a starting point, and attempts to locate a local minimizer
      of the functional.  The return value of the method is the
      termination code, which indicates why the algorithm terminated.
      See the documentation for the termination codes for details.
\item {\bf SetScaling} This method takes a {\bf HCL\_LinearOp\_d}
      and a {\bf HCL\_LinearSolver\_d}, or alternatively, just a
      {\bf HCL\_LinearOp\_d} (in the second case, the operator must
      have {\bf InvImage} implemented), and uses the operator to define
      a new inner product.  The operator must be symmetric and positive
      definite; the new inner product is defined in terms of the
      default inner product of the vector space: <x,y> = (x,Sy).
      The optimization is carried out using the new inner product.
\item {\bf UnSetScaling} This method resets the inner product to the
      default.
\item {\bf Parameters} This method returns a reference to the
      parameter table, and allows the programmer to access or
      change the parameters that control the iteration.
\item {\bf LastEval} This method returns a reference to the
      evaluation object at the best point found by the algorithm,
      thus giving the calling routine access to the function value
      and gradient at the final point.  It is an error to call
      {\bf LastEval} before {\bf Minimize}.
\end{itemize}
*/
class HCL_UMin_d: public HCL_Base
{
   public:

   // Default constructor.
      HCL_UMin_d(){}
   // Copy constructor.
      HCL_UMin_d( const HCL_UMin_d & um ){}
   // Destructor
      virtual ~HCL_UMin_d(){}

   /**@name Return values for the method {\bf Minimize}. */
      enum
      {
   //@{
         /** Possible local minimizer. The algorithm has succeeded in
             reducing the norm of the gradient below the tolerance
             specified by the user. */
         PossibleMinimizer = 1,
         /** Possible convergence. The algorithm took a step shorter than
             the minimum step length specified by the user. */
         PossibleConvergence = 2,
         /** Line search failed. The line search was unable to reduce the
             function value. */
         LineSearchFailed = 3,
         /** Possible divergence.  The algorithm took 5 consecutive
             steps exceeding the maximum step length specified by the
             user. */
         PossibleDivergence = 4,
         /** Iteration limit reached.  The algorithm was unable to meet
             the stopping criteria in the number of iterations allowed. */
         IterationLimit = 5,
         /** Possible inaccurate gradient calculation.  The negative gradient
             was not a descent direction.  Note that the minimizer does
             perform any other check of the accuracy of the gradient
             calculation.  The {\bf HCL\_Functional\_d} class has a
             method {\bf CheckGrad} to compare the analytic gradient
             against finite differences, but the implementor of the
             functional is responsible for calling this method explicitly. */
         InaccurateGradient = 6
   //@}
      } ErrorCodes;

   /** Access to parameter table. Algorithmic parameters can be accessed
       with Parameters().GetValue( "NAME",val ) or changed with
       Parameters().PutValue( "NAME",val ). */
      virtual Table & Parameters() const = 0;

   /** LastEval returns a reference to the functional's evaluation object
       at the most recent point.  Thus, after the minimization has completed,
       the calling routine has access to the last function value and
       gradient. */
      virtual HCL_EvaluateFunctional_d & LastEval() const = 0;

   /** {\bf SetScaling} defines a new inner product in terms of
       a symmetric, positive definite operator S: <x,y> = (x,Sy). */
      virtual void SetScaling( HCL_LinearOp_d * S,
                               HCL_LinearSolver_d * lsolver=NULL ) = 0;
   /// {\bf UnSetScaling} returns the inner product to the default.
      virtual void UnSetScaling() = 0;
   /** Minimize attempts to find a local minimizer of the functional f,
       using x as a starting guess. The return value is the termination
       code, as given above. */
      virtual int Minimize( HCL_Functional_d & f,HCL_Vector_d & x ) = 0;
};

#endif
