/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#ifndef HCL_LMBFGSOP1_D_H
#define HCL_LMBFGSOP1_D_H

#include "HCL_ScaleLinearOp_d.h"
#include "HCL_LinearSolver_d.h"

/** {\bf HCL\_lmbfgsOp1\_d} implements the limited memory BFGS approximation to
the inverse Hessian of a twice-differentiable function.  This approximation
uses local changes to the gradient to gradually build up an estimate of
the Hessian for use in nonlinear optimization problems.

For details of the algorithm, see the paper

   "Updating Quasi-Newton Matrices with Limited Storage" by Jorge Nocedal,
    Math. of Computation, Vol. 35, no. 151, p.p. 773--782.

Note that the operator is an approximation to the {\it inverse} Hessian,
so the the {\bf Image} method computes the quasi-Newton step.

The BFGS approximation is based on successive rank-one perturbations to
an initial approximation; these rank-one perturbations can be represented
as outer products.  This class allows the user to provide a symmetric
positive definite operator to define an alternate inner product, which
then changes the definition of the outer product.  In the context of an
optimization problem, this is equivalent to implementing the algorithm
in the alternate inner product.  */

class HCL_lmbfgsOp1_d: public HCL_LinearOp_d
{
   private:
      HCL_VectorSpace_d * sp;
      HCL_ScaleLinearOp_d * H0;
      HCL_LinearOp_d * P;
      HCL_LinearSolver_d * lsolver;
      HCL_Vector_d * temp_vec;
      HCL_Vector_d * temp_vec1;
      int CurNum,MaxNum,CurAllocated;
      HCL_Vector_d ** Y;
      HCL_Vector_d ** S;
      double * rho;
   public:

   // Default constructor---disabled.
      HCL_lmbfgsOp1_d();
   // Copy constructor---disabled.
      HCL_lmbfgsOp1_d( const HCL_lmbfgsOp1_d & H );
   /** Usual constructor. Needs a multiple of the identity to use for the
       initial inverse Hessian approximation, the maximum number of
       updates, and, optionally, an operator to change the inner product. */
      HCL_lmbfgsOp1_d( HCL_ScaleLinearOp_d * h0,int maxnum,
             HCL_LinearOp_d * Pin = NULL,HCL_LinearSolver_d * ls = NULL );
   // Destructor.
      virtual ~HCL_lmbfgsOp1_d();

   /// Domain space access.
      HCL_VectorSpace_d & Domain() const;
   /// Range space access.
      HCL_VectorSpace_d & Range() const;

   /// Access to the scale, which is dynamically changed by the operator.
      double GetScale() const;
   /** Update requires the current and next $x$ and the
       current and next gradient. */
      void Update( const HCL_Vector_d & x,const HCL_Vector_d & xnext,
                   const HCL_Vector_d & g,const HCL_Vector_d & gnext );
   /// Reset sets the operator to the initial inverse Hessian approximation.
      void Reset();

   /// Image computes the image of the operator on x, giving y.
      virtual void Image( const HCL_Vector_d & x,
                                HCL_Vector_d & y ) const;
   /// AdjImage computes the image of the adjoint on y, giving x.
      virtual void AdjImage( const HCL_Vector_d & y,
                                   HCL_Vector_d & x ) const;
   /** {\bf InvImage} computes the action of the inverse on y, giving x
       (not implemented). */
      virtual void InvImage(const HCL_Vector_d & y,
                                  HCL_Vector_d & x) const;
   /** {\bf InvAdjImage} computes the action of the inverse adjoint on x,
       giving y (not implemented). */
      virtual void InvAdjImage(const HCL_Vector_d & x,
                                     HCL_Vector_d & y) const;

   /// Write prints out some useful information about the object.
      virtual ostream & Write( ostream & str ) const;
};

#endif
