#ifndef _ABSTBL_D
#define _ABSTBL_D

#include "HCL_Base.h"
#include "Std_includes.h"

/** Interface for associative array, or hash table, data
    access. Access to data is through {\tt Get} and {\tt Put}
    methods. These methods take two arguments: the first is a
    character string specifying the {\bf key}, to which the table
    associates the {\bf value} returned in the second table. The {\bf
    value} may be any primitive numerical type, or a string. {\tt Get}
    methods, which are {\bf const}, extract the value, implicitly
    type-converting it if necessary. The {\tt Put} methods replace the
    values.  */

class AbstractTable : public HCL_Base {

public:

  AbstractTable(){};

  AbstractTable(const AbstractTable & T) {
    cerr<<"Error: no copy constructor for class AbstractTable"<<endl;
  }

  /// destructor
  virtual ~AbstractTable(){}

  /** Replication: creates and returns pointer to independent copy */
  virtual AbstractTable * Replicate() const = 0;

  /** Return the size of the table = number of keys */
  virtual int Size() const = 0;

   /** Return the ith key */
  virtual char * Key( int & i) = 0;

  /**@name Read access by key. */
  //@{
  /// extract value as int
  virtual int GetValue(const char *key, int &value) const = 0;
  /// extract value as float
  virtual int GetValue(const char *key, float &value) const = 0;
  /// extract value as double
  virtual int GetValue(const char *key, double &value) const = 0;
  /// extract value as string
  virtual int GetValue(const char *key, char *value, int length = 0) const = 0;
  //@}

  /**@name Write access by key. */
  //@{
  /// write value as int
  virtual int PutValue(const char *key, int value) = 0;
  /// write value as float
  virtual int PutValue(const char *key, float value) = 0;
  /// write value as double
  virtual int PutValue(const char *key, double value) = 0;
  /// write value as character string
  virtual int PutValue(const char *key, const char *value) = 0;
  //@}
   
  /**@name Read access by indexed key. 
    Provides easy access to keys
    with integral indices, such as "n1", which would be accessed with
    key="n", ind=1. */
  //@{
  /// extract value as int
  virtual int GetArrayValue(const char *key, int ind, int &value) const = 0;
  /// read value as float
  virtual int GetArrayValue(const char *key, int ind, float &value) const = 0;
  /// read value as double
  virtual int GetArrayValue(const char *key, int ind, double &value) const = 0;
  /// read value into character string
  virtual int GetArrayValue(const char *key, int ind, char *value, int length = 0) const = 0;
  //@}

  /**@name Read access by indexed key. 
    Provides easy access to keys
    with integral indices, such as "n1", which would be accessed with
    key="n", ind=1. */
  //@{
  /// write int value
  virtual int PutArrayValue(const char *key, int ind, int value) = 0;
  /// write float value
  virtual int PutArrayValue(const char *key, int ind, float value) = 0;
  /// write double value
  virtual int PutArrayValue(const char *key, int ind, double value) = 0;
  /// write character string value
  virtual int PutArrayValue(const char *key, int ind, const char *value) = 0;
  //@}

   /** Print the contents of the table 
     as "key=value" pairs to output
     stream s. ( s<<table; is equivalent ) 
     @param s Output stream */
  /** Print contents of tableto output stream.*/
  virtual ostream &Write(ostream&s) const = 0;

};

#endif

