/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "HCL_CMinAL_d.h"

#define POSSIBLE_MINIMIZER 1
#define POSSIBLE_CONVERGENCE 2
#define ITERATION_LIMIT 3

HCL_CMinAL_d::HCL_CMinAL_d()
{
   cerr << "Error: HCL_CMinAL_d default constructor called"
        << endl;
   exit(1);
}

HCL_CMinAL_d::HCL_CMinAL_d( const HCL_CMinAL_d & x )
{
   cerr << "Error: HCL_CMinAL_d copy constructor called"
        << endl;
   exit(1);
}

HCL_CMinAL_d::HCL_CMinAL_d( HCL_UMin_d * um,char * fname )
{
   umin = um;um->IncCount();
   if( fname )
      ParamTable = new Table( fname,"CMinAL CMin",20 );
   else
      ParamTable = new Table;
   Feval = NULL;
}

HCL_CMinAL_d::~HCL_CMinAL_d()
{
   HCL_delete( Feval );
   HCL_delete( ParamTable );
   HCL_delete( umin );
}

void HCL_CMinAL_d::Display()
{
   if( DispFlag > 0 )
   {
      cout << endl << "===========HCL_CMinAL_d===========" << endl;
      cout << "iteration: " << Itn << endl;
      cout << " f(x) = " << fx << endl;
      cout << "||H(x)|| = " << Hxnorm << endl;
      cout << "||grad Lag(x)|| = " << gradlnorm << endl;
      cout << "Penalty parameter = " << PenParam << endl;
      cout << endl << "======================================" << endl;
   }
}

int HCL_CMinAL_d::StoppingTest( HCL_EvalALFcnl_d & eval )
{
   if( Itn >= MaxItn )
      return ITERATION_LIMIT;

   fx = eval.fValue();
   eval.ConstraintViolation( *Hx );Hxnorm = Hx->Norm();
   eval.GradientLag( *gradl );gradlnorm = gradl->Norm();
   if( gradlnorm < GradTol && Hxnorm < ConstrTol )
      return POSSIBLE_MINIMIZER;

   return 0;
}

void HCL_CMinAL_d::Initialize( const HCL_Op_d & H )
{
   lambda = H.Range().Member();
   lambda->Zero();
   Hx = H.Range().Member();
   gradl = H.Domain().Member();
   if( Parameters().GetValue( "InitPenParam",InitPenParam ) )
      InitPenParam = 0.0;
   PenParam = InitPenParam;
   if( Parameters().GetValue( "PenParamIncFac",PenParamIncFac ) )
      PenParamIncFac = 4.0;
   if( Parameters().GetValue( "IncPenaltyThreshold",IncPenaltyThreshold ) )
      IncPenaltyThreshold = 0.25;
   if( Parameters().GetValue( "MaxPenParam",MaxPenParam ) )
      MaxPenParam = 1.0e8;
   if( Parameters().GetValue( "UMinGradTolDecFac",UMinGradTolDecFac ) )
      UMinGradTolDecFac = 0.5;
   if( Parameters().GetValue( "GradDecThreshold",GradDecThreshold ) )
      GradDecThreshold = 0.9;
   if( Parameters().GetValue( "UMinMinGradTol",UMinMinGradTol ) )
      UMinMinGradTol = 1.0e-8;
   if( Parameters().GetValue( "MaxItn",MaxItn ) )
      MaxItn = 10;
   if( Parameters().GetValue( "GradTol",GradTol ) )
      GradTol = 1.0e-4;
   if( Parameters().GetValue( "ConstrTol",ConstrTol ) )
      ConstrTol = 1.0e-4;
   if( Parameters().GetValue( "SpecFlag",SpecFlag ) )
      SpecFlag = 0;
   if( Parameters().GetValue( "DispFlag",DispFlag ) )
      DispFlag = 0;
   if( Parameters().GetValue( "DumpFlag",DumpFlag ) )
      DumpFlag = 0;
   if( Parameters().GetValue( "DispPrecision",DispPrecision ) )
      DispPrecision = 6;
   OldDispPrec = cout.precision( DispPrecision );
   Itn = 0;
   TermCode = 0;
}

void HCL_CMinAL_d::Clean()
{
   cout.precision( OldDispPrec );
   HCL_delete( gradl );
   HCL_delete( Hx );
   HCL_delete( lambda );
}

int HCL_CMinAL_d::Minimize( HCL_Functional_d & f,HCL_Op_d & H,
                                HCL_Vector_d & x )
{
   // Check for consistency of the various vector spaces

   if( f.Domain() != H.Domain() )
   {
      cerr << "Error in HCL_CMinAL_d::Minimize: domains of objective "
           << endl << "function and constraint function do not agree"
           << endl;
      cerr << "Domain of objective function is" << endl;
      f.Domain().Write( cerr );
      cerr << "Domain of constraint function is" << endl;
      H.Domain().Write( cerr );
      exit(1);
   }

   if( f.Domain() != x.Space() )
   {
      cerr << "Error in HCL_CMinAL_d::Minimize: starting vector is not "
           << endl << "in domain of objective function"
           << endl;
      cerr << "Domain of objective function is" << endl;
      f.Domain().Write( cerr );
      cerr << "Space of starting vector is" << endl;
      x.Space().Write( cerr );
      exit(1);
   }

   // Initialize algorithmic parameters, temporary vectors, etc.

   Initialize( H );

   // Create the augmented Lagrangian functional

   HCL_ALFcnl_d F( &f,&H,lambda,PenParam );

   // Get the initial value of the augmented Lagrangian

   Feval = (HCL_EvalALFcnl_d*)F.Evaluate( x );

   // SpecFlag (special flag) is for the developer's use; it causes
   // the code to check the gradient of the augmented Lagrangian function.

   if( SpecFlag == 1 )
      F.CheckGrad( x,Feval->GradientRef(),cout );

   // Check to see if we have started at a point that approximately
   // satisfies the first-order conditions.
   TermCode = StoppingTest( *Feval );

   // If the parameter InitPenParameter was not set by the user,
   // choose an initial penalty parameter equal to the ratio of the objective
   // function value to the norm of the constraint violation.

   if( PenParam==0.0 )
   {
      HCL_FPArch_d arch;
      if( Hxnorm <= 10*arch.eps || abs(fx) <= sqrt(arch.eps) )
         PenParam = 1.0;
      else
         PenParam = abs(fx)/Hxnorm;
      F.Parameters().PutValue( "PenParam",PenParam );
   }

   // Display progress (if DispFlag is set)

   Display();

   while( !TermCode )
   {
      // Delete the old evaluation object

      HCL_delete( Feval );

      // Minimize the augmented Lagrangian function with the current
      // value of the penalty parameter and Lagrange multiplier.

      umin->Minimize( F,x );
      Itn++;

      // Get the value of the augmented Lagrangian at the approximate
      // minimum.

      Feval = &(HCL_EvalALFcnl_d&)(umin->LastEval());
      Feval->IncCount();

      // Update the multiplier estimate

      Feval->MultiplierUpdate( *lambda );

      // Save the previous norms of the constraint violation and gradient of
      // the Lagrangian,  and check the stopping criteria.

      double oldHxnorm = Hxnorm;
      double oldgradlnorm = gradlnorm;
      TermCode = StoppingTest( *Feval );

      // If the constraint violation has not be reduced by a factor
      // of IncPenaltyThreshold, then increase the penalty parameter
      // by a factor of PenParamIncFac.

      if( !TermCode && Hxnorm > IncPenaltyThreshold*oldHxnorm )
      {
         PenParam *= PenParamIncFac;
         F.Parameters().PutValue( "PenParam",PenParam );
         if( DispFlag > 1 )
            cout << "HCL_CMinAL_d: Insufficient decrease in constraint "
                    "violation; increasing penalty parameter" << endl;
      }

      // If the norm of the gradient of the Lagrangian has not be reduced
      // by a factor of GradDecThreshold, then decrease the gradient tolerance
      // for the UMin object by a factor of UMinGradTolDecFac.

      if( !TermCode && gradlnorm > GradDecThreshold*oldgradlnorm )
      {
         double UMinGradTol;
         umin->Parameters().GetValue( "GradTol",UMinGradTol );
         UMinGradTol *= UMinGradTolDecFac;
         umin->Parameters().PutValue( "GradTol",UMinGradTol );
         if( DispFlag > 1 )
            cout << "HCL_CMinAL_d: Insufficient decrease in gradient of the "
                 << endl << "              Lagrangian; decreasing UMin "
                    "GradTol to " << UMinGradTol << endl;
      }

      // Display progress (if DispFlag is set)

      Display();
   }
   Clean();
   return TermCode;
}

HCL_EvalALFcnl_d & HCL_CMinAL_d::LastEval()
{
   if( Feval == NULL )
   {
      cerr << "Error in HCL_CMinAL_d::LastEval: not set yet" << endl;
      exit(1);
   }
   return *Feval;
}

ostream & HCL_CMinAL_d::Write(ostream & str) const
{
   str << "HCL_CMinAL_d augmented Lagrangian algorithm class" << endl;
   str << "Parameter table:" << endl;
   return ParamTable->Write( str );
}

