/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/
#include "HCL_Basis_d.h"

HCL_Basis_d::HCL_Basis_d()
{
   cerr << "Error: HCL_Basis_d default constructor called" << endl;
   exit(1);
}

HCL_Basis_d::HCL_Basis_d( const HCL_Basis_d & )
{
   cerr << "Error: HCL_Basis_d copy constructor called" << endl;
   exit(1);
}

HCL_Basis_d::HCL_Basis_d( HCL_VectorSpace_d * S,int m )
{
   n = m;
   nmax = n;
   sp = S;S->IncCount();
   vlist = new HCL_Vector_d*[n];
   int i;
   for( i=0;i<n;i++ ) vlist[i] = sp->Member();
}

HCL_Basis_d::~HCL_Basis_d()
{
   int i;
   for( i=0;i<n;i++ )
      HCL_delete( vlist[i] );
   delete[] vlist;
   HCL_delete( sp );
}

HCL_VectorSpace_d & HCL_Basis_d::Space() const
{
   return *sp;
}

int HCL_Basis_d::Dim()
{
   return n;
}

HCL_Vector_d & HCL_Basis_d::operator()( int i )
{
   return *(vlist[i-1]);
}

void HCL_Basis_d::LinearCombination( const HCL_RnVector_d & a,
                                           HCL_Vector_d & x ) const
{
   if( a.Dim() != n )
   {
      cerr << "Error in HCL_Basis_d::LinearCombination: wrong number of "
              "coefficients" << endl;
      exit(1);
   }
   if( x.Space() != Space() )
   {
      cerr << "Error in HCL_Basis_d::LinearCombination: vector is not in "
              "space" << endl;
      exit(1);
   }

   x.Mul( a(1),*(vlist[0]) );
   int i;
   for( i=2;i<=n;i++ )
      x.AddScale( a(i),*(vlist[i-1]) );
}

int HCL_Basis_d::ModifiedGramSchmidt()
{
   int i,j;
   HCL_FPArch_d arch;

   double mnorm = vlist[0]->Norm();
   if( mnorm <= arch.small )
      return 1;
   vlist[0]->Mul( 1.0/mnorm );

   for( j=1;j<n;j++ )
   {
      double mnorm0 = vlist[j]->Norm();
      if( mnorm0 > arch.small )
      {
         for( i=0;i<j;i++ )
         {
            double dot = vlist[i]->Inner( *(vlist[j]) );
            vlist[j]->AddScale( -dot,*(vlist[i]) );
         }
         mnorm = vlist[j]->Norm();
      }
      if( mnorm0 <= arch.small || mnorm < arch.eps*mnorm0 )
         return j+1;
      vlist[j]->Mul( 1.0/mnorm );
   }
   return 0;
}

int HCL_Basis_d::ModifiedGramSchmidtWithReduction()
{
   int i,j;
   HCL_FPArch_d arch;

   double mnorm = vlist[0]->Norm();
   if( mnorm <= arch.small )
   {
      HCL_delete( vlist[0] );
      for( i=1;i<n;i++ )
         vlist[i-1] = vlist[i];
      n--;
   }
   else
      vlist[0]->Mul( 1.0/mnorm );

   double tol = pow(arch.eps,2.0/3.0);
   j = 1;
   while( j < n )
   {
      double mnorm0 = vlist[j]->Norm();
      if( mnorm0 > arch.small )
      {
         for( i=0;i<j;i++ )
         {
            double dot = vlist[i]->Inner( *(vlist[j]) );
            vlist[j]->AddScale( -dot,*(vlist[i]) );
         }
         mnorm = vlist[j]->Norm();
      }
      if( mnorm0 <= arch.small || mnorm < tol*mnorm0 )
      {
         HCL_delete( vlist[j] );
         for( i=j+1;i<n;i++ )
            vlist[i-1] = vlist[i];
         n--;
      }
      else
      {
         vlist[j]->Mul( 1.0/mnorm );
         j++;
      }
   }
   return 0;
}

void HCL_Basis_d::OrthoCheck() const
{
   int i,j;
   for( i=1;i<=n;i++ )
   {
      for( j=1;j<=n;j++ )
         cout << setprecision(4) << setw(10)
              <<  vlist[i-1]->Inner( *(vlist[j-1]) ) << "  ";
      cout << endl;
   }
}

ostream & HCL_Basis_d::Write( ostream & str ) const
{
   str << "HCL_Basis_d: basis for subspace of dimension " << n << "of" << endl;
   return Space().Write( str );
}

HCL_BasisProductSpace_d::HCL_BasisProductSpace_d()
{
   cerr << "Error: HCL_BasisProductSpace_d default constructor called" << endl;
   exit(1);
}

HCL_BasisProductSpace_d::HCL_BasisProductSpace_d( const HCL_BasisProductSpace_d & )
{
   cerr << "Error: HCL_BasisProductSpace_d copy constructor called" << endl;
   exit(1);
}

HCL_BasisProductSpace_d::HCL_BasisProductSpace_d(
                         HCL_ProductSpace_d * S,int m )
{
   n = m;
   sp = S;S->IncCount();
   vlist = new HCL_ProductVector_d*[n];
   int i;
   for( i=0;i<n;i++ ) vlist[i] = (HCL_ProductVector_d*)(sp->Member());
}

HCL_BasisProductSpace_d::~HCL_BasisProductSpace_d()
{
   int i;
   for( i=0;i<n;i++ )
      HCL_delete( vlist[i] );
   delete[] vlist;
   HCL_delete( sp );
}

HCL_ProductSpace_d & HCL_BasisProductSpace_d::Space() const
{
   return *sp;
}

int HCL_BasisProductSpace_d::Dim()
{
   return n;
}

HCL_ProductVector_d & HCL_BasisProductSpace_d::operator()( int i )
{
   return *(vlist[i-1]);
}

void HCL_BasisProductSpace_d::LinearCombination(
                     const HCL_RnVector_d & a,HCL_ProductVector_d & x ) const
{
   if( a.Dim() != n )
   {
      cerr << "Error in HCL_BasisProductSpace_d::LinearCombination: wrong "
              "number of coefficients" << endl;
      exit(1);
   }
   if( x.Space() != Space() )
   {
      cerr << "Error in HCL_BasisProductSpace_d::LinearCombination: vector "
              "is not in space" << endl;
      exit(1);
   }

   x.Mul( a(1),*(vlist[0]) );
   int i;
   for( i=2;i<=n;i++ )
      x.AddScale( a(i),*(vlist[i-1]) );
}

int HCL_BasisProductSpace_d::ModifiedGramSchmidt()
{
   int i,j;
   HCL_FPArch_d arch;

   double mnorm = vlist[0]->Norm();
   if( mnorm <= arch.small )
      return 1;
   vlist[0]->Mul( 1.0/mnorm );

   for( j=1;j<n;j++ )
   {
      double mnorm0 = vlist[j]->Norm();
      if( mnorm0 > arch.small )
      {
         for( i=0;i<j;i++ )
         {
            double dot = vlist[i]->Inner( *(vlist[j]) );
            vlist[j]->AddScale( -dot,*(vlist[i]) );
         }
         mnorm = vlist[j]->Norm();
      }
      if( mnorm0 <= arch.small || mnorm < arch.eps*mnorm0 )
         return j+1;
      vlist[j]->Mul( 1.0/mnorm );
   }
   return 0;
}

int HCL_BasisProductSpace_d::ModifiedGramSchmidtWithReduction()
{
   int i,j;
   HCL_FPArch_d arch;

   double mnorm = vlist[0]->Norm();
   if( mnorm <= arch.small )
   {
      HCL_delete( vlist[0] );
      for( i=1;i<n;i++ )
         vlist[i-1] = vlist[i];
      n--;
   }
   else
      vlist[0]->Mul( 1.0/mnorm );

   double tol = pow(arch.eps,2.0/3.0);
   j = 1;
   while( j < n )
   {
      double mnorm0 = vlist[j]->Norm();
      if( mnorm0 > arch.small )
      {
         for( i=0;i<j;i++ )
         {
            double dot = vlist[i]->Inner( *(vlist[j]) );
            vlist[j]->AddScale( -dot,*(vlist[i]) );
         }
         mnorm = vlist[j]->Norm();
      }
      if( mnorm0 <= arch.small || mnorm < tol*mnorm0 )
      {
         HCL_delete( vlist[j] );
         for( i=j+1;i<n;i++ )
            vlist[i-1] = vlist[i];
         n--;
      }
      else
      {
         vlist[j]->Mul( 1.0/mnorm );
         j++;
      }
   }
   return 0;
}

void HCL_BasisProductSpace_d::OrthoCheck() const
{
   int i,j;
   for( i=1;i<=n;i++ )
   {
      for( j=1;j<=n;j++ )
         cout << setprecision(4) << setw(10)
              << vlist[i-1]->Inner( *(vlist[j-1]) ) << "  ";
      cout << endl;
   }
}

ostream & HCL_BasisProductSpace_d::Write( ostream & str ) const
{
   str << "HCL_BasisProductSpace_d: basis for subspace of dimension " << n
       << "of" << endl;
   return Space().Write( str );
}

