/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0 - July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef HCL_BASE_H_
#define HCL_BASE_H_

#include "Std_includes.h"

/** {\bf HCL\_Base} is the base class for all HCL classes.
    The purpose of having a common base class is to implement reference
    counting.  {\bf HCL\_Base} contains an integer called ReferenceCount;
    whenever a programmer wants to copy a pointer to a HCL object
    rather than copying the object itself, he must call the method
    IncCount(), which increments ReferenceCount.  The object cannot
    be deleted unless its ReferenceCount is 1; however, {\em in order
    for this scheme to work, the function {\bf HCL\_delete} must be used
    in place of the operator {\bf delete}.}  Note that the programmer
    who increments the reference count is responsible for calling
    {\bf HCL\_delete} for that object (just as a programmer who
    calls {\bf new} is responsible for calling {\bf delete}).

    The most common situation in which one wants to copy a pointer to
    a HCL object is in a constructor.  Suppose an object is constructed
    which depends on a HCL object.  It may not be desirable to make
    a copy of the HCL object (due to time or space overhead).  One can
    just copy the pointer to the object and increment its reference count;
    then the HCL object cannot be deleted until its reference count
    has be decremented by a call to {\bf HCL\_delete}.

    Example:
    \begin{verbatim}
MyObject::MyObject( HCL_Object * obj,... ) // constructor for MyObject
{
   hclobj = obj;   // hclobj is a data member of type HCL_Object*
   obj->IncCount();  // Increment obj's reference count
   // etc.
}

MyObject::~MyObject()  // destructor for MyObject
{
   HCL_delete( hclobj ); // Decrement obj's reference count
   // etc.
}
\end{verbatim} 

    The point of HCL's reference counting is to provide protection against
    an object being deleted by one part of the code while another part of
    the code is still using it.  In the above example, the object pointed
    to by {\bf hclobj} will not disappear while the {\bf MyObject} object
    still points to it. 

    {\bf HCL\_Base} also mandates that every instantiable HCL object have
a {\bf Write} method which sends a description of the object to an output stream. Of course the class implementor will decide what ``description of the object'' means in each instance. The most useful meaning seems to be: whatever is necessary to aid in debugging.

*/

class HCL_Base
{
   private:

   /** The integer ReferenceCount is the number of objects pointing at this
       object. */
      int ReferenceCount;

   public:

   /** Default constructor. This is called whenever a HCL object is
       constructed; it just sets ReferenceCount to 1. */
      HCL_Base();
   // Copy constructor---disabled.
      HCL_Base( const HCL_Base & );
   // Destructor.
      virtual ~HCL_Base();

   /** Increment reference count. This must be called whenever a pointer
       to a HCL object is copied.  There must be a matching call to the
       global function {\bf HCL\_delete}; otherwise a memory leak will
       result. */
      void IncCount() const;
   /** Decrement reference count.  This is called by {\bf HCL\_delete} and
       need never be invoked directly. */
      void DecCount() const;
   /** Access to the reference count.  This is used by {\bf HCL\_delete}. */
      int Count() const;
   /** Description of object.  {\bf Write} should be implemented to
       print a description of the object that will be useful
       for debugging.  If a complete description is lengthy,
       an implementor should consider just printing out a useful
       subset.  This method will likely be used to send information
       to the screen when a error condition is encountered. */
      virtual ostream & Write(ostream &) const = 0;
};

/** To be called in place of the operator {\bf delete} (for HCL objects only).
    Each call to {\bf new} (for a HCL object) and each invocation of
    {\bf IncCount} must be matched by a call to {\bf HCL\_delete}. */
void HCL_delete( HCL_Base * p );

// Redefinition of operator {\bf new} to explicitly call {\it malloc}.
//void * operator new( size_t s );
// Redefinition of operator {\bf delete} to explicitly call {\it free}.
//void operator delete( void * p );

#endif
