/*
                 THE HILBERT CLASS LIBRARY                        
                Release 0.8, September 1997                       
                                                                  
       copyright (C) 1995, 1996, 1997 by Mark S. Gockenbach,      
              Matthew Petro, and William W. Symes                 
                                                                  
The source code contained in this file FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#ifndef TABLE_H
#define TABLE_H

#include "absTable.H"

#define KEY_LENGTH 32      // maximum length of parameter names and of 
			   // prefixes to parameter names

#define MAX_LINE_LENGTH 256  // no input file name shall be longer than this


/** ASCII Parameter Table Object. This class implements the HCL
    associative array (hash table) interface defined in the {\bf
    AbstractTable} base class.

    The principal Table constructor reads the Table's internal data
    from a "config" file. Each line of the config file is either of
    the form "Key = Value" or "Object::Key = Value".  The first form
    initializes the parameter associated to "Key" in the table to the
    value "Value". The second form facilitates the use of one config
    file for several different objects. If the "Object" matches one of
    the object names provided by the user (as argument to the
    constructor, see below), then the parameter is read/written as
    before. Otherwise, the line is ignored.  The type of "Value" can
    be integer, float, double, or string. String values are optionally
    enclosed by double-quotes.

    Because the keys (strings) are private data for this class, it is
    a bit on the profligate side with storage.  */

class Table : public AbstractTable {

private:

   class str_table {
      
   public:
      
//      typedef struct { char key[KEY_LENGTH], *value; } pair;
      struct pair { char key[KEY_LENGTH], *value; };
      
      pair *table;
      int table_size, allocated_size;
   
      int find(const char *key, int &ind) const;
      int insert(const char *key, char *value, int ind);
      
      str_table(int size);
      
      ~str_table();
      
      int put(const char *key, char *value);
      
      char *get(const char *key) const; 
      char *key(int & i);
      
   };
   
   
   str_table prefixes;
   str_table table;
   
public:
   
   /**@name Constructors, destructor */
   //@{

   /** Create a Table object and initialize it from a file
       @param fname Name of file to read parameters from or NULL.
       @param prefixes Space-seperated list of object names. This 
	      determines which lines in the file apply to this object. 
	      (See the class description.)
       @param size Initial table size or 0. The table will grow as needed, but 
	      it is somewhat more efficient to provide a good size to start 
	      with.
    */
   Table(char *fname = NULL, char *prefixes = NULL, int size = 16);
   /** Copy constructor */
   Table( const Table & S );
   /** Destructor */
   virtual ~Table();
   //@}

   AbstractTable * Replicate() const {
     return new Table(*((Table *)this));
   }

   /** Merge entries from a file into the parameter table
       @param fname File name. */
   int Merge(char *fname);

   int Size() const {
     return table.table_size;
   }

   char * Key( int & i) {
     return table.key(i);
   }

   char * strValue( char * key ) {
     return table.get(key);
   }

   int GetValue(const char *key, int &value) const;
   int GetValue(const char *key, double &value) const;
   int GetValue(const char *key, float &value) const;
   int GetValue(const char *key, char *value, int length = 0) const;

   int PutValue(const char *key, int value);
   int PutValue(const char *key, float value);
   int PutValue(const char *key, double value);
   int PutValue(const char *key, const char *value);

   int GetArrayValue(const char *key, int ind, int &value) const;
   int GetArrayValue(const char *key, int ind, float &value) const;
   int GetArrayValue(const char *key, int ind, double &value) const;
   int GetArrayValue(const char *key, int ind, char *value, int length = 0) const;

   int PutArrayValue(const char *key, int ind, int value);
   int PutArrayValue(const char *key, int ind, float value);
   int PutArrayValue(const char *key, int ind, double value);
   int PutArrayValue(const char *key, int ind, const char *value);

   ostream &Write(ostream&s) const;

};

ostream& operator<<(ostream&s, Table&t);

#endif




