// Shannon Scott
// 7/15/1999
// General symmetric matrix class using the LAPACK packed symmetric
// matrix format.  I arbitrarily chose to store the upper triangle
// (LAPACK supports either choice).  Accessing elements of the upper
// matrix is slightly (slightly) simpler than accessing elements of
// the lower triangle.

#ifndef GENSYMMAT_H
#define GENSYMMAT_H

extern int matrix_rnd_seed[];

class GenSymMatrix_d : public Matrix_d
{
public:
  GenSymMatrix_d(void);
  GenSymMatrix_d(int size);
  GenSymMatrix_d(int i, int j);
  GenSymMatrix_d(const GenSymMatrix_d& copy);
  GenSymMatrix_d(char * fname);

  ~GenSymMatrix_d(void);

  int zero(void)           { return setAll( 0.0 ); };
  int ones(void)           { return setAll( 1.0 ); };
  int setAll(double val);
  int random(void);

  double  operator()(int i, int j) const;
  double& operator()(int i, int j);

  double  get(int i, int j) const;
  void    set(int i, int j, double val);

  int numRows(void) const { return size; };
  int numCols(void) const { return size; };

  double normFrobenius(void) const;
  double normInf(void) const;
  double normOnes(void) const;

  int cholesky(UpperTriangle_d& U);
  int solvePositiveDefinite(GeneralMatrix_d& X, const GeneralMatrix_d& B);
  int solve(GeneralMatrix_d& X, const GeneralMatrix_d& B);

  int eig(double* val) const;                   
  int eig(double* val, GeneralMatrix_d& X ) const;

  double*  data;
  
  friend ostream& operator<< (ostream&, const GenSymMatrix_d&);
private:
  int size;
  int nAlloc;  // This value is not really necessary, but it
		    // speeds up some processes (since the amount of
		    // data stored is not as simple as n*m).
};


// These inline functions seem awfuly redundant (look at how similar
// they are!), but I want them to be inlined, and allow the use of
// methods and operators, so...
inline double GenSymMatrix_d::operator()(int i, int j) const
{
  REQUIRE( i <= size, "Row index out of range." );
  REQUIRE( j <= size, "Column index out of range." );

  if( i <= j )
    return data[ i + j * (j - 1) / 2 - 1 ];
  else
    return data[ j + i * (i - 1) / 2 - 1 ];
}

inline double& GenSymMatrix_d::operator()(int i, int j)
{
  REQUIRE( i <= size, "Row index out of range." );
  REQUIRE( j <= size, "Column index out of range." );

  if( i <= j )
    return data[ i + j * (j - 1) / 2 - 1 ];
  else
    return data[ j + i * (i - 1) / 2 - 1 ];
}

inline double GenSymMatrix_d::get(int i, int j) const
{
  REQUIRE( i <= size, "Row index out of range." );
  REQUIRE( j <= size, "Column index out of range." );

  if( i <= j )
    return data[ i + j * (j - 1) / 2 - 1 ];
  else
    return data[ j + i * (i - 1) / 2 - 1 ];
}

inline void GenSymMatrix_d::set(int i, int j, double val)
{
  REQUIRE( i <= size, "Row index out of range." );
  REQUIRE( j <= size, "Column index out of range." );

  if( i <= j )
    data[ i + j * (j - 1) / 2 - 1 ] = val;
  else
    data[ j + i * (i - 1) / 2 - 1 ] = val;
}

#endif
