// Shannon Scott   5/10/1999
// General (dense) matrix class, double precision
// Matrix indexing is one based.
// Everything that does not return a value will return an error-value.
// This is 0 if successful.

#ifndef _GENERAL_MATRIX_D_H
#define _GENERAL_MATRIX_D_H

#ifndef bool
#define bool int
#define false 0
#define true 1
#endif

extern int matrix_rnd_seed[];

class GeneralMatrix_d : public Matrix_d
{
  GeneralMatrix_d(void){};                             
public:
  GeneralMatrix_d(int rows, int cols);      
  GeneralMatrix_d(const GeneralMatrix_d& copy);       
  GeneralMatrix_d(const GeneralMatrix_d& copy, Index rows, Index cols);
  
  GeneralMatrix_d(const Matrix_d& copy);
  GeneralMatrix_d(const Matrix_d& copy, Index rows, Index cols);

  GeneralMatrix_d(const Permutation&          P);
  GeneralMatrix_d(const GenSymMatrix_d&       A);
  GeneralMatrix_d(const BandedMatrix_d&       B);
  GeneralMatrix_d(const UpperTriangle_d&      U);
  GeneralMatrix_d(const LowerTriangle_d&      L);  
  GeneralMatrix_d(const SparseRCSMatrix_d&    A);  
  GeneralMatrix_d(const SymSparseRCSMatrix_d& A);
  
  // Wrap an array to form a vector.  Delete memory when through if
  // delete_mem = true.
  GeneralMatrix_d(int veclen, double* vec, bool delete_mem = false);

  // Wrap an array to form a matrix (column-major order).  Delete
  // memory when through if delete_mem = true.
  GeneralMatrix_d(int rows, int cols,double* mat,
		  bool delete_mem = false);
  // Read a matrix from a file.  The file is plain text, containing the two
  // integers m (number of rows), n (number of columns), followed by the
  // entries of the matrix by columns.
  GeneralMatrix_d( char * fname );

  ~GeneralMatrix_d(void);

  double  get(int i, int j) const;          // Only get value
  void    set(int i, int j, double val);    // Only set value
  double& element(int i, int j);            // Free access to value

  double  operator()(int i, int j) const;
  double& operator()(int i, int j);  

  // Return pointer to internal memory (be careful!)
  double* getData(void){ return data; };

  // Change the flag deciding whether memory is deleted on destruction
  void setDeleteMemFlag(bool _delete_mem)
    { delete_mem = _delete_mem; };
  
  // Memory returned must be deleted!
  GeneralMatrix_d* get(void) const;
  GeneralMatrix_d* get(Index rows, Index cols) const;
  
  // Maybe a better idea.  No need to necessarilly delete memory,
  // but has the overhead of a small copy operation (not much).
  //GeneralMatrix_d get(void) const;
  
  int set(const GeneralMatrix_d& B);
  int set(const GeneralMatrix_d& B, Index rows, Index cols);
  int set(Index rowsA, Index colsA, const GeneralMatrix_d& B,
	  Index rowsB = Index(), Index colsB = Index());

  int set(const Matrix_d&);
  int set(const Matrix_d& B, Index rows, Index cols);
  int set(Index rowsA, Index colsA, const Matrix_d& B,
	  Index rowsB = Index(), Index colsB = Index());

  int zero(void)          { return setAll(0.0); };
  int ones(void)          { return setAll(1.0); };
  int diag(double val);
  int setAll(double val);
  int random(void);

  inline int numRows(void) const { return nRows; };
  inline int numCols(void) const { return nCols; };

  friend ostream& operator<< (ostream&, const GeneralMatrix_d&);

  // Old operations (may be phased out)
  int add(const GeneralMatrix_d& A, const GeneralMatrix_d& B);
  //int add(const Matrix& A, const Matrix& B);
  int mult(const GeneralMatrix_d& B);
  int mult(const GeneralMatrix_d& A, GeneralMatrix_d& B);
  
  // Add a single matrix to this
  int add(const Matrix_d&               B);
  int add(const GenSymMatrix_d&       B);
  int add(const BandedMatrix_d&       B);
  int add(const GeneralMatrix_d&      B);
  int add(const UpperTriangle_d&      U);
  int add(const LowerTriangle_d&      L);
  int add(const SparseRCSMatrix_d&    B);
  int add(const SymSparseRCSMatrix_d& B);

  // Add a single transposed matrix
  int addTrans(const Matrix_d&               B);
  int addTrans(const BandedMatrix_d&       B);
  int addTrans(const GeneralMatrix_d&      B);
  int addTrans(const UpperTriangle_d&      U);
  int addTrans(const LowerTriangle_d&      L);
  int addTrans(const SparseRCSMatrix_d&    B);
  int addTrans(const GenSymMatrix_d&       B){ return add( B ); };
  int addTrans(const SymSparseRCSMatrix_d& B){ return add( B ); };

  // Add matrix to this, C += a * A
  int add(const GenSymMatrix_d&       A, double a);
  int add(const BandedMatrix_d&       A, double a);
  int add(const GeneralMatrix_d&      A, double a);
  int add(const SparseRCSMatrix_d&    A, double a);
  int add(const SymSparseRCSMatrix_d& A, double a);

  // Add transpose of matrix to this, C += a * A'
  int addTrans(const BandedMatrix_d&       A, double a);
  int addTrans(const GeneralMatrix_d&      A, double a);
  int addTrans(const SparseRCSMatrix_d&    A, double a);
  int addTrans(const GenSymMatrix_d&       A, double a)
    { return add( A, a ); };
  int addTrans(const SymSparseRCSMatrix_d& A, double a)
    { return add( A, a ); };
  
  // Multiply two matrices, C = A*B
  int mult(const UpperTriangle_d&      A, const GeneralMatrix_d& B);
  int mult(const LowerTriangle_d&      A, const GeneralMatrix_d& B);
  int mult(const GenSymMatrix_d&       A, const GeneralMatrix_d& B);
  int mult(const GeneralMatrix_d&      A, const GeneralMatrix_d& B);
  int mult(const SymSparseRCSMatrix_d& A, const GeneralMatrix_d& B);
  int mult(const SparseRCSMatrix_d&    A, const GeneralMatrix_d& B);
  int mult(const BandedMatrix_d&       A, const GeneralMatrix_d& B);
  
  // Multiply tranposed matrix A by matrix B
  int multTrans(const GeneralMatrix_d&      A, const GeneralMatrix_d& B);
  int multTrans(const UpperTriangle_d&      A, const GeneralMatrix_d& B);
  int multTrans(const LowerTriangle_d&      A, const GeneralMatrix_d& B);
  int multTrans(const SparseRCSMatrix_d&    A, const GeneralMatrix_d& B);
  int multTrans(const SymSparseRCSMatrix_d& A, const GeneralMatrix_d& B)
    { return mult( A, B ); };
  int multTrans(const GenSymMatrix_d&       A, const GeneralMatrix_d& B)
    { return mult( A, B ); };
  int multTrans(const BandedMatrix_d&       A, const GeneralMatrix_d& B);
  
  // Multiply matrices,  C = a * A * B + b * C
  int mult(const GeneralMatrix_d& A,
	   const GeneralMatrix_d& B,
	         double a, double b);
  
  // Produce image, Y = A * X, where A is this.
  int image(const GeneralMatrix_d& X, GeneralMatrix_d& Y) const;
  
  int solve( GeneralMatrix_d& X, const GeneralMatrix_d& B) const;
  int transposeSolve( GeneralMatrix_d& X, const GeneralMatrix_d& B) const;
  int LU( UpperTriangle_d& U, LowerTriangle_d& L, Permutation& P) const;
  int QR(GeneralMatrix_d& Q, GeneralMatrix_d& R) const;
  int eig(double* val) const;                   
  int eig(double* val, GeneralMatrix_d& X ) const;
  int svd(GeneralMatrix_d& U, GeneralMatrix_d& S, GeneralMatrix_d& V) const;
  
  double condInf(void) const;
  double condOnes(void) const;  
  double normFrobenius(void) const;
  double normInf(void) const;
  double normOnes(void) const;

  int transpose(GeneralMatrix_d& Atrans) const;
  
  // Scalar operations
  int add(double val);
  int mult(double val);

  double* data;

  friend class SparseRCSMatrix_d;
  friend class UpperTriangle_d;
  friend class LowerTriangle_d;
  friend class Permutation;
  friend class GenSymMatrix_d;
  friend class SymSparseRCSMatrix_d;
protected:
  int     nRows;
  int     nCols;
  bool    delete_mem;
};


inline double GeneralMatrix_d::get(int i, int j) const
{
  REQUIRE( i <= nRows, "Row index out of bounds." );
  REQUIRE( j <= nCols, "Column index out of bounds." );
  return data[ (i-1) + (j-1) * nRows ];
};

inline void GeneralMatrix_d::set(int i, int j, double val)
{
  REQUIRE( i <= nRows, "Row index out of bounds." );
  REQUIRE( j <= nCols, "Column index out of bounds." );
  data[ (i-1) + (j-1) * nRows ] = val;
};

inline double& GeneralMatrix_d::element(int i, int j)
{
  REQUIRE( i <= nRows, "Row index out of bounds." );
  REQUIRE( j <= nCols, "Column index out of bounds." );
  return data[i-1 + nRows*(j-1)];
};

inline double GeneralMatrix_d::operator()(int i, int j) const
{
  REQUIRE( i <= nRows, "Row index out of bounds." );
  REQUIRE( j <= nCols, "Column index out of bounds." );
  return data[i-1 + nRows*(j-1)];
};

inline double& GeneralMatrix_d::operator()(int i, int j)
{
  REQUIRE( i <= nRows, "Row index out of bounds." );
  REQUIRE( j <= nCols, "Column index out of bounds." );
  return data[i-1 + nRows*(j-1)];
};

#endif











