// Shannon Scott
// 7/21/1999
// symmetric row-compressed sparse matrix class.
// sparse_rcs_d.H
// 8/2/1999, Converted rowstart array to ones-based indexing for
// compatibility with Yale Sparse Pack (written in Fortran).

#include "mat_inc_d.H"

extern "C" void F77NAME(sdrv)( int * n,int * p,int * ip,int * ia,int * ja,
                       double * a,double * b,double * z,int * nsp,int * isp,
                       double * rsp,int * esp,int * path,int * flag );

//*************************************************************************
SymSparseRCSMatrix_d::SymSparseRCSMatrix_d(int i, int j)
{
  nRows = i;
  nCols = j;

  nAlloc = nCols;    // Just a guess to start with.
  nNonZero = 0;

  data   = new double  [ nAlloc ];
  column = new int[ nAlloc ];

  rowstart = new int[ nRows + 1 ];

  for(int k=0; k < nRows+1; ++k)
    rowstart[k] = 1;
}

//************************************************************************
SymSparseRCSMatrix_d::SymSparseRCSMatrix_d(int i, int j,
					   int _nNonZero)
{
  nRows = i;
  nCols = j;

  nAlloc = _nNonZero;
  nNonZero = 0;

  data   = new double  [ nAlloc ];
  column = new int[ nAlloc ];

  rowstart = new int[ nRows + 1 ];

  for(int k=0; k < nRows+1; ++k)
    rowstart[k] = 1;
}

//*************************************************************************
SymSparseRCSMatrix_d::SymSparseRCSMatrix_d(const SymSparseRCSMatrix_d& copy)
{
  nRows = copy.nRows;
  nCols = copy.nCols;
  nNonZero = copy.nNonZero;
  nAlloc = copy.nAlloc;

  data = new double[ nAlloc ];
  column = new int[ nAlloc ];
  rowstart = new int[ nRows + 1 ];

  int i;
  for(i=0; i < nAlloc; ++i)
    {
      data[i] = copy.data[i];
      column[i] = copy.column[i];
    }
  for(i=0; i < nRows+1; ++i)
    rowstart[i] = copy.rowstart[i];
}

//*********************************************************************
SymSparseRCSMatrix_d::SymSparseRCSMatrix_d(const Matrix_d& copy)
{
  nRows = copy.numRows();
  nCols = copy.numCols();
  nNonZero = 0;
  nAlloc = nRows;
  
  data = new double[ nAlloc ];
  column = new int[ nAlloc ];
  rowstart = new int[ nRows + 1 ];

  for(int j=1; j <= nCols; ++j)
    {
      for(int i=1; i <= nRows; ++i)
	{
	  double val = copy.get(i,j);
	  if( val )
	    {
	      if( i <= j )
		set(i, j, val);
	      else
		set(j, i, val);
	    }
	}
    }
}

//*********************************************************************
SymSparseRCSMatrix_d::SymSparseRCSMatrix_d(const Matrix_d& copy,
					   Index rows, Index cols)
{
  nRows = copy.numRows();
  nCols = copy.numCols();
  nNonZero = 0;
  nAlloc = nRows;
  
  data = new double[ nAlloc ];
  column = new int[ nAlloc ];
  rowstart = new int[ nRows + 1 ];

  set(copy, rows, cols);
}

//**********************************************************************
SymSparseRCSMatrix_d::SymSparseRCSMatrix_d(const GeneralMatrix_d& A)
{
  nRows = A.numRows();
  nCols = A.numCols();
  nNonZero = 0;
  nAlloc = nRows;
  
  data = new double[ nAlloc ];
  column = new int[ nAlloc ];
  rowstart = new int[ nRows + 1 ];
  
  for(int j=0; j < nCols; ++j)
    {
      for(int i=0; i < nRows; ++i)
	{
	  double val = A.data[ i + nRows * j ];
	  if( val )
	    {
	      if( i <= j )
		set(i+1, j+1, val);
	      else
		set(j+1, i+1, val);
	    }
	}
    }
}

//************************************************************************
SymSparseRCSMatrix_d::~SymSparseRCSMatrix_d(void)
{
  delete[] data;
  delete[] column;
  delete[] rowstart;
}

//*********************************************************************
double SymSparseRCSMatrix_d::get(int i, int j) const
{
  REQUIRE( i <= nRows, "Row index out of bounds.");
  REQUIRE( j <= nCols, "Column index out of bounds.");

  if( i > j )
    {
      int temp = i;
      i = j;
      j = temp;
    }
  
  // Note: this, and all related methods for this class, use a linear
  // search technique.  But, since the bounds of each row are known,
  // you could use a binary search to speed things up.  I decided
  // against this for this first version.
  int k;
  for(k=rowstart[i-1]-1; k < rowstart[i]-1; ++k)
    {
      if( column[k] == j )
	return data[k];
    }
  return 0.0;
}

//***********************************************************************
void SymSparseRCSMatrix_d::set(int i, int j, double val)
{
  REQUIRE( i <= nRows, "Row index out of bounds.");
  REQUIRE( j <= nCols, "Column index out of bounds.");
  
  int k, n;

  if( i > j )
    {
      int temp = i;
      i = j;
      j = temp;
    }

  // We need to find the correct index at which to place the new
  // value.  If there is already a non-zero value at this location,
  // return a reference to it, otherwise find the place to insert it
  // (the value of k at the end).
  for(k=rowstart[i-1]-1; k < rowstart[i]-1; ++k)
    {
      if( column[k] == j )
	{
	  data[k] = val;
	  return;
	}
      else if( column[k] > j )
	break;
    }

  // Is the value non-zero?
  if( val == 0.0 )
    return;
  
  // Is there enough room to add another entry?
  if( nAlloc <= nNonZero )
    resize();
  
  // Shift the data down to make room for the new value.
  for( n = nNonZero; n > k; --n )
    {
      data  [n] = data  [n-1];
      column[n] = column[n-1];
    }

  // Adjust the values of rowstart
  for(n=i; n < nRows+1; ++n)
    ++ rowstart[n];

  // Increment nNonZero
  ++ nNonZero;

  // Set the value
  column[k] = j;
  data[k] = val;
}

//************************************************************************
int SymSparseRCSMatrix_d::set(const SymSparseRCSMatrix_d& B)
{
  REQUIRE( nRows == B.nRows, "Row dimensions do not agree." );
  REQUIRE( nCols == B.nCols, "Column dimensions do not agree." );

  if( nAlloc < B.nNonZero )
    {
      delete[] data;
      delete[] column;
      data = new double[ B.nAlloc ];
      column = new int[ B.nAlloc ];
      nAlloc = B.nAlloc;
    }
  
  int i;

  for(i=0; i <= nRows; ++i)
    rowstart[i] = B.rowstart[i];

  nNonZero = B.nNonZero;
  for(i=0; i < nNonZero; ++i)
    {
      data[i]   = B.data[i];
      column[i] = B.column[i];
    }

  return 0;
}

/*//*************************************************************************
int SymSparseRCSMatrix_d::set(const SymSparseRCSMatrix_d& B,
Index rows, Index cols)
{
  // Initialize index objects (check bounds)
  REQUIRE( rows.init(1, B.numRows()) == 0,
	   "Row index initialization failed." );
  
  REQUIRE( cols.init(1, B.numCols()) == 0,
	   "Column index initialization failed");
  
  REQUIRE( nRows == B.nRows, "Row dimensions do not agree." );
  REQUIRE( nCols == B.nCols, "Column dimensions do not agree." );

  int i, j, n, m;

  // Iterate through columns submatrix
  for(j=cols.getFirst(), m=0;    j > 0;   j=cols.getNext(), ++m)
    {
      // Iterate through rows of submatrix
      for(i=rows.getFirst(), n=0;   i > 0;   i=rows.getNext(), ++n)
	{
	  if( B(i,j) )
	    set(n, m, B(i,j));
	}// for rows
    }// for cols

}*/

//*************************************************************************
int SymSparseRCSMatrix_d::set(const GeneralMatrix_d& B)
{
  REQUIRE( nRows == B.nRows, "Row dimensions do not agree." );
  REQUIRE( nCols == B.nCols, "Column dimensions do not agree." );

  int i,j;

  // Clean out this matrix
  nNonZero = 0;
  for(i=0; i <= nRows; ++i)
    rowstart[i] = 1;

  // Grab the values from the upper triangle of the general matrix
  for(j=1; j <= nCols; ++j)
    {
      for(i=1; i <= j; ++i)
	{
	  if( B(i,j) )
	    set(i,j, B(i,j));
	}
    }
  return 0;
}

//*********************************************************************
int SymSparseRCSMatrix_d::set(const GeneralMatrix_d& B,
			      Index rows, Index cols)
{
  // Initialize index objects (check bounds)
  REQUIRE( rows.init(1, B.numRows()) == 0 && cols.init(1, B.numCols()) == 0,
	   "Index initialization failed");

  REQUIRE(nRows == rows.size(), "Ranges do not agree");
  REQUIRE(nCols == cols.size(), "Domains do not agree");
  
  int i, j, n, m;

  // Iterate through columns submatrix
  for(j=cols.getFirst(), m=1;
              j > 0; 
      j=cols.getNext(), ++m)
    {

      // Iterate through rows of submatrix
      for(i=rows.getFirst(), n=1; 
	         i > 0; 
	  i=rows.getNext(), ++n)
	{
	  if( B(i,j) != 0.0 )
	    set(n, m, B(i,j));
	    //else
	    //remove(n, m);
	}// for rows
    }// for cols

  return 0;
}

//*********************************************************************
int SymSparseRCSMatrix_d::set(Index rowsA, Index colsA, 
			   const GeneralMatrix_d& B,
			   Index rowsB, Index colsB)
{
  // Initialize index objects for this matrix
  REQUIRE(rowsA.init(1, nRows) == 0,
	  "Initialization of row index for matrix A failed");
  REQUIRE(colsA.init(1, nCols) == 0,
	  "Initialization of column index for matrix A failed");

  // Initialize index objects for B matrix
  REQUIRE(rowsB.init(1, B.nRows) == 0,
	  "Initialization of row index for matrix B failed");
  REQUIRE(colsB.init(1, B.nCols) == 0,
	  "Initialization of column index for matrix B failed");
  
  // Check that both sub-matrices are the same size
  REQUIRE( rowsA.size() == rowsB.size(), "Ranges do not agree" );
  REQUIRE( colsA.size() == colsB.size(), "Domains do not agree" );

  int ai, aj, bi, bj;

  // Iterate over columns of sub-matrices
  for( aj = colsA.getFirst(), bj = colsB.getFirst();
       aj > 0 && bj > 0;
       aj = colsA.getNext(), bj = colsB.getNext())
    {

      // Iterate over rows of sub-matrices
      for( ai = rowsA.getFirst(), bi = rowsB.getFirst();
	   ai > 0 && bi > 0;
	   ai = rowsA.getNext(), bi = rowsB.getNext())
	{
	  if( B(bi, bj) != 0.0 )
	    set(ai, aj, B(bi, bj));
	  //else
	  //remove(ai, aj);
	}// for rows
    }// for columns

  return 0;
}

//*********************************************************************
int SymSparseRCSMatrix_d::set(const Matrix_d& B, 
			      Index rows, Index cols)
{
  // Initialize index objects (check bounds)
  REQUIRE( rows.init(1, B.numRows()) == 0 && cols.init(1, B.numCols()) == 0,
	   "Index initialization failed");

  REQUIRE(nRows == rows.size(), "Ranges do not agree");
  REQUIRE(nCols == cols.size(), "Domains do not agree");

  int i, j, n, m;

  for(j=cols.getFirst(), m=1;
             j > 0; 
      j=cols.getNext(), ++m)
    {
      for(i=rows.getFirst(), n=1; 
	      i > 0; 
	  i=rows.getNext(), ++n)
	{
	  double val = B.get(i, j);
	  if( val != 0.0 )
	    set(n, m, val);
	  //else
	  //  remove(n, m);
	}// for rows
    }// for cols

  return 0;
}

//*********************************************************************
int SymSparseRCSMatrix_d::set(Index rowsA, Index colsA, 
			      const Matrix_d& B,
			      Index rowsB, Index colsB)
{
  // Initialize index objects for this matrix
  REQUIRE(rowsA.init(1, nRows) == 0,
	  "Initialization of row index for matrix A failed");
  REQUIRE(colsA.init(1, nCols) == 0,
	  "Initialization of column index for matrix A failed");

  // Initialize index objects for B matrix
  REQUIRE(rowsB.init(1, B.numRows()) == 0,
	  "Initialization of row index for matrix B failed");
  REQUIRE(colsB.init(1, B.numCols()) == 0,
	  "Initialization of column index for matrix B failed");
  
  // Check that both sub-matrices are the same size
  REQUIRE( rowsA.size() == rowsB.size(), "Ranges do not agree");
  REQUIRE( colsA.size() == colsB.size(), "Domains do not agree");

  int ai, aj, bi, bj;

  for( aj = colsA.getFirst(), bj = colsB.getFirst();
       aj > 0 && bj > 0;
       aj = colsA.getNext(), bj = colsB.getNext())
    {
      for( ai = rowsA.getFirst(), bi = rowsB.getFirst();
	   ai > 0 && bi > 0;
	   ai = rowsA.getNext(), bi = rowsB.getNext())
	{
	  double val = B.get(bi, bj);
	  if( val != 0.0 )
	    set(ai, aj, val);
	  //else
	  //  remove(ai, aj);
	}// for rows
    }// for columns

  return 0;
}

//***********************************************************************
double SymSparseRCSMatrix_d::operator()(int i, int j) const
{
  REQUIRE( i <= nRows, "Row index out of bounds.");
  REQUIRE( j <= nCols, "Column index out of bounds.");

  if( i > j )
    {
      int temp = i;
      i = j;
      j = temp;
    }

  // Note: this, and all related methods for this class, use a linear
  // search technique.  But, since the bounds of each row are known,
  // you could use a binary search to speed things up.  I decided
  // against this for this first version.
  int k;
  for(k=rowstart[i-1]-1; k < rowstart[i]-1; ++k)
    {
      if( column[k] == j )
	return data[k];
    }
  return 0.0;
}

//************************************************************************
double& SymSparseRCSMatrix_d::operator()(int i, int j)
{
  REQUIRE( i <= nRows, "Row index out of bounds.");
  REQUIRE( j <= nCols, "Column index out of bounds.");
  
  int k, n;

  if( i > j )
    {
      int temp = i;
      i = j;
      j = temp;
    }

  // We need to find the correct index at which to place the new
  // value.  If there is already a non-zero value at this location,
  // return a reference to it, otherwise find the place to insert it
  // (the value of k at the end).
  for(k=rowstart[i-1]-1; k < rowstart[i]-1; ++k)
    {
      if( column[k] == j )
	return data[k];
      else if( column[k] > j )
	break;
    }

  // Is there enough room to add another entry?
  if( nAlloc <= nNonZero )
    resize();
  
  // Shift the data down to make room for the new value.
  for( n = nNonZero; n > k; --n )
    {
      data  [n] = data  [n-1];
      column[n] = column[n-1];
    }

  // Adjust the values of rowstart
  for(n=i; n < nRows+1; ++n)
    ++ rowstart[n];

  // Increment nNonZero
  ++ nNonZero;

  // Zero the new location (this turns out to be important)
  data[k] = 0.0;
  
  // Return the new location
  column[k] = j;
  return data[k];
}

//************************************************************************
int SymSparseRCSMatrix_d::add(const SymSparseRCSMatrix_d& A)
{
  REQUIRE( nRows == A.nRows, "Ranges do not agree." );
  REQUIRE( nCols == A.nCols, "Domains do not agree." );
  
  int m,n;

  for(m=0; m < A.nRows; ++m)
    {
      for(n=rowstart[m]-1; n < rowstart[m+1]-1; ++n)
	{
	  (*this)(m+1, column[n]) += A.data[n];
	}
    }
  return 0;
}

/* This needs to be symmetrized
//************************************************************************
int SymSparseRCSMatrix_d::mult(const GeneralMatrix_d& X,
			       GeneralMatrix_d& B)
{
  REQUIRE( nCols == X.numRows(), "Domain does not agree." );
  REQUIRE( nRows == B.numRows(), "Range does not agree." );
  REQUIRE( X.numCols() == B.numCols(), "Dimensions of X and B do not agree." );

  int i, j, k;

  for(k=1; k <= X.numCols(); ++k)
    {
      for(i=1; i <= nRows; ++i)
	{
	  double sum = 0;
	  for(j = rowstart[i-1]; j < rowstart[i]; ++j)
	    {
	      sum += data[j] * X( column[j], k );
	    }
	  B( i, k ) = sum;
	}
    }
  
  return 0;
}
*/

/* This needs to be symmetrized
//************************************************************************
int SymSparseRCSMatrix_d::mult(const SymSparseRCSMatrix_d& X,
			       GeneralMatrix_d& B)
{
  REQUIRE( nCols == X.nRows, "Domain does not agree." );
  REQUIRE( nRows == B.numRows(), "Range does not agree." );
  REQUIRE( X.nCols == B.numCols(), "Dimensions of X and B do not agree." );

  int i, j, k;

  for(k=1; k <= X.numCols(); ++k)
    {
      for(i=1; i <= nRows; ++i)
	{
	  double sum = 0;
	  for(j = rowstart[i-1]; j < rowstart[i]; ++j)
	    {
	      sum += data[j] * X( column[j], k );
	    }
	  B( i, k ) = sum;
	}
    }
  
  return 0;
}*/

//************************************************************************
int SymSparseRCSMatrix_d::solve( GeneralMatrix_d& X,
                                 const GeneralMatrix_d& B) const
{
  REQUIRE( nCols == X.numRows(), "Domain does not agree." );
  REQUIRE( nRows == B.numRows(), "Range does not agree." );
  REQUIRE( X.numCols() == B.numCols(), "Dimensions of X and B do not agree." );
  REQUIRE( X.numCols() == 1,
           "Can only handle one right hand side at the moment" );
  int i;

   int locN = X.numRows();
   int fac = 5;
   int nsp = (3*nRows+2*nAlloc)*fac;
   double * rsp = new double[nsp];
   int * perm = new int[nRows];
   for( i=0;i<nRows;i++ ) perm[i] = i+1;
   int esp,flag;
   int path = 5;
   int done = 0;
   int IncStorageCntr=0;
   int IncStorageLimit=4;
   while( !done )
   {
      F77NAME(sdrv)( (int*)&locN,perm,perm,rowstart,column,data,
             B.data,X.data,(int*)&nsp,(int*)rsp,rsp,(int*)&esp,
             (int*)&path,(int*)&flag );
      if( flag != 0 )
      {
         int e = flag/locN;
         int r = flag-locN*e;
         switch(e)
         {
            case 2:
               cerr << "Error in SymSparseRCSMatrix_d::solve: error "
                       "code from sdrv is " << flag << endl;
               cerr << "Duplicate entry in matrix, row = " << r << endl;
               exit(1);
               break;
            case 6:
               IncStorageCntr++;
               if( IncStorageCntr <= IncStorageLimit )
               {
                  fac *= 2;
                  nsp = (3*nRows+2*nAlloc)*fac;
                  delete[] rsp;
                  rsp = new double[nsp];
               }
               else
               {
                  cerr << "Error in SymSparseRCSMatrix_d::solve: error "
                          "code from sdrv is " << flag << endl;
                  cerr << "Insufficient storage in SSF, row = " << r
                       << endl;
                  exit(1);
               }
               break;
            case 7:
               IncStorageCntr++;
               if( IncStorageCntr <= IncStorageLimit )
               {
                  fac *= 2;
                  nsp = (3*nRows+2*nAlloc)*fac;
                  delete[] rsp;
                  rsp = new double[nsp];
               }
               else
               {
                  cerr << "Error in SymSparseRCSMatrix_d::solve: error "
                          "code from sdrv is " << flag << endl;
                  cerr << "Insufficient storage in SNF" << endl;
                  exit(1);
               }
               break;
            case 8:
               cerr << "Error in SymSparseRCSMatrix_d::solve: error "
                       "code from sdrv is " << flag << endl;
               cerr << "Zero pivot, row = " << r << endl;
               exit(1);
               break;
            case 10:
               IncStorageCntr++;
               if( IncStorageCntr <= IncStorageLimit )
               {
                  fac *= 2;
                  nsp = (3*nRows+2*nAlloc)*fac;
                  delete[] rsp;
                  rsp = new double[nsp];
               }
               else
               {
                  cerr << "Error in SymSparseRCSMatrix_d::solve: error "
                          "code from sdrv is " << flag << endl;
                  cerr << "Insufficient storage in SDRV" << endl;
                  exit(1);
               }
               break;
            case 11:
               cerr << "Error in SymSparseRCSMatrix_d::solve: error "
                       "code from sdrv is " << flag << endl;
               cerr << "Illegal path specification" << endl;
               exit(1);
               break;
            default:
               cerr << "Error in SymSparseRCSMatrix_d::solve: error "
                       "code from sdrv is " << flag << endl;
               cerr << "Unknown error flag" << endl;
               exit(1);
               break;
         }
      }
      else
         done = 1;
   }
   path = 3;
   done=0;
   while( !done )
   {
      F77NAME(sdrv)( (int*)&locN,perm,perm,rowstart,column,data,
             B.data,X.data,(int*)&nsp,(int*)rsp,rsp,(int*)&esp,
             (int*)&path,(int*)&flag );
         if( flag != 0 )
         {
            int e = flag/locN;
            int r = flag-locN*e;
            switch(e)
            {
               case 2:
                  cerr << "Error in SymSparseRCSMatrix_d::solve: "
                          "error code from sdrv is " << flag << endl;
                  cerr << "Duplicate entry in matrix, row = " << r << endl;
                  exit(1);
                  break;
               case 6:
                  IncStorageCntr++;
                  if( IncStorageCntr <= IncStorageLimit )
                  {
                     fac *= 2;
                     nsp = (3*nRows+2*nAlloc)*fac;
                     delete[] rsp;
                     rsp = new double[nsp];
                  }
                  else
                  {
                     cerr << "Error in SymSparseRCSMatrix_d::solve: "
                             "error code from sdrv is " << flag << endl;
                     cerr << "Insufficient storage in SSF, row = " << r
                          << endl;
                     exit(1);
                  }
                  break;
               case 7:
                  IncStorageCntr++;
                  if( IncStorageCntr <= IncStorageLimit )
                  {
                     fac *= 2;
                     nsp = (3*nRows+2*nAlloc)*fac;
                     delete[] rsp;
                     rsp = new double[nsp];
                  }
                  else
                  {
                     cerr << "Error in SymSparseRCSMatrix_d::solve: "
                             "error code from sdrv is " << flag << endl;
                     cerr << "Insufficient storage in SNF" << endl;
                     exit(1);
                  }
                  break;
               case 8:
                  cerr << "Error in SymSparseRCSMatrix_d::solve: "
                          "error code from sdrv is " << flag << endl;
                  cerr << "Zero pivot, row = " << r << endl;
                  exit(1);
                  break;
               case 10:
                  IncStorageCntr++;
                  if( IncStorageCntr <= IncStorageLimit )
                  {
                     fac *= 2;
                     nsp = (3*nRows+2*nAlloc)*fac;
                     delete[] rsp;
                     rsp = new double[nsp];
                  }
                  else
                  {
                     cerr << "Error in SymSparseRCSMatrix_d::solve: "
                             "error code from sdrv is " << flag << endl;
                     cerr << "Insufficient storage in SDRV" << endl;
                     exit(1);
                  }
                  break;
               case 11:
                  cerr << "Error in SymSparseRCSMatrix_d::solve: "
                          "error code from sdrv is " << flag << endl;
                  cerr << "Illegal path specification" << endl;
                  exit(1);
                  break;
               default:
                  cerr << "Error in SymSparseRCSMatrix_d::solve: "
                          "error code from sdrv is " << flag << endl;
                  cerr << "Unknown error flag" << endl;
                  exit(1);
                  break;
            }
         }
         else
            done = 1;
      }
  delete[] rsp;
  delete[] perm;
  return 0;
}

/*
//************************************************************************
double SymSparseRCSMatrix_d::normFrobenius(void)
{
  double normval = 0.0;
  for(int i=0; i < nNonZero; ++i)
    normval += data[i] * data[i];
  return sqrt( normval );
}

//************************************************************************
double SymSparseRCSMatrix_d::normInf(void)
{
  double normval = 0.0;

  // Iterate over rows
  for(int i=0; i < nRows; ++i)     
    {
      double subnorm = 0.0;
      
      // Iterate over non-zero columns
      for(int j = rowstart[i]; j < rowstart[j+1]; ++j)
	{
	  subnorm += fabs( data[j] );
	}
      if( subnorm > normval)
	normval = subnorm;
    }
  return normval;
}

//***********************************************************************
double SymSparseRCSMatrix_d::normOnes(void)
{
  double normval = 0.0;

  // Iterate over columns
  for(int j=1; j <= nCols; ++j)    
    {
      double subnorm = 0.0;

      // Iterate over rows
      for(int i=0; i < nRows; ++i)
	{
	  // Find the correct entry
	  for(int k = rowstart[i]; k < rowstart[i+1]; ++k)
	    {
	      if( column[k] == j )
		{
		  subnorm += fabs( data[k] );
		  break;
		}
	      else if( column[k] > j )
		break;
	    }// for k
	}// for i

      if( subnorm > normval )
	normval = subnorm;
    }// for j
  
  return normval;
}
*/
//************************************************************************
ostream& operator<< (ostream& os, const SymSparseRCSMatrix_d& A)
{
  int i, j;

  for(i=1; i <= A.nRows; ++i)
    {
      for(j=1; j < i; ++j)
	{
	  os << A(j,i) << '\t';
	}
      for(j=i; j <= A.nCols; ++j)
	{
	  os << A(i,j) << '\t';
	}
      os << endl;
    }
  os << endl;
  return os;
}

//****************************************************************************
void SymSparseRCSMatrix_d::dump(ostream& os)
{
  int k;

  os << "data:     ";
  for(k=0; k < nNonZero; ++k)
    os << data[k] << '\t';
  os << endl;

  os << "column:   ";
  for(k=0; k < nNonZero; ++k)
    os << column[k] << '\t';
  os << endl;

  os << "rowstart: ";
  for(k=0; k < nRows+1; ++k)
    os << rowstart[k] << '\t';
  os << endl;
}

//*************************************************************************
void SymSparseRCSMatrix_d::resize(void)
{
  double*   tData    = new double  [ nAlloc * 2 ];
  int* tColumn  = new int[ nAlloc * 2 ];

  for(int i=0; i < nNonZero; ++i)
    {
      tData  [i] = data  [i];
      tColumn[i] = column[i];
    }

  delete[] data;
  delete[] column;

  data = tData;
  column  = tColumn;

  nAlloc *= 2;
}

//************************************************************************
void SymSparseRCSMatrix_d::trim(void)
{
  nAlloc = nNonZero;

  double*   temp_data = new double[ nAlloc ];
  int* temp_column = new int[ nAlloc ];

  int i;
  for(i=0; i < nAlloc; ++i)
    {
      temp_data[i] = data[i];
      temp_column[i] = column[i];
    }

  delete[] data;
  delete[] column;

  data = temp_data;
  column = temp_column;

  // Some error checking
  #ifdef _DEBUG_
  for(i=0; i <= nRows; ++i)
    REQUIRE( rowstart[i] <= nAlloc, "Invalid rowstart index." );
  #endif
}

//**************************************************************************
void SymSparseRCSMatrix_d::remove(int i, int j)
{
  int k, m;

  if( i > j )
    {
      int temp = i;
      i = j;
      j = temp;
    }

  // Find the entry, if it exists
  for( k = rowstart[i-1]-1; k < rowstart[i]-1; ++k )
    {
      if( column[k] == j )
	{
	  nNonZero --;
	  for(m = k; m < nNonZero; ++m)
	    {
	      data[m] = data[m+1];
	      column[m] = column[m+1];
	    }
	  for(m = i; m <= nRows; ++m)
	    -- rowstart[m];
	  return;
	}
      else if( column[k] > j )
	return;
    }
}

//***************************************************************************
void SymSparseRCSMatrix_d::removeZeros(void)
{
  int i, j, k;

  // Look for zero values
  for(i=0; i < nRows; ++i)
    {
      for(j = rowstart[i]-1; j < rowstart[i+1]-1; ++j)
	{
	  // Remove the zero values
	  if( data[j] == 0 )
	    {
	      nNonZero --;
	      for(k = j; k < nNonZero; ++k)
		{
		  data[k] = data[k+1];
		  column[k] = column[k+1];
		}
	      for(k = i+1; k <= nRows; ++k)
		-- rowstart[k];
	    }
	}
    }
}

//****************************************************************************
int SymSparseRCSMatrix_d::zero(void)
{
  nNonZero = 0;
  for(int k=0; k < nRows+1; ++k)
    rowstart[k] = 1;
  return 0;
}
