c==============================================================
c
c     Routine: fnlj
c
c     Purpose: Calculate the Lennard-Jones potential
c
c     Input parameters:
c
c        n          integer - number of variables, (3 times the 
c                             number of coordinates - 6)
c                             
c
c        x          real(*) - the independent variables; x = (x,y,z)
c
c        gflag      integer - boolean which flags whether the gradient
c                             is to be calculated
c
c     Output parameters:
c
c        fcur       real - the objective value
c
c        grad       real(*) - the objective gradient
c
c
c================================================================

      subroutine fnlj(n,x,fcur)

      implicit none
      double precision x(*),fcur
      integer n

c=====================================================================
c
c Local variables:
c
c    i,j   loop control variables
c
c=======================================================================

      integer i,j,ii
      double precision z
      double precision x1(3)
      double precision x2(3)

      fcur = 0.d0

      do j = 1, (n+6)/3
	 do i = j+1, (n+6)/3
	    if(j.eq.1)then
               x1(1) = 0.d0
	       x1(2) = 0.d0
	       x1(3) = 0.d0
            endif
	    if(j.eq.2)then
	       x1(1) = x(1)
	       x1(2) = 0.d0
	       x1(3) = 0.d0
            endif
	    if(j.eq.3)then
	       x1(1) = x(2)
	       x1(2) = x(3)
	       x1(3) = 0.d0
            endif
	    if(j.gt.3)then
	       do ii = 1,3
		    x1(ii) =  x( (j-3)*3+ii )
	       enddo
            endif
	    if(i.eq.2)then
	       x2(1) = x(1)
	       x2(2) = 0.d0
	       x2(3) = 0.d0
	    endif
	    if(i.eq.3)then
	       x2(1) = x(2)
	       x2(2) = x(3)
	       x2(3) = 0.d0
            endif
	    if(i.gt.3)then
	       do ii = 1, 3
		     x2(ii) =  x( (i-3)*3+ii )
               enddo
            endif
            z = 0.d0
	    do ii = 1, 3
	        z = (x1(ii)-x2(ii))**2 + z
	    enddo
	    fcur = 1.0/z**6 - 2.0/z**3 + fcur
	 enddo

      enddo

      return
      end


      subroutine gradlj(n,x,grad)

      implicit none
      integer n
      double precision x(*),grad(*)

      integer i,j,ii
      double precision z
      double precision x1(3)
      double precision x2(3)
      double precision temp

        do i = 1, n
	   grad(i) = 0.d0
        enddo
        do j = 2, (n+6)/3
	   if(j.gt.3)then
	     do ii = 1, 3
	       x1(ii) = x(3*(j-3)+ii)
	     enddo
	   elseif(j.eq.2)then
	      x1(1) = x(1)
	      x1(2) = 0.d0
	      x1(3) = 0.d0
           else
	      x1(1) = x(2)
	      x1(2) = x(3)
	      x1(3) = 0.d0
           endif
           do i = 1, (n+6)/3
	      if(i.ne.j)then
	         if(i.gt.3)then
	             do ii = 1, 3
	                x2(ii) = x(3*(i-3)+ii)
	             enddo
	         elseif(i.eq.2)then
	             x2(1) = x(1)
		     x2(2) = 0.d0
		     x2(3) = 0.d0
	         elseif(i.eq.3)then
	             x2(1) = x(2)
		     x2(2) = x(3)
		     x2(3) = 0.d0
                  else
		     x2(1) = 0.d0
		     x2(2) = 0.d0
		     x2(3) = 0.d0
                  endif
		  z = 0.d0
                  do ii = 1, 3
		     z = (x1(ii) - x2(ii))*(x1(ii) - x2(ii))+z
		  enddo
                  temp = 6.0/z**4-6.0/z**7
		  if(j.gt.3)then
		      do ii = 1, 3
			 grad(3*(j-3) + ii) = 
     +                   grad(3*(j-3)+ii)+2.0*temp*(x1(ii)-x2(ii))
		       enddo
		   elseif(j.eq.3)then
			 grad(2) = grad(2)+2.0*temp*(x1(1)-x2(1))
			 grad(3) = grad(3)+2.0*temp*(x1(2)-x2(2))
                   else
			 grad(1) = grad(1)+2.0*temp*(x1(1)-x2(1))
                   endif
                endif
           enddo
        enddo


      return
      end

      subroutine hesslj(n,x,H)

      implicit none
      integer n
      double precision x(*),H(n,*)

      integer i,j,ii,jj
      double precision z,z1,z2
      double precision x1(3)
      double precision x2(3)

C      write(*,*) 'n = ',n
      do j = 1,n
         do i=1,n
            H(i,j) = 0.d0
         enddo
      enddo

      do i = 1,(n+6)/3-1
         if(i.gt.3)then
            do ii = 1, 3
               x1(ii) = x(3*(i-3)+ii)
            enddo
         elseif(i.eq.1)then
            x1(1) = 0.d0
            x1(2) = 0.d0
            x1(3) = 0.d0
         elseif(i.eq.2)then
            x1(1) = x(1)
            x1(2) = 0.d0
            x1(3) = 0.d0
         else
            x1(1) = x(2)
            x1(2) = x(3)
            x1(3) = 0.d0
         endif
         do j = i+1,(n+6)/3
            if(j.gt.3)then
               do ii = 1, 3
                  x2(ii) = x(3*(j-3)+ii)
               enddo
            elseif(j.eq.1)then
               x2(1) = 0.d0
               x2(2) = 0.d0
               x2(3) = 0.d0
            elseif(j.eq.2)then
               x2(1) = x(1)
               x2(2) = 0.d0
               x2(3) = 0.d0
            else
               x2(1) = x(2)
               x2(2) = x(3)
               x2(3) = 0.d0
            endif
C            write(*,*) 'i,j = ',i,j
            z = 0.d0
            do ii = 1, 3
               z = (x1(ii) - x2(ii))*(x1(ii) - x2(ii))+z
            enddo
            z1 = 7.0d0/z**8-4.0d0/z**5
            z2 = 1.0d0/z**4-1.0d0/z**7
            if( i.eq.1 .and. j.eq.2 ) then
               H(1,1) = H(1,1) + 6.0d0*(z1*4*x(1)*x(1)+2.0d0*z2)
            elseif( i.eq.1 .and. j.eq.3 ) then
               H(2,3) = H(2,3) + 24.0d0*x(2)*x(3)*z1
               do ii=2,3
                  H(ii,ii) = H(ii,ii) + 6.0d0*(z1*4.0d0*x(ii)*x(ii)+
     +                                  2.0d0*z2)
               enddo
            elseif( i.eq.1 .and. j.gt.3 ) then
               do ii=3*j-8,3*j-6
                  H(ii,ii) = H(ii,ii) + 6.0d0*(z1*4.0d0*x(ii)*x(ii)+
     +                                  2.0d0*z2)
               enddo
               do ii=3*j-8,3*j-7
                  do jj=ii+1,3*j-6
                     H(ii,jj) = H(ii,jj) + 24.0d0*x(ii)*x(jj)*z1
                  enddo
               enddo
            elseif( i.eq.2 .and. j.eq.3 ) then
               H(1,1) = H(1,1) + 6.0d0*(z1*4.0d0*(x(1)-x(2))*(x(1)-x(2))
     +                         + 2.0d0*z2)
               H(2,2) = H(2,2) + 6.0d0*(z1*4.0d0*(x(1)-x(2))*(x(1)-x(2))
     +                         + 2.0d0*z2)
               H(3,3) = H(3,3) + 6.0d0*(z1*4.0d0*x(3)*x(3)
     +                         + 2.0d0*z2)
               H(1,2) = H(1,2) - 6.0d0*(z1*4.0d0*(x(1)-x(2))*(x(1)-x(2))
     +                         + 2.0d0*z2)
               H(1,3) = H(1,3) + 24.0d0*(x(1)-x(2))*x(3)*z1
               H(2,3) = H(2,3) + 24.0d0*(x(2)-x(1))*x(3)*z1
            elseif( i.eq.2 .and. j.gt.3 ) then
               H(1,1) = H(1,1) + 6.0d0*(z1*4.0d0*(x(1)-x(3*j-8))*
     +                           (x(1)-x(3*j-8))+2.0d0*z2)
               H(1,3*j-8) = H(1,3*j-8) - 6.0d0*(z1*4.0d0*
     +                      (x(1)-x(3*j-8))*(x(1)-x(3*j-8))+2.0d0*z2)
               do jj=3*j-7,3*j-6
                  H(1,jj) = H(1,jj) + 24.0d0*z1*(x(1)-x(3*j-8))*x(jj)
               enddo
               H(3*j-8,3*j-8) = H(3*j-8,3*j-8) + 6.0d0*(z1*4.0d0*
     +                      (x(3*j-8)-x(1))*(x(3*j-8)-x(1))+2*z2)
               do jj=3*j-7,3*j-6
                  H(3*j-8,jj) = H(3*j-8,jj) + 24.0d0*z1*
     +                          (x(3*j-8)-x(1))*x(jj)
                  H(jj,jj) = H(jj,jj) + 6.0d0*(z1*4.0d0*x(jj)*x(jj)
     +                                + 2.0d0*z2)
               enddo
               H(3*j-7,3*j-6) = H(3*j-7,3*j-6) + 24.0d0*z1*
     +                          x(3*j-7)*x(3*j-6)
            elseif( i.eq.3 ) then
               H(2,2) = H(2,2) + 6.0d0*(z1*4.0d0*(x(2)-x(3*j-8))*
     +                           (x(2)-x(3*j-8))+2.0d0*z2)
               H(3,3) = H(3,3) + 6.0d0*(z1*4.0d0*(x(3)-x(3*j-7))*
     +                           (x(3)-x(3*j-7))+2.0d0*z2)
               H(2,3) = H(2,3) + 24*(x(2)-x(3*j-8))*(x(3)-x(3*j-7))*z1
               H(2,3*j-8) = H(2,3*j-8) - 6.0d0*(z1*4.0d0*
     +                       (x(3*j-8)-x(2))*(x(3*j-8)-x(2))+2.0d0*z2)
               H(2,3*j-7) = H(2,3*j-7) + 24.0d0*(x(2)-x(3*j-8))*
     +                       (x(3*j-7)-x(3))*z1
               H(2,3*j-6) = H(2,3*j-6) + 24.0d0*(x(2)-x(3*j-8))*
     +                       x(3*j-6)*z1
               H(3,3*j-8) = H(3,3*j-8) + 24.0d0*(x(3)-x(3*j-7))*
     +                       (x(3*j-8)-x(2))*z1
               H(3,3*j-7) = H(3,3*j-7) - 6.0d0*(z1*4.0d0*
     +                      (x(3)-x(3*j-7))*(x(3)-x(3*j-7))+2.0d0*z2)
               H(3,3*j-6) = H(3,3*j-6) + 24.0d0*(x(3)-x(3*j-7))*
     +                       x(3*j-6)*z1
               H(3*j-8,3*j-8) = H(3*j-8,3*j-8) + 6.0d0*(z1*4.0d0*
     +                       (x(3*j-8)-x(2))*(x(3*j-8)-x(2))+2.0d0*z2)
               H(3*j-7,3*j-7) = H(3*j-7,3*j-7) + 6.0d0*(z1*4.0d0*
     +                       (x(3*j-7)-x(3))*(x(3*j-7)-x(3))+2.0d0*z2)
               H(3*j-6,3*j-6) = H(3*j-6,3*j-6) + 6.0d0*(z1*4.0d0*
     +                       x(3*j-6)*x(3*j-6)+2.0d0*z2)
               H(3*j-8,3*j-7) = H(3*j-8,3*j-7) + 24.0d0*z1*
     +                       (x(3*j-8)-x(2))*(x(3*j-7)-x(3))
               H(3*j-8,3*j-6) = H(3*j-8,3*j-6) + 24.0d0*z1*
     +                       x(3*j-6)*(x(3*j-8)-x(2))
               H(3*j-7,3*j-6) = H(3*j-7,3*j-6) + 24.0d0*z1*
     +                       x(3*j-6)*(x(3*j-7)-x(3))
            else
               do ii=8,6,-1
                  H(3*i-ii,3*i-ii) = H(3*i-ii,3*i-ii) + 6.0d0*(z1*4.0d0
     +                               *(x(3*j-ii)-x(3*i-ii))*
     +                                (x(3*j-ii)-x(3*i-ii))+2.0d0*z2)
                  H(3*j-ii,3*j-ii) = H(3*j-ii,3*j-ii) + 6.0d0*(z1*4.0d0
     +                               *(x(3*j-ii)-x(3*i-ii))*
     +                                (x(3*j-ii)-x(3*i-ii))+2.0d0*z2)
                  H(3*i-ii,3*j-ii) = H(3*i-ii,3*j-ii) - 6.0d0*(z1*4.0d0
     +                               *(x(3*i-ii)-x(3*j-ii))*
     +                                (x(3*i-ii)-x(3*j-ii))+2.0d0*z2)
               enddo
               do ii=7,6,-1
                  H(3*j-8,3*j-ii) = H(3*j-8,3*j-ii) + 24.0d0*z1*
     +                   (x(3*j-8)-x(3*i-8))*(x(3*j-ii)-x(3*i-ii))
                  H(3*i-8,3*j-ii) = H(3*i-8,3*j-ii) + 24.0d0*z1*
     +                   (x(3*i-8)-x(3*j-8))*(x(3*j-ii)-x(3*i-ii))
                  H(3*i-8,3*i-ii) = H(3*i-8,3*i-ii) + 24.0d0*z1*
     +                   (x(3*i-8)-x(3*j-8))*(x(3*i-ii)-x(3*j-ii))
               enddo
               H(3*i-7,3*i-6) = H(3*i-7,3*i-6) + 24.0d0*z1*
     +                   (x(3*i-7)-x(3*j-7))*(x(3*i-6)-x(3*j-6))
               do ii=8,6,-2
                  H(3*i-7,3*j-ii) = H(3*i-7,3*j-ii) + 24.0d0*z1*
     +                    (x(3*i-7)-x(3*j-7))*(x(3*j-ii)-x(3*i-ii))
               enddo
               do ii=8,7,-1
                  H(3*i-6,3*j-ii) = H(3*i-6,3*j-ii) + 24.0d0*z1*
     +                    (x(3*i-6)-x(3*j-6))*(x(3*j-ii)-x(3*i-ii))
               enddo
               H(3*j-7,3*j-6) = H(3*j-7,3*j-6) + 24.0d0*z1*
     +                    (x(3*j-7)-x(3*i-7))*(x(3*j-6)-x(3*i-6))
            endif
         enddo
      enddo

      do i=2,n
         do j=1,i-1
            H(i,j) = H(j,i)
         enddo
      enddo

      return
      end


