/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0, July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

/* This is a test program that accompanies the report

   "Implementing functionals in HCL", Technical Report TR99-24,
   Department of Computational and Applied Mathematics, Rice University,
   PO Box 1892, Houston, TX 77251-1892.

The reader will find a postscript version of this report in
HCLROOT/doc/postscript/fcnl.ps, also accessible through the HCL
Tutorials page.

The purpose of this program is to exercise the functional class
LennardJonesEnergy, which is thoroughly discussed in the above report.
This class represents the Lennard-Jones potential energy function
as an HCL_Functional_d. For precise definition of this function, 
and references to its use in molecular dynamics, see the report.
*/

/*
The header file lj.h defines the classes implementing the
Lennard-Jones potential. Including it here gives the main program
access to these classes.  Examining it, you will see that it includes
the header files for various HCL classes needed in the definition
as well as standard C/C++ headers.

The optimization algorithm access class HCL_UMinMethod_d header
file includes many other files as well, and so makes the HCL 
unconstrained minimization package available to this program.
*/

#include "lj.h"
#include "HCL_UMinMethod_d.H"

/* Required for run-time check that two vector spaces have
   the same type. See Note 1 below for more on this. */

char HCL_RnSpace_d::Type;

/* This simple main program is structured to read the number of atoms
   in the Lennard-Jones model from the command line, as the second
   argument. */

int main( int argc,char ** argv )  {

/* to remind the user... */

  if( argc != 2 ) {
    cerr << "usage: main.x N" << endl;
    exit(1);
  }

/* The main constructor of the Lennard-Jones potential function 
   takes the number of atoms as its sole argument */

  int N = atoi( argv[1] );
  LennardJonesEnergy_d f( N );

/* Create two vectors in the domain of the operator. See Note 2 below for
   more on this construction. The memory is allocated "dynamically", from
   the heap, so is persistent beyond the scope of these statements and
   must be managed by the program. The HCL convention signalling creation
   of persistent storage that must be managed is to return a pointer, 
   rather than a reference or value. */
  
  HCL_Vector_d * x  = f.Domain().Member();
  HCL_Vector_d * dx = f.Domain().Member();
   
/* fill x with random numbers - this method uses the standard library
   random number generator, hence the need to reseed for each
   execution. */

  x->Random();

/* fill dx with random numbers, scale by 1/1000 */

  dx->Random();
  dx->Mul(1.0e-3);

/* The values returned by the CheckGrad method should lie on a 
   parabola, as they represent the difference between the computation
   of the Gradient() method and centered differences. See the
   reference manual for the menaing of the arguments. 
   */
     
  cout << "Checking gradient" << endl;
  f.CheckGrad( *x,*dx,cout );

/* HCL functions in general have Hessians as well. If the
   computation of the Hessian were to fail for any reason,
   the retuned pointer would be NULL, so checking its value 
   is a natural way to catch errors before using it.

   The pointer once again signals the creation of persistent (dynamic)
   storage, which must be managed.
*/

  HCL_GeneralMatOp_d * H = (HCL_GeneralMatOp_d*)f.Hessian( *x );
  if (H) {

    /* All HCL classes have a write method, which dumps to an
       output stream whatever information the programmer feels
       is useful. */

    cout <<endl<< "Hessian at x: " << endl;
    H->Write( cout );

    /* This main program has no more specific need for the Hessian
       after this. Since the Hessian object has been allocated
       dynamically, the program must explicitly deallocate the memory
       it uses. Instead of calling the standard C++ deallocator
       (``delete''), call the HCL deallocator HCL_delete to maintain
       reference counting. This standard programming practice helps
       prevent ``dangling pointers'' (which point to memory already
       deallocated) and ``memory leaks'' (in which memory fails to be
       deallocated when no longer needed).  */
  
    HCL_delete( H );
  }
  else {
    cerr<<"Error: main program of fcnl tutorial"<<endl;
    cerr<<"failed to generate Hessian operator"<<endl;
    exit(1);
  }

  /* Fill dx with random numbers, scale by 10e-2 */
  
  dx->Random();
  dx->Mul(1.0e-2);

  /* check the correctness of the Hessian calculation - similar to
     gradient check. */

  cout<<endl << "Checking Hessian"<< endl;
  f.CheckHess( *x,*dx,cout );

  /* Use the UMinMethod class to construct the object which
     encapsulates the unconstrained minimization algorithm. 
     UMinMethod provides an alternate construction of these
     objects, as opposed to the "usual" construction which
     takes two lines rather than one - see the op example.
     The constructor takes as arguments strings specifying 
     \begin{itemize}
     \item the choice of method: BFGS or NLCG or...
     \item the choice line search method (for first order methods): 
     DennisSchnabel or Fletcher or...
     \item the name of the parameter file.
     \end{itemize}
     The parameter file lists convergence parameters etc. For example
     the parameter DispFlag regulates the amount of stuff dumped out 
     on the screen. See Note 3 for more on this. */

  HCL_UMinMethod_d umin("BFGS","DennisSchnabel","par");

  /* UMin classes have one main method - Minimize, which takes
     two arguments:
     \begin{itemize}
     \item the function (reference to HCL_Functional_d, of which
     LennardJonesEnergy is a derived class; in C++, you can pass
     values (i.e. names of objects) where references are called 
     for in the parameter list)
     \item the initial guess, in which the final estimate will be
     returned - reference to an HCL_Vector_d.
     */

  umin.Minimize( f, *x );

  /* Note that the minimization is unlikely to have produced anything
     like the global minimum energy of the Lennard-Jones potential,
     since quasi-Newton methods like BFGS produce local minima and
     this function has LOTS of them. */

  /* clean up - again using the HCL deallocator. Not important at the
     end of a main program, since the end of the process will
     deallocate all of its memory, but essential especially in methods
     called repeatedly. Remember that x and dx are pointers to
     dynamically allocated memory, so the program must manage this
     memory explicitly. */

  HCL_delete(x);
  HCL_delete(dx);

}

/* Note 1. The only instance of explicit runtime typechecking built
   into HCL concerns vector spaces. Before accessing the particular data
   (dimensions, sample rates, other auxiliary quantities) that
   characterize a vector space, one must be sure that the access methods
   for these data are well defined. Since HCL vector spaces are often
   referenced as instances of HCL_VectorSpace, which generally defines
   none of these access methods, a cast is necessary to make the access
   methods available: you must declare that the space is not just an
   HCL_VectorSpace, but is actually the type of vector space which has
   the particular access methods which you want to use to generate data
   for comparison. To be sure that the cast will succeed, you must know
   that the vector space actually is of this type. The role of the static
   character variable Type is to facilitate this check: since it common
   to all objects of the class. Therefore its address is characteristic
   of the class, and can be checked against another such Type address to
   verify that two spaces are actually of the same type.  Static
   variables must be declared in the main block of each program, hence
   the presence of this declaration. 

   A future version of HCL, using the dynamic cast facility of the 
   ISO standard, may obviate the need for the Type trick. */

/* Note 2. We need two vectors in the domain of f with which to run
   various checks. The domain of f is a particular HCL_VectorSpace,
   namely an instance of HCL_RnVector_d. However the particular data
   structures and internal details of the domain are not required at
   this level - only features common to all functions, vectors,
   etc. appear in this main program. So we use a so-called virtual
   constructor, the Member() method of the vector space class, to
   construct the required vectors. This is possible because every HCL
   function knows its domain, and returns a reference to it with the
   Domain() method. HCL policy is to return pointers to objects
   allocated from the heap, to signal the programmer that the calling
   level of the program is responsible for managing this memory, and
   in particular for deallocating it when it's no longer
   needed. Member() allocates memory from the heap (``dynamically'' at
   runtime) rather than from the stack (``statically'' at compile
   time). Since the Member() function is virtual, the nature of the
   calling vector space object does not have to be specified in detail
   at this point: the compiler will resolve this call by going as far
   down the class hierarchy as possible. So secretly x and dx will
   have all features of the domain of f, even those features which
   HCL_Vectors do not in general share.  */

/* Note 3. You will note that in the parameter file par, the
   parameters for the UMin method (i.e. the quasiNewton update) are
   tagged with the prefix "UMin::", while those for the LineSearch
   algorithm are tagged with "LineSearch::". Thus the parameter file
   contains lines like

        LineSearch::MaxItn = 100
        UMin::MaxItn = 10

   The first indicates that the maximum number of iterations allowed
   in each invocation of the line search algorithm is 100, and the second
   indicates that the maximum number of iterations allowed in the
   execution of the minimization algorithm is 20.

   The meanings and default values for all HCL algorithm parameters
   are documented in the HCL reference manual pages.  */
