C     SUBROUTINE  assemble
C
C     PURPOSE
C       This subroutines assembles the stiffness matrix K and the
C       loading vector F for the finite element formulation of the
C       elliptic BVP
C
C              -div(a(x,y)u) = f(x,y) in Omega
C                         u  = 0 on boundary of Omega
C
C       Here Omega is a connected open (polygonal) set in the plane,
C       the coefficient a(x,y) is represented by a continuous piecewise
C       linear function, the forcing function f(x,y) is implemented in
C       the Fortran function FN below, and the approximate solution u(x,y)
C       is the Galerkin approximation from the space of continuous piecewise
C       linear functions satisfying the boundary conditions.

      subroutine assemble( M,N,L,NodeList,NodePtrs,ElList,
     +                     a,ik,jk,k,nzk,maxnzk,F )

C     ARGUMENTS

      implicit none
      integer M,N,L,nzk,maxnzk
      integer NodePtrs(N),ElList(L,3),ik(*),jk(*)
      double precision NodeList(M,4),a(M),k(*),F(N)

C     MEANING OF ARGUMENTS
C
C     INPUT
C     -----
C     M (integer)
C       number of nodes in mesh
C
C     N (integer)
C       number of "free" nodes in mesh (i.e. nodes not associated
C       with a Dirichlet condition)
C
C     L (integer)
C       number of triangles in mesh
C
C     NodeList (Mx4 array)
C       Each row contains the following information about the
C       corresponding node: x-coord., y-coord.,
C                          (if boundary, index of node in boundary vector,
C                           otherwise, zero),
C                          (if free, index of node in vector of unknowns,
C                           otherwise, zero).
C
C     NodePtrs (array of length N)
C       NodePtrs(i) is the index into NodeList of the ith free node.
C       So, for example, NodeList(NodePtrs(i),1:2) are the
C       coordinates of the ith free node.
C
C     ElList: (Lx3 array)
C       Each row corresponds to one element and contains
C       pointers to the nodes of the triangle in NodeList.
C
C     a (array of length M)
C       Nodal values of the piecewise linear function a(x,y) appearing
C       in the PDE.
C
C     maxnzk (integer)
C       The maximum allowable number of nonzeros in the sparse matrix K
C
C     OUTPUT
C     ------
C
C     ik (integer array)
C       Row pointers for the sparse matrix K in row-compressed storage
C       (K is the stiffness matrix in the finite element equation)
C
C     jk (integer array)
C       Column indices for the sparse matrix K in row-compressed storage
C
C     k (double precision array)
C       Nonzero elements of the sparse matrix K in row-compressed storage
C
C     nzk (integer)
C       The number of nonzeros in the sparse matrix K
C
C     F (array of length N)
C       The loading vector for the finite element equation


C     LOCAL VARIABLES

      double precision quad3,Fquad
      integer nn,kk(3),i,j,i1,i2,i3,ii
      double precision v

C     Main loop: adds the contribution from each element

      do i=1,N
         F(i) = 0.0d0
      enddo

      nzk = 0
      do nn=1,L

C        This element contributes to at most 6 entries in the upper
C        triangle of the finite element matrix

         i1 = ElList(nn,1)
         i2 = ElList(nn,2)
         i3 = ElList(nn,3)

C        Sort the pointers

         if (i1.gt.i2) then
            ii=i1
            i1=i2
            i2=ii
         endif
         if (i2.gt.i3) then
            ii=i2
            i2=i3
            i3=ii
         endif
         if (i1.gt.i2) then
            ii=i1
            i1=i2
            i2=ii
         endif

         kk(1) = NodeList(i1,4)
         kk(2) = NodeList(i2,4)
         kk(3) = NodeList(i3,4)

C        Add the contributions to the load vector

         do i=1,3
            if (kk(i).gt.0) then
               v = Fquad(NodeList(i1,1),NodeList(i1,2),
     +                      NodeList(i2,1),NodeList(i2,2),
     +                      NodeList(i3,1),NodeList(i3,2),i)
               F(kk(i)) = F(kk(i))+v
            endif
         enddo

C        Add the contributions to the stiffness matrix

         do i=1,3
            do j=i,3
               if (kk(i).gt.0 .and. kk(j).gt.0) then

                  v = quad3(NodeList(i1,1),NodeList(i1,2),
     +                      NodeList(i2,1),NodeList(i2,2),
     +                      NodeList(i3,1),NodeList(i3,2),
     +                      a(i1),a(i2),a(i3),i,j)
                  call add2symcrs(N,ik,jk,k,nzk,maxnzk,v,kk(i),kk(j))
               endif
            enddo
         enddo

      enddo

      return
      end

C     SUBROUTINE  pileassemble
C
C     PURPOSE
C       This subroutines assembles the stiffness matrix K and the
C       loading vector F for the finite element formulation of the
C       elliptic BVP (Planar Isotropic Linear Elasticity---PILE)
C
C              -div(sigma) = f(x,y) in Omega,
C                        sigma*u  = g on boundary of Omega,
C               sigma = 2*mu*epsilon+lambda*trace(epsilon)*I,
C               epsilon = (1/2)*(grad(epsilon)+grad(epsilon)^T)
C
C       Here Omega is a connected open (polygonal) set in the plane,
C       the coefficients mu(x,y) and lambda(x,y) are represented by
C       continuous piecewise linear function, the forcing function f(x,y)
C       is implemented in the Fortran function FN below, and the approximate
C       solution u(x,y) = (u_1(x,y),u_2(x,y)) is the Galerkin approximation
C       from the space of continuous piecewise linear functions.

      subroutine pileassemble( M,N,L,NodeList,NodePtrs,ElList,
     +                         mu,lam,ik,jk,k,nzk,maxnzk,F )

C     ARGUMENTS

      implicit none
      integer M,N,L,nzk,maxnzk
      integer NodePtrs(N),ElList(L,6),ik(*),jk(*)
      double precision NodeList(M,4),mu(M),lam(M),k(*),F(N)

C     MEANING OF ARGUMENTS
C
C     INPUT
C     -----
C     M (integer)
C       number of nodes in mesh
C
C     N (integer)
C       number of "free" nodes in mesh (N=M in this case)
C
C     L (integer)
C       number of triangles in mesh
C
C     NodeList (Mx4 array)
C       Each row contains the following information about the
C       corresponding node: x-coord., y-coord.,
C                          (if boundary, index of node in boundary vector,
C                           otherwise, zero),
C                          (if free, index of node in vector of unknowns,
C                           otherwise, zero).
C
C     NodePtrs (array of length N)
C       NodePtrs(i) is the index into NodeList of the ith free node.
C       So, for example, NodeList(NodePtrs(i),1:2) are the
C       coordinates of the ith free node.
C
C     ElList: (Lx6 array)
C       Each row corresponds to one element.  The first three entries are
C       pointers to the nodes of the triangle in NodeList.  Entries
C       4,5,6 correspond to the three edges and indicate whether
C       an edge is on the boundary (1) or not (0).  The flag for
C       the edge joining nodes i,j (i,j=1,2,3, i\neq j) is in entry
C       i+j+1.
C
C     mu (array of length M)
C       Nodal values of the piecewise linear function mu(x,y) appearing
C       in the PDE.
C
C     lam (array of length M)
C       Nodal values of the piecewise linear function lambda(x,y) appearing
C       in the PDE.
C
C     maxnzk (integer)
C       The maximum allowable number of nonzeros in the sparse matrix K
C
C     OUTPUT
C     ------
C
C     ik (integer array)
C       Row pointers for the sparse matrix K in row-compressed storage
C       (K is the stiffness matrix in the finite element equation)
C
C     jk (integer array)
C       Column indices for the sparse matrix K in row-compressed storage
C
C     k (double precision array)
C       Nonzero elements of the sparse matrix K in row-compressed storage
C
C     nzk (integer)
C       The number of nonzeros in the sparse matrix K
C
C     F (array of length 2N)
C       The loading vector for the finite element equation


C     LOCAL VARIABLES

      double precision quad3b,Gvquad
      integer nn,kk(3),i,j,i1,i2,i3,ii,ii1,ii2,kk1,kk2
      double precision v

C     Initialize the load vector and the stiffness matrix

      do i=1,2*N
         F(i) = 0.0d0
      enddo

      nzk = 0

C     Main loop: adds the contribution from each element

      do nn=1,L

C        Get the pointers to the vertices of this triangle

         i1 = ElList(nn,1)
         i2 = ElList(nn,2)
         i3 = ElList(nn,3)

C        write(*,*) 'Element ',nn
C         write(*,*) '   Vertices: ',i1,i2,i3

C        Sort the pointers

         if (i1.gt.i2) then
            ii=i1
            i1=i2
            i2=ii
         endif
         if (i2.gt.i3) then
            ii=i2
            i2=i3
            i3=ii
         endif
         if (i1.gt.i2) then
            ii=i1
            i1=i2
            i2=ii
         endif

C         write(*,*) '   Vertices (sorted): ',i1,i2,i3

C        Get the indices of the vertices in the list of free variables
         
         kk(1) = NodeList(i1,4)
         kk(2) = NodeList(i2,4)
         kk(3) = NodeList(i3,4)

C         write(*,*) '   Indices: ',kk(1),kk(2),kk(3)

C        Add the contributions to the stiffness matrix
C
C        This element contributes to at most 21 entries in the upper
C        triangle of the finite element matrix.  Note that the matrix
C        is a 2x2 block matrix; the (1,1) block is the "u_1,u_1" block,
C        the (1,2) block is the "u_1,u_2" block, and the (2,2) block is
C        the "u_2,u_2" block.

C        Handle the (1,1) block.  This element contributes to at most
C        6 entries in the upper triangle of the (1,1) block.

         do i=1,3
            do j=i,3
               if (kk(i).gt.0 .and. kk(j).gt.0) then

                  v = quad3b(NodeList(i1,1),NodeList(i1,2),
     &                       NodeList(i2,1),NodeList(i2,2),
     &                       NodeList(i3,1),NodeList(i3,2),
     &                       2*mu(i1)+lam(i1),2*mu(i2)+lam(i2),
     &                       2*mu(i3)+lam(i3),i,1,j,1)+
     &                quad3b(NodeList(i1,1),NodeList(i1,2),
     &                       NodeList(i2,1),NodeList(i2,2),
     &                       NodeList(i3,1),NodeList(i3,2),
     &                       mu(i1),mu(i2),mu(i3),i,2,j,2)
                  call add2symcrs(2*N,ik,jk,k,nzk,maxnzk,v,kk(i),kk(j))
               endif
            enddo
         enddo

C        Handle the (1,2) block.  This element contributes to at most
C        9 entries in the upper triangle of the (1,2) block.

         do i=1,3
            do j=1,3
               if (kk(i).gt.0 .and. kk(j).gt.0) then

                  v = quad3b(NodeList(i1,1),NodeList(i1,2),
     &                       NodeList(i2,1),NodeList(i2,2),
     &                       NodeList(i3,1),NodeList(i3,2),
     &                       mu(i1),mu(i2),mu(i3),i,2,j,1)+
     &                quad3b(NodeList(i1,1),NodeList(i1,2),
     &                       NodeList(i2,1),NodeList(i2,2),
     &                       NodeList(i3,1),NodeList(i3,2),
     &                       lam(i1),lam(i2),lam(i3),i,1,j,2)

C                  if (kk(i).eq.25 .and. kk(j).eq.25) then
C                     write(*,*) '      Adding ',v,' to K(25,146)'
C                  endif

                  call add2symcrs(2*N,ik,jk,k,nzk,maxnzk,v,
     &                            kk(i),N+kk(j))
               endif
            enddo
         enddo

C        Handle the (2,2) block.  This element contributes to at most
C        6 entries in the upper triangle of the (2,2) block.

         do i=1,3
            do j=i,3
               if (kk(i).gt.0 .and. kk(j).gt.0) then

                  v = quad3b(NodeList(i1,1),NodeList(i1,2),
     &                       NodeList(i2,1),NodeList(i2,2),
     &                       NodeList(i3,1),NodeList(i3,2),
     &                       mu(i1),mu(i2),mu(i3),i,1,j,1)+
     &                quad3b(NodeList(i1,1),NodeList(i1,2),
     &                       NodeList(i2,1),NodeList(i2,2),
     &                       NodeList(i3,1),NodeList(i3,2),
     &                       2*mu(i1)+lam(i1),2*mu(i2)+lam(i2),
     &                       2*mu(i3)+lam(i3),i,2,j,2)
                  call add2symcrs(2*N,ik,jk,k,nzk,maxnzk,v,
     &                            N+kk(i),N+kk(j))
               endif
            enddo
         enddo

C        Add the contributions to the load vector from the forcing function
C        in the PDE

CC         do i=1,3
CC            if (kk(i).gt.0) then

C              Handle the x (first) component of the forcing function

CC               v = Fvquad(1,NodeList(i1,1),NodeList(i1,2),
CC     &                      NodeList(i2,1),NodeList(i2,2),
CC     &                      NodeList(i3,1),NodeList(i3,2),i)

C               write(*,*) 'Contribution of forcing function:'
C               write(*,*) '   kk(i),v',kk(i),v

CC               F(kk(i)) = F(kk(i))+v

C              Handle the y (second) component of the forcing function

CC               v = Fvquad(2,NodeList(i1,1),NodeList(i1,2),
CC     &                      NodeList(i2,1),NodeList(i2,2),
CC     &                      NodeList(i3,1),NodeList(i3,2),i)

C               write(*,*) 'Contribution of forcing function:'
C               write(*,*) '   N+kk(i),v',N+kk(i),v

CC               F(N+kk(i)) = F(N+kk(i))+v

CC            endif
CC         enddo

C        Add the contributions to the load vector from the boundary traction

         do i=1,2
            do j=i+1,3

               if (ElList(nn,i+j+1).eq.1) then

                  ii1 = ElList(nn,i)
                  ii2 = ElList(nn,j)
                  kk1 = NodeList(ii1,4)
                  kk2 = NodeList(ii2,4)

                if (kk1.ne.0 .and. kk2.ne.0) then

C                  write(*,*) 'Edge joining vertices ',ii1,ii2,
C     &                       ' is on boundary'

C                 Handle the x (first) component of the traction

                  v = Gvquad(1,NodeList(ii1,1),NodeList(ii1,2),
     &                      NodeList(ii2,1),NodeList(ii2,2),1)

C                  write(*,*) ' x-component:'
C                  write(*,*) '   kk1,v',kk1,v

                  F(kk1) = F(kk1)+v

                  v = Gvquad(1,NodeList(ii1,1),NodeList(ii1,2),
     &                      NodeList(ii2,1),NodeList(ii2,2),2)

C                  write(*,*) '   kk2,v',kk2,v

                  F(kk2) = F(kk2)+v

C                 Handle the y (second) component of the traction

                  v = Gvquad(2,NodeList(ii1,1),NodeList(ii1,2),
     &                      NodeList(ii2,1),NodeList(ii2,2),1)

C                  write(*,*) ' y-component:'
C                  write(*,*) '   N+kk1,v',N+kk1,v

                  F(N+kk1) = F(N+kk1)+v
                  v = Gvquad(2,NodeList(ii1,1),NodeList(ii1,2),
     &                      NodeList(ii2,1),NodeList(ii2,2),2)

C                  write(*,*) '   N+kk2,v',N+kk2,v

                  F(N+kk2) = F(N+kk2)+v

                endif
               endif

            enddo
         enddo

      enddo

      return
      end

C     SUBROUTINE  assemble2
C
C     PURPOSE
C       This subroutines assembles the "partial adjoint" matrix B
C       for the finite element formulation of the elliptic BVP
C
C              -div(a(x,y)u) = f(x,y) in Omega
C                         u  = 0 on boundary of Omega
C
C       Here Omega is a connected open (polygonal) set in the plane,
C       the coefficient a(x,y) is represented by a continuous piecewise
C       linear function, the forcing function f(x,y) is implemented in
C       the Fortran function FN below, and the approximate solution u(x,y)
C       is the Galerkin approximation from the space of continuous piecewise
C       linear functions satisfying the boundary conditions.

      subroutine assemble2( M,N,L,NodeList,NodePtrs,ElList,
     +                     u,ib,jb,b,nzb,maxnzb )

C     ARGUMENTS

      implicit none
      integer M,N,L,nzb,maxnzb
      integer NodePtrs(N),ElList(L,3),ib(*),jb(*)
      double precision NodeList(M,4),u(N),b(*)

C     MEANING OF ARGUMENTS
C
C     INPUT
C     -----
C     M (integer)
C       number of nodes in mesh
C
C     N (integer)
C       number of "free" nodes in mesh (i.e. nodes not associated
C       with a Dirichlet condition)
C
C     L (integer)
C       number of triangles in mesh
C
C     NodeList (Mx4 array)
C       Each row contains the following information about the
C       corresponding node: x-coord., y-coord.,
C                          (1 if interior, 0 if boundary),
C                          (if free, index of node in vector of unknowns,
C                           otherwise, zero).
C
C     NodePtrs (array of length N)
C       NodePtrs(i) is the index into NodeList of the ith free node.
C       So, for example, NodeList(NodePtrs(i),1:2) are the
C       coordinates of the ith free node.
C
C     ElList: (Lx3 array)
C       Each row corresponds to one element and contains
C       pointers to the nodes of the triangle in NodeList.
C
C     u (array of length N)
C       Nodal values of the piecewise linear function u(x,y) appearing
C       in the PDE.
C
C     OUTPUT
C     ------
C
C     ib (integer array)
C       Row pointers for the sparse matrix B in row-compressed storage
C
C     jb (integer array)
C       Column indices for the sparse matrix B in row-compressed storage
C
C     b (double precision array)
C       Nonzero elements of the sparse matrix B in row-compressed storage
C
C     nzb (integer)
C       The number of nonzeros in the sparse matrix B
C
C     maxnzb (integer)
C       The maximum allowable number of nonzeros in the sparse matrix B

C     LOCAL VARIABLES

      double precision quad3a
      integer nn,kk(3),ll(3),i,j,ii
      double precision v,v1,v2,v3

C     Main loop: adds the contribution from each element

      nzb = 0
      do nn=1,L

C        This element contributes to at most 9 entries in the matrix

         kk(1) = ElList(nn,1)
         kk(2) = ElList(nn,2)
         kk(3) = ElList(nn,3)

C        Sort the pointers

         if (kk(1).gt.kk(2)) then
            ii=kk(1)
            kk(1)=kk(2)
            kk(2)=ii
         endif
         if (kk(2).gt.kk(3)) then
            ii=kk(2)
            kk(2)=kk(3)
            kk(3)=ii
         endif
         if (kk(1).gt.kk(2)) then
            ii=kk(1)
            kk(1)=kk(2)
            kk(2)=ii
         endif

         ll(1) = NodeList(kk(1),4)
         ll(2) = NodeList(kk(2),4)
         ll(3) = NodeList(kk(3),4)

         if (ll(1) .gt. 0) then
            v1 = u(ll(1))
         else
            v1 = 0.0
         endif
         if (ll(2) .gt. 0) then
            v2 = u(ll(2))
         else
            v2 = 0.0
         endif
         if (ll(3) .gt. 0) then
            v3 = u(ll(3))
         else
            v3 = 0.0
         endif

C        Add the contributions to the stiffness matrix

C         write(*,*) 'nn = ',nn
C         write(*,*) 'kk = ',kk(1),kk(2),kk(3)
C         write(*,*) 'll = ',ll(1),ll(2),ll(3)
         do i=1,3
            do j=1,3
               if (ll(i).gt.0) then

C                  write(*,*) '   i,j = ',i,j
                  v = quad3a(NodeList(kk(1),1),NodeList(kk(1),2),
     +                      NodeList(kk(2),1),NodeList(kk(2),2),
     +                      NodeList(kk(3),1),NodeList(kk(3),2),
     +                      j,v1,v2,v3,i)
C                  write(*,*) '   Adding ',v,' to B(',ll(i),
C     +                       ',',kk(j),')'
            call add2nonsymcrs(N,M,ib,jb,b,nzb,maxnzb,v,ll(i),kk(j))
               endif
            enddo
         enddo

      enddo

      return
      end

C     SUBROUTINE  add2symcrs
C
C     PURPOSE
C       This subroutine adds to an entry in a symmetric sparse matrix stored in
C       Compressed-Row Storage.  That is, it replaces A(k,l) by A(k,l)+x.
C       If A(k,l) is zero upon input, the routine creates a new nonzero
C       entry in the matrix (provided X .NE. 0.0).

      subroutine add2symcrs(n,iA,jA,A,nzA,maxnzA,x,k,l)

C     ARGUMENTS

      integer n,nzA,k,l,maxnzA
      integer iA(n+1),jA(*)
      double precision x
      double precision A(*)

C
C     INPUT
C     -----
C
C     n (integer)
C       Number of columns (and rows) in A
C
C     maxnzA (integer)
C       The maximum allowable number of nonzeros in the sparse matrix A
C
C     x (double precision)
C       Value to be added to A(k,l)
C
C     i (integer)
C       row index of entry to be changed
C
C     j (integer)
C       column index of entry to be changed
C
C
C     INPUT/OUTPUT ARGUMENTS
C     ----------------------
C
C     iA (integer array)
C       Row pointers for the sparse matrix A in row-compressed storage
C
C     jA (integer array)
C       Column indices for the sparse matrix A in row-compressed storage
C
C     A (double precision array)
C       Nonzero elements of the sparse matrix A in row-compressed storage
C
C     nzA (integer)
C       The number of nonzeros in the sparse matrix A
C

C     LOCAL VARIABLES

      integer i,p

      if ( k .lt. 1 .or. k .gt. n .or. l .lt. 1 .or. l .gt. n ) then
         write(*,*) 'Error in add2symcrs: invalid indices: ',k,l
         write(*,*) 'Matrix has size ',n
         stop
      endif

      if (x .eq. 0.0) then
         return
      endif

      if (nzA .eq. 0) then
         do i=1,n+1
            iA(i) = 1
         enddo
      endif


      if (iA(k) .eq. nzA+1 .or. 
     +    (iA(k+1).eq.nzA+1 .and. ja(nzA).lt.l)) then
         if (nzA.eq.maxnzA) then
            write(*,*) 'Error in add2symcrs: maxnzA reached'
            write(*,*) 'Adding to entry (',k,',',l,')'
            stop
         endif
         A(nzA+1) = x
         jA(nzA+1) = l
         do i=k+1,n+1
            iA(i) = iA(i)+1
         enddo
         nzA = nzA+1
      elseif (iA(k+1) .eq. iA(k)) then
         if (nzA .eq. maxnzA) then
            write(*,*) 'Error in add2symcrs: maxnzA reached'
            write(*,*) 'Adding to entry (',k,',',l,')'
            stop
         endif
         do i=nza+1,ia(k)+1,-1
            A(i) = A(i-1)
            jA(i) = jA(i-1)
         enddo
         A(iA(k)) = x
         ja(iA(k)) = l
         do i=k+1,n+1
            iA(i) = iA(i)+1
         enddo
         nzA = nzA+1
      else
         p = 0
         do i=iA(k),iA(k+1)-2
            if (jA(iA(k)+p).lt.l) then
               p = p+1
            endif
         enddo
         if (jA(iA(k)+p).eq.l) then
            A(iA(k)+p) = A(iA(k)+p)+x
         elseif (jA(iA(k)+p).lt.l) then
            if (nzA .eq. maxnzA) then
               write(*,*) 'Error in add2symcrs: maxnzA reached'
               write(*,*) 'Adding to entry (',k,',',l,')'
               stop
            endif
            do i=nzA+1,iA(k)+p+2,-1
               A(i) = A(i-1)
               jA(i) = jA(i-1)
            enddo
            A(iA(k)+p+1) = x
            jA(iA(k)+p+1) = l
            do i=k+1,n+1
               iA(i) = iA(i)+1
            enddo
            nzA = nzA+1
         else
            if (nzA .eq. maxnzA) then
               write(*,*) 'Error in add2symcrs: maxnzA reached'
               write(*,*) 'Adding to entry (',k,',',l,')'
               stop
            endif
            do i=nzA+1,iA(k)+p+1,-1
               A(i) = A(i-1)
               jA(i) = jA(i-1)
            enddo
            A(iA(k)+p) = x
            jA(iA(k)+p) = l
            do i=k+1,n+1
               iA(i) = iA(i)+1
            enddo
            nzA = nzA+1
         endif
      endif

      return
      end

C     SUBROUTINE  add2nonsymcrs
C
C     PURPOSE
C       This subroutine adds to an entry in a nonsymmetric sparse matrix
C       stored in Compressed-Row Storage.  That is, it replaces A(k,l) by
C       A(k,l)+x.  If A(k,l) is zero upon input, the routine creates a new
C       nonzero entry in the matrix (provided X .NE. 0.0).

      subroutine add2nonsymcrs(m,n,iA,jA,A,nzA,maxnzA,x,k,l)

C     ARGUMENTS

      integer m,n,nzA,k,l,maxnzA
      integer iA(n+1),jA(*)
      double precision x
      double precision A(*)

C
C     INPUT
C     -----
C
C     n (integer)
C       Number of columns (and rows) in A
C
C     maxnzA (integer)
C       The maximum allowable number of nonzeros in the sparse matrix A
C
C     x (double precision)
C       Value to be added to A(k,l)
C
C     i (integer)
C       row index of entry to be changed
C
C     j (integer)
C       column index of entry to be changed
C
C
C     INPUT/OUTPUT ARGUMENTS
C     ----------------------
C
C     iA (integer array)
C       Row pointers for the sparse matrix A in row-compressed storage
C
C     jA (integer array)
C       Column indices for the sparse matrix A in row-compressed storage
C
C     A (double precision array)
C       Nonzero elements of the sparse matrix A in row-compressed storage
C
C     nzA (integer)
C       The number of nonzeros in the sparse matrix A
C

C     LOCAL VARIABLES

      integer i,p

      if ( k .lt. 1 .or. k .gt. m .or. l .lt. 1 .or. l .gt. n ) then
         write(*,*) 'Error in add2nonsymcrs: invalid indices: ',k,l
         write(*,*) 'Matrix has size ',m,n
         stop
      endif

      if (x .eq. 0.0) then
         return
      endif

      if (nzA .eq. 0) then
         do i=1,m+1
            iA(i) = 1
         enddo
      endif


      if (iA(k) .eq. nzA+1 .or. 
     +    (iA(k+1).eq.nzA+1 .and. ja(nzA).lt.l)) then
         if (nzA.eq.maxnzA) then
            write(*,*) 'Error in add2nonsymcrs: maxnzA reached'
            write(*,*) 'Adding to entry (',k,',',l,')'
            stop
         endif
         A(nzA+1) = x
         jA(nzA+1) = l
         do i=k+1,m+1
            iA(i) = iA(i)+1
         enddo
         nzA = nzA+1
      elseif (iA(k+1) .eq. iA(k)) then
         if (nzA .eq. maxnzA) then
            write(*,*) 'Error in add2nonsymcrs: maxnzA reached'
            write(*,*) 'Adding to entry (',k,',',l,')'
            stop
         endif
         do i=nza+1,ia(k)+1,-1
            A(i) = A(i-1)
            jA(i) = jA(i-1)
         enddo
         A(iA(k)) = x
         ja(iA(k)) = l
         do i=k+1,m+1
            iA(i) = iA(i)+1
         enddo
         nzA = nzA+1
      else
         p = 0
         do i=iA(k),iA(k+1)-2
            if (jA(iA(k)+p).lt.l) then
               p = p+1
            endif
         enddo
         if (jA(iA(k)+p).eq.l) then
            A(iA(k)+p) = A(iA(k)+p)+x
         elseif (jA(iA(k)+p).lt.l) then
            if (nzA .eq. maxnzA) then
               write(*,*) 'Error in add2nonsymcrs: maxnzA reached'
               write(*,*) 'Adding to entry (',k,',',l,')'
               stop
            endif
            do i=nzA+1,iA(k)+p+2,-1
               A(i) = A(i-1)
               jA(i) = jA(i-1)
            enddo
            A(iA(k)+p+1) = x
            jA(iA(k)+p+1) = l
            do i=k+1,m+1
               iA(i) = iA(i)+1
            enddo
            nzA = nzA+1
         else
            if (nzA .eq. maxnzA) then
               write(*,*) 'Error in add2nonsymcrs: maxnzA reached'
               write(*,*) 'Adding to entry (',k,',',l,')'
               stop
            endif
            do i=nzA+1,iA(k)+p+1,-1
               A(i) = A(i-1)
               jA(i) = jA(i-1)
            enddo
            A(iA(k)+p) = x
            jA(iA(k)+p) = l
            do i=k+1,m+1
               iA(i) = iA(i)+1
            enddo
            nzA = nzA+1
         endif
      endif

      return
      end

C     FUNCTION   Fquad
C
C     PURPOSE
C       Computes the integral over a triangle of the forcing function
C       FN times a piecewise linear basis function.
C
      double precision function Fquad(x1,y1,x2,y2,x3,y3,i)

C     ARGUMENTS

      double precision x1,y1,x2,y2,x3,y3
      integer i

C     INPUT
C
C     x1,y1 (double precision)
C       Coordinates of the first vertex of the triangle
C
C     x2,y2 (double precision)
C       Coordinates of the second vertex of the triangle
C
C     x3,y3 (double precision)
C       Coordinates of the third vertex of the triangle
C
C     i (integer)
C       Local index of basis function (i = 1, 2, or 3)
C       Basis function is 1 at the ith node and 0 at the other two

C     LOCAL VARIABLES

      double precision x13,x23,y13,y23,J
      double precision u1,u2,u3,v1,v2,v3
      double precision l1,l2,l3

      double precision FN
      double precision twothirds,onesixth

      twothirds = 2.0d0/3.0d0
      onesixth = 1.0d0/6.0d0

C     Compute the Jacobian

      x13 = x1-x3
      x23 = x2-x3
      y13 = y1-y3
      y23 = y2-y3
      J = abs(x13*y23-y13*x23)

C     Compute the quadrature nodes (on the reference triangle with
C     vertices (1,0), (0,1), (0,0), the nodes are (2/3,1/6),
C     (1/6,2/3), (1/6,1/6))

      u1 = x3+twothirds*x13+onesixth*x23
      v1 = y3+twothirds*y13+onesixth*y23
      u2 = x3+onesixth*x13+twothirds*x23
      v2 = y3+onesixth*y13+twothirds*y23
      u3 = x3+onesixth*x13+onesixth*x23
      v3 = y3+onesixth*y13+onesixth*y23

C     Get the values of the linear function at the quadrature nodes

      if (i.eq.1) then
         l1 = twothirds
         l2 = onesixth
         l3 = onesixth
      elseif (i.eq.2) then
         l1 = onesixth
         l2 = twothirds
         l3 = onesixth
      elseif (i.eq.3) then
         l1 = onesixth
         l2 = onesixth
         l3 = twothirds
      endif

C     Estimate the integral

      Fquad = J*onesixth*(FN(x1,y1)*l1+FN(x2,y2)*l2+FN(x3,y3)*l3)

      return
      end

C     FUNCTION   quad3
C
C     PURPOSE
C       Computes the integral over a triangle of the product of a
C       linear function times the inner product of the gradients of
C       two basis functions.
C
      double precision function quad3(x1,y1,x2,y2,x3,y3,
     +                                a1,a2,a3,j,k)

C     ARGUMENTS

      double precision x1,y1,x2,y2,x3,y3,a1,a2,a3
      integer j,k

C     INPUT
C
C     x1,y1 (double precision)
C       Coordinates of the first vertex of the triangle
C
C     x2,y2 (double precision)
C       Coordinates of the second vertex of the triangle
C
C     x3,y3 (double precision)
C       Coordinates of the third vertex of the triangle
C
C     a1,a2,a3 (double precision)
C       nodal values of the linear function
C
C     j,k (integers)
C       Local indices of basis functions (j,k = 1, 2, or 3)
C       Basis function is 1 at the ith node and 0 at the other two
C       The integrand is the linear function times the dot product
C       of the gradients of the jth and kth basis functions.

C     LOCAL VARIABLES

      double precision x13,x23,y13,y23,Jac
      double precision u1,u2,u3,v1,v2,v3
      double precision l1,l2,l3
      double precision A(3,3),b(3,2),fac
      integer ipiv(3),info
      double precision twothirds,onesixth

      twothirds = 2.0d0/3.0d0
      onesixth = 1.0d0/6.0d0

C     Compute the Jacobian

      x13 = x1-x3
      x23 = x2-x3
      y13 = y1-y3
      y23 = y2-y3
      Jac = abs(x13*y23-y13*x23)

C     Compute the quadrature nodes (on the reference triangle with
C     vertices (1,0), (0,1), (0,0), the nodes are (2/3,1/6),
C     (1/6,2/3), (1/6,1/6))

      u1 = x3+twothirds*x13+onesixth*x23
      v1 = y3+twothirds*y13+onesixth*y23
      u2 = x3+onesixth*x13+twothirds*x23
      v2 = y3+onesixth*y13+twothirds*y23
      u3 = x3+onesixth*x13+onesixth*x23
      v3 = y3+onesixth*y13+onesixth*y23

C     Get the values of the linear function at the quadrature nodes


      l1 = twothirds*a1+a2*onesixth+a3*onesixth
      l2 = a1*onesixth+twothirds*a2+a3*onesixth
      l3 = a1*onesixth+a2*onesixth+twothirds*a3


C     Get the gradients of the jth and kth basis functions

      A(1,1)=1.0d0 
      A(2,1)=1.0d0
      A(3,1)=1.0d0
      A(1,2)=x1
      A(2,2)=x2
      A(3,2)=x3
      A(1,3)=y1
      A(2,3)=y2
      A(3,3)=y3

      if (j.eq.1) then
         b(1,1)=1.0d0
         b(2,1)=0.0d0
         b(3,1)=0.0d0
      elseif (j.eq.2) then
         b(1,1)=0.0d0
         b(2,1)=1.0d0
         b(3,1)=0.0d0
      elseif (j.eq.3) then
         b(1,1)=0.0d0
         b(2,1)=0.0d0
         b(3,1)=1.0d0
      endif
      if (k.eq.1) then
         b(1,2)=1.0d0
         b(2,2)=0.0d0
         b(3,2)=0.0d0
      elseif (k.eq.2) then
         b(1,2)=0.0d0
         b(2,2)=1.0d0
         b(3,2)=0.0d0
      elseif (k.eq.3) then
         b(1,2)=0.0d0
         b(2,2)=0.0d0
         b(3,2)=1.0d0
      endif

      call dgesv(3,2,A,3,ipiv,b,3,info)

C     Compute the dot product of the gradients

      fac = b(2,1)*b(2,2)+b(3,1)*b(3,2)

C     Estimate the integral

      quad3 = fac*Jac*onesixth*(l1+l2+l3)

      return
      end

C     FUNCTION   quad3a
C
C     PURPOSE
C       Computes the integral over a triangle of the product of a
C       piecewise linear basis function times the inner product of the
C       gradients of a linear function times a basis function.

      double precision function quad3a(x1,y1,x2,y2,x3,y3,j,
     +                                a1,a2,a3,k)

C     ARGUMENTS

      double precision x1,y1,x2,y2,x3,y3,a1,a2,a3
      integer j,k

C     INPUT
C
C     x1,y1 (double precision)
C       Coordinates of the first vertex of the triangle
C
C     x2,y2 (double precision)
C       Coordinates of the second vertex of the triangle
C
C     x3,y3 (double precision)
C       Coordinates of the third vertex of the triangle
C
C     a1,a2,a3 (double precision)
C       nodal values of the linear function
C
C     j,k (integers)
C       Local indices of basis functions (j,k = 1, 2, or 3)
C       Basis function is 1 at the ith node and 0 at the other two
C       The integrand is the linear function times the dot product
C       of the gradients of the jth and kth basis functions.

C     LOCAL VARIABLES

      double precision x13,x23,y13,y23,Jac
      double precision l1,l2,l3
      double precision A(3,3),b(3,2),fac
      integer ipiv(3),info
      double precision twothirds,onesixth

      twothirds = 2.0d0/3.0d0
      onesixth = 1.0d0/6.0d0

C     Compute the Jacobian

      x13 = x1-x3
      x23 = x2-x3
      y13 = y1-y3
      y23 = y2-y3
      Jac = abs(x13*y23-y13*x23)

C     Get the values of the first basis function at the quadrature nodes

      if (j.eq.1) then
         l1 = twothirds
         l2 = onesixth
         l3 = onesixth
      elseif (j.eq.2) then
         l1 = onesixth
         l2 = twothirds
         l3 = onesixth
      elseif (j.eq.3) then
         l1 = onesixth
         l2 = onesixth
         l3 = twothirds
      endif

C     Get the gradients of the linear function and kth basis function

      A(1,1)=1.0d0 
      A(2,1)=1.0d0
      A(3,1)=1.0d0
      A(1,2)=x1
      A(2,2)=x2
      A(3,2)=x3
      A(1,3)=y1
      A(2,3)=y2
      A(3,3)=y3

      b(1,1) = a1
      b(2,1) = a2
      b(3,1) = a3
      if (k.eq.1) then
         b(1,2)=1.0d0
         b(2,2)=0.0d0
         b(3,2)=0.0d0
      elseif (k.eq.2) then
         b(1,2)=0.0d0
         b(2,2)=1.0d0
         b(3,2)=0.0d0
      elseif (k.eq.3) then
         b(1,2)=0.0d0
         b(2,2)=0.0d0
         b(3,2)=1.0d0
      endif

      call dgesv(3,2,A,3,ipiv,b,3,info)

C     Compute the dot product of the gradients

      fac = b(2,1)*b(2,2)+b(3,1)*b(3,2)

C     Estimate the integral

      quad3a = fac*Jac*onesixth*(l1+l2+l3)

      return
      end

C     FUNCTION   quad3b
C
C     PURPOSE
C       Computes the integral over a triangle of a linear function
C       times the product of partial derivatives of two basis functions.
C
      double precision function quad3b(x1,y1,x2,y2,x3,y3,
     +                                a1,a2,a3,j,dj,k,dk)

C     ARGUMENTS

      double precision x1,y1,x2,y2,x3,y3,a1,a2,a3
      integer j,k,dj,dk

C     INPUT
C
C     x1,y1 (double precision)
C       Coordinates of the first vertex of the triangle
C
C     x2,y2 (double precision)
C       Coordinates of the second vertex of the triangle
C
C     x3,y3 (double precision)
C       Coordinates of the third vertex of the triangle
C
C     a1,a2,a3 (double precision)
C       nodal values of the linear function
C
C     j,k (integers)
C       Local indices of basis functions (j,k = 1, 2, or 3)
C       Basis function i is 1 at the ith node and 0 at the other two
C
C     dj,dk (integers)
C       Index of partial derivatives for basis functions j,k
C       (dj,dk=1,2).
C
C       The integrand is the linear function times the partial derivative
C       of the jth basis function w.r.t the dj'th variable times the partial
C       derivative of the kth basis function w.r.t the dk'th variable.


C     LOCAL VARIABLES

      double precision x13,x23,y13,y23,Jac
      double precision u1,u2,u3,v1,v2,v3
      double precision l1,l2,l3
      double precision A(3,3),b(3,2),fac
      integer ipiv(3),info
      double precision twothirds,onesixth

      twothirds = 2.0d0/3.0d0
      onesixth = 1.0d0/6.0d0

C     Compute the Jacobian

      x13 = x1-x3
      x23 = x2-x3
      y13 = y1-y3
      y23 = y2-y3
      Jac = abs(x13*y23-y13*x23)

C     Compute the quadrature nodes (on the reference triangle with
C     vertices (1,0), (0,1), (0,0), the nodes are (2/3,1/6),
C     (1/6,2/3), (1/6,1/6))

      u1 = x3+twothirds*x13+onesixth*x23
      v1 = y3+twothirds*y13+onesixth*y23
      u2 = x3+onesixth*x13+twothirds*x23
      v2 = y3+onesixth*y13+twothirds*y23
      u3 = x3+onesixth*x13+onesixth*x23
      v3 = y3+onesixth*y13+onesixth*y23

C     Get the values of the linear function at the quadrature nodes


      l1 = twothirds*a1+a2*onesixth+a3*onesixth
      l2 = a1*onesixth+twothirds*a2+a3*onesixth
      l3 = a1*onesixth+a2*onesixth+twothirds*a3

C     Get the gradients of the jth and kth basis functions

      A(1,1)=1.0d0 
      A(2,1)=1.0d0
      A(3,1)=1.0d0
      A(1,2)=x1
      A(2,2)=x2
      A(3,2)=x3
      A(1,3)=y1
      A(2,3)=y2
      A(3,3)=y3

      if (j.eq.1) then
         b(1,1)=1.0d0
         b(2,1)=0.0d0
         b(3,1)=0.0d0
      elseif (j.eq.2) then
         b(1,1)=0.0d0
         b(2,1)=1.0d0
         b(3,1)=0.0d0
      elseif (j.eq.3) then
         b(1,1)=0.0d0
         b(2,1)=0.0d0
         b(3,1)=1.0d0
      endif
      if (k.eq.1) then
         b(1,2)=1.0d0
         b(2,2)=0.0d0
         b(3,2)=0.0d0
      elseif (k.eq.2) then
         b(1,2)=0.0d0
         b(2,2)=1.0d0
         b(3,2)=0.0d0
      elseif (k.eq.3) then
         b(1,2)=0.0d0
         b(2,2)=0.0d0
         b(3,2)=1.0d0
      endif

      call dgesv(3,2,A,3,ipiv,b,3,info)
C      write(*,*) 'Output of dgesv:'
C      do ii=1,3
C         write(*,*) (b(ii,jj),jj=1,2)
C      enddo
C      write(*,*) 'j,k: ',j,k
C      write(*,*) 'dj,dk: ',dj,dk
C      write(*,*) 'b(dj+1,1),b(dk+1,2): ',b(dj+1,1),b(dk+1,2)

C     Compute the product of the partial derivatives

      fac = b(dj+1,1)*b(dk+1,2)

C     Estimate the integral

C      write(*,*) '         In quad3b: fac,Jac,l1,l2,l3 ',
C     &            fac,Jac,l1,l2,l3
      quad3b = fac*Jac*onesixth*(l1+l2+l3)

      return
      end

C     FUNCTION   Fvquad
C
C     PURPOSE
C       Computes the integral over a triangle of a component of a
C       (vector-valued) forcing function FNv times a piecewise linear
C       basis function.
C
      double precision function Fvquad(k,x1,y1,x2,y2,x3,y3,i)

C     ARGUMENTS

      double precision x1,y1,x2,y2,x3,y3
      integer i,k

C     INPUT
C
C     k (integer)
C       Indicates which component of f (k=1,2)
C
C     x1,y1 (double precision)
C       Coordinates of the first vertex of the triangle
C
C     x2,y2 (double precision)
C       Coordinates of the second vertex of the triangle
C
C     x3,y3 (double precision)
C       Coordinates of the third vertex of the triangle
C
C     i (integer)
C       Local index of basis function (i = 1, 2, or 3)
C       Basis function is 1 at the ith node and 0 at the other two

C     LOCAL VARIABLES

      double precision x13,x23,y13,y23,J
      double precision u1,u2,u3,v1,v2,v3
      double precision l1,l2,l3

      double precision FNv
      double precision twothirds,onesixth

      twothirds = 2.0d0/3.0d0
      onesixth = 1.0d0/6.0d0

C     Compute the Jacobian

      x13 = x1-x3
      x23 = x2-x3
      y13 = y1-y3
      y23 = y2-y3
      J = abs(x13*y23-y13*x23)

C     Compute the quadrature nodes (on the reference triangle with
C     vertices (1,0), (0,1), (0,0), the nodes are (2/3,1/6),
C     (1/6,2/3), (1/6,1/6))

      u1 = x3+twothirds*x13+onesixth*x23
      v1 = y3+twothirds*y13+onesixth*y23
      u2 = x3+onesixth*x13+twothirds*x23
      v2 = y3+onesixth*y13+twothirds*y23
      u3 = x3+onesixth*x13+onesixth*x23
      v3 = y3+onesixth*y13+onesixth*y23

C     Get the values of the linear function at the quadrature nodes

      if (i.eq.1) then
         l1 = twothirds
         l2 = onesixth
         l3 = onesixth
      elseif (i.eq.2) then
         l1 = onesixth
         l2 = twothirds
         l3 = onesixth
      elseif (i.eq.3) then
         l1 = onesixth
         l2 = onesixth
         l3 = twothirds
      endif

C     Estimate the integral

      Fvquad = J*onesixth*(FNv(k,x1,y1)*l1+FNv(k,x2,y2)*l2+
     &                     FNv(k,x3,y3)*l3)

      return
      end

C     FUNCTION   Gvquad
C
C     PURPOSE
C       Computes the integral over a line segment of a component of a
C       (vector-valued) function GNv times a piecewise linear
C       basis function.
C
      double precision function Gvquad(k,x1,y1,x2,y2,i)

C     ARGUMENTS

      double precision x1,y1,x2,y2
      integer i,k

C     INPUT
C
C     k (integer)
C       Indicates which component of g (k=1,2)
C
C     x1,y1 (double precision)
C       Coordinates of the first endpoint of the segment
C
C     x2,y2 (double precision)
C       Coordinates of the second endpoint of the segment
C
C     i (integer)
C       Local index of basis function (i = 1, 2)
C       Basis function is 1 at the ith node and 0 at the other.

C     LOCAL VARIABLES

      double precision v1,v2,s,t,t1,t2,w1,z1,w2,z2,p1,p2

      double precision GNv

C     Compute the vector in the direction of the line segment
C     and its length

      v1 = x2-x1
      v2 = y2-y1
      s = sqrt(v1*v1+v2*v2)

C     Compute the quadrature nodes in [0,1]

      t = 1.0/sqrt(3.0)
      t1 = 0.5*(1.0-t)
      t2 = 0.5*(1.0+t)

C     Compute the quadrature nodes on the segment

      w1 = x1+v1*t1
      z1 = y1+v2*t1
      w2 = x1+v1*t2
      z2 = y1+v2*t2

C     Compute the values of the linear function

      if (i.eq.1) then
         p1 = t2
         p2 = t1
      else
         p1 = t1
         p2 = t2
      endif

C     Estimate the integral

      Gvquad = 0.5*s*(GNv(k,w1,z1)*p1+GNv(k,w2,z2)*p2)

      return
      end

C FUNCTION FN

      double precision function FN(x,y)

      double precision x,y

      FN = 1.0d0

      return
      end

C Test Problem 1: u_1(x,y)=x+2xy, u_2(x,y)=y^2
C                 mu=2, lambda=1
C                 f(x,y)=(0,-16)
C                 g(x,y):
C                    lower boundary: (-4x,-1)
C                    right boundary: (5+12y,4)
C                    upper boundary: (4x,13)
C                    left boundary: (-5-12y,0)
C
C                 Omega = unit square

C FUNCTION FNv

      double precision function FNv(i,x,y)

      double precision x,y,pi
      integer i

      pi = 3.14159265358979d0
      if (i.eq.1) then

C        Test Problem 1
C         FNv = 0.0d0
C        Test Problem 2
C         FNv = 0.0d0
C        Test Problem 3
C         FNv = 10.0d0*y*(1.0d0-y)+4.0d0*x*(1.0d0-x)-3.0d0*pi*pi*
C     &         cos(pi*x)*cos(pi*y)
C        Test Problem 4
C         FNv = -3.0d0*exp(x)-2.0d0*(1.0d0+x)*exp(x)-(x+y)*exp(x)-1.0d0
C        Test Problem I
         FNv = 0.0d0

      else

C        Test Problem 1
C         FNv = -16.0d0
C        Test Problem 2
C         FNv = -12.0d0
C        Test Problem 3
C         FNv = -3.0d0*(1.0d0-2.0d0*x)*(1.0d0-2.0d0*y)+7.0d0*pi*pi*
C     &          sin(pi*x)*sin(pi*y)
C        Test Problem 4
C         FNv = -exp(x)-2.0d0
C        Test Problem I
         FNv = 0.0d0

      endif

      return
      end

C FUNCTION GNv

      double precision function GNv(i,x,y)

      double precision x,y,e
      integer i

      e = exp(1.0d0)
      if (i.eq.1) then

      if (.false.) then
C        Test Problem 1
         if (fabs(y).lt.1.0d-14) then
            GNv = -4.0d0*x
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 5.0d0+12.0d0*y
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 4.0d0*x
         elseif (fabs(x).lt.1.0e-14) then
            GNv = -(5.0d0+12.0d0*y)
         else
           write(*,*) 'Invalid point in GNv: ',x,y
         endif
      elseif (.false.) then
C        Test Problem 2
         if (fabs(y).lt.1.0d-14) then
            GNv = -4.0d0*x
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 4.0d0+8.0d0*y
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 4.0d0*x
         elseif (fabs(x).lt.1.0e-14) then
            GNv = -(4.0d0+8.0d0*y)
         else
            write(*,*) 'Invalid point in GNv: ',x,y
         endif
      elseif (.false.) then
C        Test Problem 4
         if (fabs(y).lt.1.0d-14) then
            GNv = -1.0d0-x
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 4.0d0*e+(1.0d0+e)*(y+1.0d0)
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 1.0d0+x
         elseif (fabs(x).lt.1.0e-14) then
            GNv = -2.0d0-2.0d0*y
         else
            write(*,*) 'Invalid point in GNv: ',x,y
         endif
      elseif (.true.) then
C        Test Problem I
         if (fabs(y).lt.1.0d-14) then
            GNv = -1.0d0
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 1.0d0
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 1.0d0
         elseif (fabs(x).lt.1.0e-14) then
            GNv = -1.0d0
         else
            write(*,*) 'Invalid point in GNv: ',x,y
         endif
      endif

      elseif (i.eq.2) then

      if (.false.) then
C        Test Problem 1
         if (fabs(y).lt.1.0d-14) then
            GNv = -1.0d0
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 4.0d0
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 13.0d0
         elseif (fabs(x).lt.1.0e-14) then
            GNv = 0.0d0
         else
            write(*,*) 'Invalid point in GNv: ',x,y
         endif
      elseif (.false.) then
C        Test Problem 2
         if (fabs(y).lt.1.0d-14) then
            GNv = 0.0d0
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 4.0d0
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 8.0d0
         elseif (fabs(x).lt.1.0e-14) then
            GNv = 0.0d0
         else
            write(*,*) 'Invalid point in GNv: ',x,y
         endif
      elseif (.false.) then
C        Test Problem 4
         if (fabs(y).lt.1.0d-14) then
            GNv = -2.0d0*(1.0d0+x)-x*(exp(x)+1.0d0)
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 2.0d0
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 2.0d0*(1.0d0+x)+(x+1.0d0)*(exp(x)+1.0d0)
         elseif (fabs(x).lt.1.0e-14) then
            GNv = -1.0d0
         else
            write(*,*) 'Invalid point in GNv: ',x,y
         endif
      elseif (.true.) then
C        Test Problem I
         if (fabs(y).lt.1.0d-14) then
            GNv = -1.0d0
         elseif (fabs(x-1.0d0).lt.1.0e-14) then
            GNv = 1.0d0
         elseif (fabs(y-1.0d0).lt.1.0e-14) then
            GNv = 1.0d0
         elseif (fabs(x).lt.1.0e-14) then
            GNv = -1.0d0
         else
            write(*,*) 'Invalid point in GNv: ',x,y
         endif
      endif

      endif

      return
      end

C FUNCTION: fabs

      double precision function fabs(x)

      double precision x

      if (x.ge.0) then
         fabs = x
      else
         fabs = -x
      endif

      return
      end
