/*
                 THE HILBERT CLASS LIBRARY                        
                  Release 1.0, July 2000

                  copyright (C) 1995-2000
           Mark S. Gockenbach and William W. Symes                 
                                                                  
The source code contained in this file is FREE SOFTWARE; you can
redistribute it and/or modify it under the terms of the GNU Library
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later
version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

A copy of the GNU Library General Public License resides in the root
directory of this package, as LICENSE.html, and can be viewed
conveniently with most web browsers, or less conveniently as an ASCII
text file. To receive a printed copy contact the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

*/

#include "EllipticSolOp3_d.h"
#include "HCL_LS_d.h"
#include "HCL_UMinTR_d.h"
#include "HCL_TrustRegionSolvers_d.h"

/* This is a test program that accompanies the tutorial

   "Implementing functionals in HCL", Technical Report TR99-22,
   Department of Computational and Applied Mathematics, Rice University,
   PO Box 1892, Houston, TX 77251-1892.

The purpose of this program is to exercise the operator class
EllipticSolOp3_d, which is thoroughly discussed in the above report.
This class represents the solution operator G of the standard elliptic
BVP (in divergence form)

          -div(a* grad( u)) = f in (0,1)x(0,1),
                          u = 0 on the boundary.

Both the coefficient a and the solution u are discretized by piecewise
linear finite finite elements, and G maps the coefficient a to the solution
u.

This program creates the an instance of the operator G and the coefficient
a, applies G to compute u (the data), and then solves for a by minimizing
the Output Least-Squares (OLS) functional

             ||G(a)-u||^2.
*/

// Every class derived from HCL_VectorSpace_d must have a Type field,
// which is a static char.  This allows a primitive form of RunTime Type
// Information.  Since the field is static, it must be declared in the
// main program---a constant irritation!

char HCL_RnSpace_d::Type;

main( int argc,char ** argv)
{
   if( argc != 2 )
   {
      cerr << "usage: main.x jobfile" << endl;
      exit(1);
   }

   // Create a Table from the jobfile; this table extracts keyword-value
   // pairs from the given file.  The only relevant keyword-value pair
   // needed for this program is the keyword "MeshFile", whose value is
   // the name of the file containing a description of the triangulation
   // of the domain.

   Table ParamTable( argv[1] );

   // Extract the name of the mesh file from the parameter table.

   char name[81];
   if( ParamTable.GetValue( "MeshFile",name ) )
   {
      cerr << "Error: 'MeshFile' missing from the job file " << argv[1]
           << endl;
      cerr << "Contents of job file:" << endl;
      ParamTable.Write( cerr );
      exit(1);
   }

   // This declaration calls the constructor for the operator class, which
   // reads the mesh description from the input file.

   EllipticSolOp3_d G( name );

   // For test purposes, we create a coefficient a.  This is done by
   // creating a vector in the domain of the operator, and then setting
   // all of its components to 1.0.

   HCL_Vector_d * a = G.Domain().Member();
   a->Fill( 1.0 );

   // Next, we create a vector to hold the data.

   HCL_Vector_d * u = G.Range().Member();

   // Compute the data u = G(a).

   G.Image( *a,*u );

   // Perturb a by 10% to get a starting value for the optimization.

   HCL_Vector_d * a1 = G.Domain().Member();
   a1->Random();
   a1->ScaleAdd( a->Norm()/a1->Norm()/10.0,*a );

   // Use the HCL least-squares functional class to represent the OLS
   // functional.

   HCL_LeastSquaresFcnl_d J( &G,u );

   // Now create a minimization algorithm object to solve the OLS
   // problem. We will use the HCL trust region algorithm (HCL_UMinTR_d)
   // with the Steihaug-Toint algorithm (HCL_TRCG_d) for solving the trust
   // region subproblem.
   //
   // Note that both algorithm classes can read algorithmic parameters
   // from an optional file; any parameters not defined in the file
   // (or all parameters if the file is not given) are assigned default
   // values.
   //
   // The parameters for a trust region solver are tagged with the
   // prefix "TRSolver::", while those for an unconstrained minimization
   // algorithm are tagged with "UMin::".  Thus the job file contains
   // lines like
   //
   //      TRSolver::MaxItn = 100
   //      UMin::MaxItn = 10
   //
   // The first indicates that the maximum number of iterations allowed
   // in each invocation of the trust region solver is 100, and the second
   // indicates that the maximum number of iterations allowed in the
   // execution of the minimization algorithm is 20.
   //
   // All parameters (their meanings and default values) are documented in
   // the HCL reference manual.
   

   HCL_TRCG_d trsolver( argv[1] );
   HCL_UMinTR_d umin( &trsolver,argv[1] );

   // Now solve the problem.  The following command instructs the minimization
   // object to attempt to find a local minimizer of J, using the current
   // value of a1 as the starting vector, and overwriting a1 with the computed
   // solution.

   umin.Minimize( J,*a1 );

   // Finally, compare the computed solution to the exact solution.

   a1->Sub( *a );
   cout << "Relative error in computed solution: " << a1->Norm()/a->Norm()
        << endl;
 
   // When a dynamically-allocated HCL object is no longer needed, it
   // is deallocated using the function HCL_delete.

   HCL_delete( u );
   HCL_delete( a );
   HCL_delete( a1 );
}
